// (c) 2001 Stefan Haustein
// Contributors: Bjorn Aadland

package org.kxml.wap;

import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import org.kxml.Attribute;
import org.kxml.Xml;
import org.kxml.io.ParseException;
import org.kxml.parser.AbstractXmlParser;
import org.kxml.parser.ParseEvent;
import org.kxml.parser.StartTag;
import org.kxml.parser.Tag;

/** Still Todo:
 * <ul>
 * <li>implement Processing Instructions</li>
 * <li>implement support for more than one codepages</li>
 * </ul>
 */

public class WbxmlParser extends AbstractXmlParser {
    
    InputStream in;
    
    // all known code pages and their tables
    Hashtable otherAttrStartTables = new Hashtable();
    Hashtable otherAttrValueTables = new Hashtable();
    Hashtable otherTagTables = new Hashtable();
    
    String[] attrStartTable;
    String[] attrValueTable;
    String[] tagTable;
    String stringTable;
    
    int version;
    int publicIdentifierId;
    String publicIdentifier=null;
    int charSet;
    
    
    StartTag current;
    ParseEvent next;
    boolean whitespace;
    
    private PublicIDEntry documentPubID=null;
    private PublicIDEntry currentPubID=null;
    
    
    public WbxmlParser(InputStream in) throws Exception {
        
        int publicIdentifierOffset=-1;
        
        this.in = in;
        
        version = readByte();
        publicIdentifierId = readInt();
        
        if (publicIdentifierId == 0)
            publicIdentifierOffset=readInt();
        
        charSet = readInt(); // skip charset
        
        int strTabSize = readInt();
        
        StringBuffer buf = new StringBuffer(strTabSize);
        
        for (int i = 0; i < strTabSize; i++)
            buf.append((char)readByte());
        
        stringTable = buf.toString();
        
        // initialisation of tables and public identifier objects
        WbxmlInitialiser initialiser=null;
        if(publicIdentifierId!=0) {
            initialiser=WbxmlInitialiserFactory.getInitialiserByPublicIdentifierCode(publicIdentifierId);
        }
        else {
            if(publicIdentifierOffset>=0) {
                extractInlinePublicID(publicIdentifierOffset);
                if (publicIdentifier!=null) {
                    initialiser=WbxmlInitialiserFactory.getInitialiserByPublicIdentifier(publicIdentifier);
                }
            }
        }
        if(initialiser!=null) {
            initialiser.initialise(this);
        }
        
        // load the public identifier
        if(publicIdentifier==null) {
            documentPubID=getPublicIDEntryByCode(publicIdentifierId);
        }
        else {
            documentPubID=getPublicIDEntry(publicIdentifier);
        }
        
    }
    
    private void extractInlinePublicID(int anOffset) {
        publicIdentifier=this.extractFromStrT(anOffset);
    }
    
    public ParseEvent peek() throws IOException {
        
        String s;
        
        if (next != null)
            return next;
        
        if (current != null && current.getDegenerated()) {
            next = new Tag
            (Xml.END_TAG, current,
            current.getNamespace(),
            current.getName());
            
            current = current.getParent();
            return next;
        }
        
        ParseEvent result = null;
        
        do {
            int id = in.read();
            
            switch (id) {
                case -1:
                    if (current != null)
                        throw new RuntimeException("unclosed elements: " + current);
                    
                    next = new ParseEvent(Xml.END_DOCUMENT, null);
                    break;
                    
                case Wbxml.SWITCH_PAGE:
                    int page = readByte();
                    switchPage(page);
                    break;
                    
                case Wbxml.END:
                    next = new Tag(Xml.END_TAG, current,
                    current.getNamespace(),
                    current.getName());
                    
                    current = current.getParent();
                    break;
                    
                case Wbxml.ENTITY:
                    next = new ParseEvent(Xml.TEXT, "" + (char)readInt());
                    break;
                    
                case Wbxml.STR_I: {
                    s = readStrI();
                    next = new ParseEvent(whitespace ? Xml.WHITESPACE : Xml.TEXT, s);
                    break;
                }
                
                case Wbxml.EXT_I_0:
                case Wbxml.EXT_I_1:
                case Wbxml.EXT_I_2:
                case Wbxml.EXT_T_0:
                case Wbxml.EXT_T_1:
                case Wbxml.EXT_T_2:
                case Wbxml.EXT_0:
                case Wbxml.EXT_1:
                case Wbxml.EXT_2:
                case Wbxml.OPAQUE:
                    next = parseWapExtension(id);
                    break;
                    
                case Wbxml.PI:
                    throw new RuntimeException("PI curr. not supp.");
                    // readPI;
                    // break;
                    
                case Wbxml.STR_T: {
                    int pos = readInt();
                    int end = stringTable.indexOf('\0', pos);
                    next = new ParseEvent
                    (Xml.TEXT, stringTable.substring(pos, end));
                    break;
                }
                
                default:
                    next = parseElement(id);
            }
        }
        while (next == null);
        
        return next;
    }
    
    public ParseEvent read() throws IOException {
        if (next == null)
            peek();
        ParseEvent result = next;
        next = null;
        return result;
    }
    
    
    /**
     * Switches to code page <code>page</code>
     *
     * @param page the page to switch to
     */
    private void switchPage(int page){
        Integer ipage = new Integer(page);
        attrStartTable = (String[]) otherAttrStartTables.get(ipage);
        attrValueTable = (String[]) otherAttrValueTables.get(ipage);
        tagTable = (String[]) otherTagTables.get(ipage);
        
        currentPubID=getPublicIDEntryByIndex(page);
    }
    
    /** For handling wap extensions in attributes, overwrite this
    method, call super and return a corresponding TextEvent. */
    
    public ParseEvent parseWapExtension(int id) throws IOException {
        
        switch (id) {
            case Wbxml.EXT_I_0:
            case Wbxml.EXT_I_1:
            case Wbxml.EXT_I_2:
                return new WapExtensionEvent(id, readStrI());
                
            case Wbxml.EXT_T_0:
            case Wbxml.EXT_T_1:
            case Wbxml.EXT_T_2:
                return new WapExtensionEvent(id, new Integer(readInt()));
                
            case Wbxml.EXT_0:
            case Wbxml.EXT_1:
            case Wbxml.EXT_2:
                return new WapExtensionEvent(id, null);
                
            case Wbxml.OPAQUE: {
                int len = readInt();
                byte[] buf = new byte[len];
                
                for (int i = 0; i < len; i++)   // enhance with blockread!
                    buf[i] = (byte)readByte();
                
                return new WapExtensionEvent(id, buf);
            } // case OPAQUE
        } // SWITCH
        
        throw new IOException("illegal id!");
    }
    
    public Vector readAttr() throws IOException {
        
        Vector result = new Vector();
        
        int id = readByte();
        
        while (id != 1) {
            
            String name = resolveId(attrStartTable, id);
            StringBuffer value;
            
            int cut = name.indexOf('=');
            
            if (cut == -1)
                value = new StringBuffer();
            else {
                value = new StringBuffer(name.substring(cut + 1));
                name = name.substring(0, cut);
            }
            
            id = readByte();
            while (id > 128 || id == Wbxml.ENTITY
            || id == Wbxml.STR_I || id == Wbxml.STR_T
            || (id >= Wbxml.EXT_I_0 && id <= Wbxml.EXT_I_2)
            || (id >= Wbxml.EXT_T_0 && id <= Wbxml.EXT_T_2)) {
                
                switch (id) {
                    case Wbxml.ENTITY:
                        value.append((char)readInt());
                        break;
                        
                    case Wbxml.STR_I:
                        value.append(readStrI());
                        break;
                        
                    case Wbxml.EXT_I_0:
                    case Wbxml.EXT_I_1:
                    case Wbxml.EXT_I_2:
                    case Wbxml.EXT_T_0:
                    case Wbxml.EXT_T_1:
                    case Wbxml.EXT_T_2:
                    case Wbxml.EXT_0:
                    case Wbxml.EXT_1:
                    case Wbxml.EXT_2:
                    case Wbxml.OPAQUE:
                        
                        ParseEvent e = parseWapExtension(id);
                        if (!(e.getType() != Xml.TEXT
                        && e.getType() != Xml.WHITESPACE))
                            throw new RuntimeException("parse WapExtension must return Text Event in order to work inside Attributes!");
                        
                        value.append(e.getText());
                        
                        break;
                        
                    case Wbxml.STR_T:
                        value.append(readStrT());
                        break;
                        
                    default:
                        value.append(resolveId(attrValueTable, id));
                }
                
                id = readByte();
            }
            
            result.addElement(new Attribute(null, name, value.toString()));
        }
        
        return result;
    }
    
    String resolveId(String[] tab, int id) throws IOException {
        int idx = (id & 0x07f) - 5;
        if (idx == -1)
            return readStrT();
        if (idx < 0 || tab == null
        || idx >= tab.length || tab[idx] == null)
            throw new IOException("id " + id + " undef.");
        
        return tab[idx];
    }
    
    StartTag parseElement(int id) throws IOException {
        
        String tag = resolveId(tagTable, id & 0x03f);
        
        String ns = null;
        
        Vector attributes = null;
        
        // load the attributes if appropriate
        if((id & 128) != 0) {
            attributes=readAttr();
        }
        
        // now pickup the current namespace
        if(documentPubID!=null) {
            if(documentPubID!=currentPubID) {
                if(currentPubID==null) { currentPubID=documentPubID; }
                ns=currentPubID.getNameSpace();
            }
        }
        
        if(ns!=null) {
            if(attributes==null) {
                attributes=new Vector();
            }
            int i=0;
            Attribute attr=null;
            for(i=0;i<attributes.size();i++) {
                attr = (Attribute)attributes.elementAt(i);
                if(attr.getName().equals("xmlns")) {
                    break;
                }
            }
            
            // if i less than the length of the Vector than we've
            // found a namespace
            if(i<attributes.size()) {
                attributes.remove(i);
                attributes.insertElementAt(new Attribute(null, "xmlns", ns),i);
            }
            else {
                attributes.insertElementAt(new Attribute(null, "xmlns", ns),0);
            }
            
        }
        
        // ok, now let's care about attrs etc
        
        try {
            current = new StartTag
            (current, // previous
            ns, // namespace
            tag, // name
            attributes, // attributes
            (id & 64) == 0, // degenerated
            processNamespaces);  // processing
        }
        catch (Exception e) {
            throw new ParseException(null, e, -1, -1);
        }
        
        return current;
    }
    
    int readByte() throws IOException {
        int i = in.read();
        if (i == -1)
            throw new IOException("Unexpected EOF");
        return i;
    }
    
    int readInt() throws IOException {
        int result = 0;
        int i;
        
        do {
            i = readByte();
            result = (result << 7) | (i & 0x7f);
        }
        while ((i & 0x80) != 0);
        
        return result;
    }
    
    String readStrI() throws IOException {
        StringBuffer buf = new StringBuffer();
        boolean wsp = true;
        while (true) {
            int i = in.read();
            if (i == -1)
                throw new IOException("Unexpected EOF");
            if (i == 0)
                break;
            if (i > 32)
                wsp = false;
            buf.append((char)i);
        }
        this.whitespace = wsp;
        return buf.toString();
    }
    
    String readStrT() throws IOException {
        int pos = readInt();
        return extractFromStrT(pos);
    }
    
    private String extractFromStrT(int pos) {
        int end = stringTable.indexOf('\0', pos);
        
        end=end<0?stringTable.length():end;
        
        return stringTable.substring(pos, end);
    }
    
    /**
     * Sets the tag table for a given page.
     *	The first string in the array defines tag 5, the second tag 6 etc.
     *  Currently, only page 0 is supported
     *
     * @param page the page referring to
     * @param tagTable a table filled with tags
     */
    public void setTagTable(int page, String [] tagTable)  {
        otherTagTables.put(new Integer(page),tagTable);
        if(page == 0){
            this.tagTable = tagTable;
        }
    }
    
    
    /**
     * Sets the attribute start Table for a given page.
     *	The first string in the array defines attribute
     *  5, the second attribute 6 etc.
     *  Please use the character '=' (without quote!) as delimiter
     *  between the attribute name and the (start of the) value
     *
     * @param page the page referring to
     * @param attrStartTable a table filled with attribute names
     */
    public void setAttrStartTable(int page, String [] attrStartTable) {
        otherAttrStartTables.put(new Integer(page),attrStartTable);
        if(page == 0){
            this.attrStartTable = attrStartTable;
        }
    }
    
    /**
     * Sets the attribute value Table for a given page.
     *	The first string in the array defines attribute value 0x85,
     *  the second attribute value 0x86 etc.
     *
     * @param page the page referring to
     * @param attrValueTable a table filled with attribute values
     */
    public void setAttrValueTable(int page, String [] attrValueTable) {
        otherAttrValueTables.put(new Integer(page),attrValueTable);
        if(page == 0){
            this.attrValueTable = attrValueTable;
        }
    }
    
    /* Public Identifier information lookup */
    private HashMap myPublicIDMap=null;
    private HashMap myPublicIDCodeMap=null;
    private HashMap myPublicIDIndexMap=null;
    
    void addPublicIDEntry(PublicIDEntry anEntry) {
        if(myPublicIDMap==null) { myPublicIDMap=new HashMap(); }
        if(myPublicIDCodeMap==null) { myPublicIDCodeMap=new HashMap(); }
        if(myPublicIDIndexMap==null) { myPublicIDIndexMap=new HashMap(); }
        
        if(anEntry.getPublicIdentifier()!=null) {
            myPublicIDMap.put(anEntry.getPublicIdentifier().toUpperCase(), anEntry);
        }
        
        if(anEntry.getPublicIDCode()!=null) {
            myPublicIDMap.put(anEntry.getPublicIDCode(), anEntry);
        }
        
        myPublicIDIndexMap.put(new Integer(anEntry.getStrTableIndex()), anEntry);
        
    }
    
    public PublicIDEntry getPublicIDEntry(String anIdentifier) {
        return (PublicIDEntry) myPublicIDMap.get(anIdentifier.toUpperCase());
    }
    
    public PublicIDEntry getPublicIDEntryByCode(int aCode) {
        return (PublicIDEntry) myPublicIDCodeMap.get(new Integer(aCode));
    }
    
    public PublicIDEntry getPublicIDEntryByIndex(int anIndex) {
        
        return (PublicIDEntry) myPublicIDIndexMap.get(new Integer(anIndex));
    }
    
}






