/*

 $Id: AbstractClient.java,v 1.12 2002/08/09 15:16:32 stefano_fornari Exp $

 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.client;

import sync4j.framework.core.*;

import java.util.Vector;
import java.util.Enumeration;

/**
 *
 *  abstract data synchronization client
 *
 *  @author Sean C. Sullivan
 *
 */
public abstract class AbstractClient
{
    private final IClientConnectionFactory[] m_connectionFactories;
    private final Vector m_clientListeners = new Vector();
    private final String m_strServerAddress;

    /**
     *
     *  @param connectionFactories array length must be greater than zero.
     *           Elements of the array must be non-null.
     *  @param strServerAddress server address. Must be non-null.
     *
     */
    public AbstractClient(
            final IClientConnectionFactory[] connectionFactories,
            final String strServerAddress)
    {
        // todo : code here
        if (connectionFactories == null)
        {
            throw new NullPointerException(
                    "IClientConnectionFactory[] parameter is null");
        }
        m_connectionFactories = (IClientConnectionFactory[]) connectionFactories.clone();
        if (m_connectionFactories.length < 1)
        {
            throw new IllegalArgumentException(
                    "IClientConnectionFactory[] array contains < 1 elements");
        }
        for (int i = 0; i < m_connectionFactories.length; i++)
        {
            if (m_connectionFactories[i] == null)
            {
                throw new NullPointerException(
                        "IClientConnectionFactory["
                        + i
                        + "] is null");
            }
        }

        if (strServerAddress == null)
        {
            throw new NullPointerException(
                    "strServerAddress parameter is null");
        }
        m_strServerAddress = strServerAddress;
    }

    // todo : add more methods

    /**
     *
     * @return client connection factory objects
     *
     */
    protected IClientConnectionFactory[] getConnectionFactories()
    {
        return (IClientConnectionFactory[]) m_connectionFactories.clone();
    }

    /**
     *
     * @param listener object that implements the proper callback interface.
     *              Must be non-null
     * @see #removeClientListener
     *
     */
    public void addClientListener(final IClientListener listener)
    {
        if (listener == null)
        {
            throw new NullPointerException(
                    "IClientListener parameter is null");
        }
        m_clientListeners.addElement(listener);
    }

    /**
     *
     * @param listener object was previously registered by a call to
     *      {@link #addClientListener}
     *
     * @see #addClientListener
     *
     */
    public void removeClientListener(final IClientListener listener)
    {
        if (listener != null)
        {
            m_clientListeners.removeElement(listener);
        }
    }

    /**
     *
     * @param evt event.  Must be non-null.
     *
     */
    protected void fireClientEvent(final ClientEvent evt)
    {
        if (evt == null)
        {
            throw new NullPointerException("evt parameter is null");
        }

        Enumeration e = m_clientListeners.elements();
        while (e.hasMoreElements())
        {
            IClientListener listener;
            listener = (IClientListener) e.nextElement();
            listener.syncClientUpdate(evt);
        }
    }

    /**
     *
     * connect to the data synchronization server
     *
     */
    protected IClientConnection connectToServer()
            // todo : throws exception(s)
    {
        IClientConnectionFactory[] factories;
        factories = this.getConnectionFactories();

        IClientConnection conn = null;

        for (int i = 0; i < factories.length; i++)
        {
            try
            {
                conn = factories[i].getClientConnection(
                        this.getServerAddress());
            }
            catch (Exception ex)
            {
                // ignore
            }
        }

        if (conn == null)
        {
            // todo : throw a real exception here
            throw new RuntimeException();
        }

        return conn;
    }

    /**
     *
     * @return the data synchronization server's address
     *
     */
    public String getServerAddress()
    {
        return m_strServerAddress;
    }

    /**
     *
     * @param type must be non-null
     *
     */
    public /* todo : make final? */ void doSync(final SyncType type)
            // todo : throw Exception(s)
    {
        if (type == null)
        {
            throw new NullPointerException("SyncType parameter is null");
        }

        final IClientConnection conn = this.connectToServer();

        if (type == SyncType.TWO_WAY)
        {
            this.doTwoWaySync(conn);
        }
        else if (type == SyncType.SLOW)
        {
            this.doSlowSync(conn);
        }
        else if (type == SyncType.ONE_WAY_FROM_CLIENT)
        {
            this.doOneWayFromClientSync(conn);
        }
        else if (type == SyncType.REFRESH_FROM_CLIENT)
        {
            this.doRefreshFromClientSync(conn);
        }
        else if (type == SyncType.ONE_WAY_FROM_SERVER)
        {
            this.doOneWayFromServerSync(conn);
        }
        else if (type == SyncType.REFRESH_FROM_SERVER)
        {
            this.doRefreshFromServerSync(conn);
        }
        // todo : add code for Server alerted sync ... - ?
        // todo : else if ...
        // todo : else ...
    }

    protected void doSlowSync(IClientConnection conn)
    {
        System.out.println(conn);
    }

    protected void doTwoWaySync(IClientConnection conn)
    {
        System.out.println(conn);
    }

    protected void doOneWayFromClientSync(IClientConnection conn)
    {
        System.out.println(conn);
    }

    protected void doRefreshFromClientSync(IClientConnection conn)
    {
        System.out.println(conn);
    }

    protected void doOneWayFromServerSync(IClientConnection conn)
    {
        System.out.println(conn);
    }

    protected void doRefreshFromServerSync(IClientConnection conn)
    {
        System.out.println(conn);
    }

}
