/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.config;

import java.io.InputStream;
import java.io.IOException;

import java.net.URL;

import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.util.Collections;

import sync4j.framework.tools.beans.BeanFactory;
import sync4j.framework.config.ConfigurationException;

import org.apache.commons.lang.StringUtils;

/**
 * Incapsulates all the configuration information about a <i>Sync4jEngine</i>.
 * Configuration parameters are stored as properties in a Map structure. The
 * key is the name of the parameter and it is strucured in dotted separated
 * namespaces. The value of the parameter is the value of the key and can be of
 * any type. Accessor methods are provided to get the value of the parameter in
 * a particular type (such as String, int, double, boolean, ...).<br>
 * Access to getXXX() and setXXX() methods are synchronized.
 *
 * @author  Stefano Fornari @ funambol
 *
 * @version $Id: Configuration.java,v 1.6 2003/12/15 16:54:25 luigiafassina Exp $
 */
public class Configuration
implements java.io.Serializable {

    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data

    private static final Map internalMap = Collections.synchronizedMap(new HashMap());

    // -------------------------------------------------------------- properties

    private transient ClassLoader classLoader = null;

    /** Getter for property classLoader.
     * @return Value of property classLoader.
     *
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /** Setter for property classLoader.
     * @param classLoader New value of property classLoader.
     *
     */
    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    // ------------------------------------------------------------ Constructors

    public Configuration() {
    }

    // ---------------------------------------------------------- Public methods

    public String getStringValue(String name, String def)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        return (value == null) ? def : value;
    }

    public String getStringValue(String name)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return value;
    }

    /**
     * Turns a character separated values property into an array of String
     * objects given the separator. For example:
     * <pre>
     *      one;two;trhee
     * </pre>
     * will result in
     * <pre
     *      {"one", "two", "trhee"}
     * </pre>
     *
     * @param name the name of the property
     * @param separator the separator character
     *
     * @return an array of Strings
     */
    public String[] getStringArray(String name, char separator)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        //
        // NOTE: String.split() is not used for performance reason. We do not
        //       need a regular expression here.
        //
        return StringUtils.split(value, String.valueOf(separator));
    }

    public int getIntValue(String name, int def)
    throws ConfigurationException {
        Integer value = (Integer)internalMap.get(name);

        return (value == null) ? def : value.intValue();
    }

    public int getIntValue(String name)
    throws ConfigurationException {
        Integer value = (Integer)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return value.intValue();
    }

    public boolean getBoolValue(String name, boolean def)
    throws ConfigurationException {
        Boolean value = (Boolean)internalMap.get(name);

        return (value == null) ? def : value.booleanValue();
    }

    public boolean getBoolValue(String name)
    throws ConfigurationException {
        Boolean value = (Boolean)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return value.booleanValue();
    }

    public Object getValue(String name, Object def)
    throws ConfigurationException {
        Object value = internalMap.get(name);

        return (value == null) ? def : value;
    }

    public Object getValue(String name)
    throws ConfigurationException {
        Object value = internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return value;
    }

    /**
     * It is supposed that the value of the parameter is a class name. It returns
     * an instance of that class, created with <i>newInstance()</i>.
     *
     * @param name the name of the parameter
     *
     * @return an instance of the class specified by the parameter
     *
     * @throws ConfigurationException in case of errors
     */
    public Object getClassInstance(String name)
    throws ConfigurationException {
       String value = (String)internalMap.get(name);

       if (value == null) {
           throw new ConfigurationException("No " + name + " property defined");
       }

       try {
           return Class.forName(value).newInstance();
       } catch (Exception e) {
           throw new ConfigurationException( "Error in creating an instance of "
                                           + value
                                           , e                                );
       }
    }

    /**
     * It is supposed that the value of the parameter is a bean name. It returns
     * that bean, created with <i>BeanFactory.getBeanInstance()</i>.
     *
     * @param name the name of the parameter
     *
     * @return an instance of the class specified by the parameter
     *
     * @throws ConfigurationException in case of errors
     */
    public Object getBeanInstance(String name)
    throws ConfigurationException {
       String value = (String)internalMap.get(name);

       if (value == null) {
           throw new ConfigurationException("No " + name + " property defined");
       }

       try {
           return BeanFactory.getBeanInstance(classLoader, value);
       } catch (Exception e) {
           throw new ConfigurationException( "Error in creating an instance of "
                                           + value
                                           , e                                );
       }
    }

    /**
     * Create and return an array of bean as specified by a list of bean names.
     *
     * @param name the name of the paramenter
     * @param separator the separator character
     *
     */
    public Object[] getBeanArray(String name, char separator)
    throws ConfigurationException {
        String[] values = getStringArray(name, separator);

        Object[] ret = new Object[values.length];

        int i = 0;
        try {
            for (; i<ret.length; ++i) {
                ret[i] = BeanFactory.getBeanInstance(classLoader, values[i]);
            }
        } catch (Exception e) {
            throw new ConfigurationException ( "Error in creating the bean "
                                             + values[i]
                                             , e                           );
        }

        return ret;
    }

    public synchronized void load(InputStream is) throws IOException {
        Properties properties = new Properties();

        properties.load(is);

        internalMap.putAll(properties);
    }

    public synchronized void load(String uri)
    throws IOException {
        load(new URL(uri).openStream());
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append(getClass().getName()).append(" - ");
        sb.append(internalMap.toString()).append(" - ");
        sb.append("classLoader: " + classLoader);

        return sb.toString();
    }

}
