/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltAtomic&gt tag in the SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari
 *  
 *  @version $Id: AtomicCommand.java,v 1.4 2003/12/15 16:55:09 luigiafassina Exp $
 *
 */
public final class AtomicCommand
    extends AbstractCommand
{
    public static String COMMAND_NAME = "Atomic";

    private boolean m_bNoResponse;
    private Meta m_meta;
    private AbstractCommand[] m_commands;

    /**
     *
     *  @param cmdid command identifier. Must be non-null.
     *  @param bNoResponse true if no response is required.
     *  @param m meta data. May be null.
     *  @param commands array of commands. Must be non-null. Elements of the array must be non-null.
     *                  Array length must be greater than zero. The element objects
     *                  must be of the following types: {@link AddCommand}, {@link DeleteCommand},
     *                  {@link CopyCommand}, {@link MapCommand}, {@link ReplaceCommand},
     *                  {@link SequenceCommand}, or {@link SyncCommand}
     *
     */
    public AtomicCommand(
            final CommandIdentifier cmdid,
            final boolean bNoResponse,
            final Meta m,
        final AbstractCommand[] commands)
    {
        super(cmdid);

        if (commands == null)
        {
            throw new NullPointerException(
                    "commands array parameter is null");
        }

        m_bNoResponse = bNoResponse;
        // todo : different Meta handling code?
        m_meta = m;
        m_commands = (AbstractCommand[]) commands.clone();
        if (m_commands.length < 1)
        {
            throw new IllegalArgumentException(
              "AbstractCommand[] array length is < 1");
        }
        for (int i = 0; i < m_commands.length; i++)
        {
            if (m_commands[i] == null)
            {
                throw new NullPointerException(
                "AbstractCommand array contains a null value, index = "
                        + i);
            }
            else if ((!(m_commands[i] instanceof AddCommand))
                    && (!(m_commands[i] instanceof DeleteCommand))
                    && (!(m_commands[i] instanceof CopyCommand))
                    && (!(m_commands[i] instanceof MapCommand))
                    && (!(m_commands[i] instanceof ReplaceCommand))
                    && (!(m_commands[i] instanceof SequenceCommand))
                    && (!(m_commands[i] instanceof SyncCommand)))
            {
                throw new IllegalArgumentException(
                        "illegal nested command: "
                        + m_commands[i]);
            }
        }
        // Note: the SyncML represent DTD says that Atomic
        //       commands are allowed inside of another Atomic
        //       command. However, the SyncML PDF
        //       documentation says that it is illegal to
        //       nest Atomic commands.  syncml.org confirmed via
        //       email that nested Atomic commands are illegal.
    }

    /**
     *
     *  @param eAtomic element &ltAtomic&gt
     *
     */
    AtomicCommand(final Element eAtomic)
            throws RepresentationException
    {
        // todo : verify that eAtomic is <Atomic>
        this(Util.extractCommandId(eAtomic, true),
                Util.extractNoResp(eAtomic, false),
                Util.extractMeta(eAtomic, false),
            Util.extractAbstractCommands(eAtomic, true));
    }

    /**
     *
     *  @return true if no response is required
     *
     */
    public boolean getNoResponse()
    {
        return m_bNoResponse;
    }

    /**
     *
     *  @return meta object or null
     *
     */
    public Meta getMeta()
    {
        return m_meta;
    }

    /**
     *
     *  @return an array of command objects. The return value is guaranteed to be non-null.
     *
     */
    public AbstractCommand[] getCommands()
    {
        return (AbstractCommand[]) m_commands.clone();
    }

    String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<Atomic>");
        sbXML.append(this.getCommandIdentifier().toXML());
        if (this.getNoResponse())
        {
            sbXML.append("<NoResp/>\n");
        }
        if (this.getMeta() != null)
        {
            sbXML.append(this.getMeta().toXML());
        }
        AbstractCommand[] cmds;
        cmds = this.getCommands();
        for (int i = 0; i < cmds.length; i++)
        {
            sbXML.append(cmds[i].toXML());
        }
        sbXML.append("</Atomic>\n");
        return sbXML.toString();
    }

        public String getName() {
            return AtomicCommand.COMMAND_NAME;
        }


}
