/*
 Copyright (c) 2003-2004 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 * This class represents an authentication challenge element. It  corresponds 
 * to the &ltChal&gt element in SyncML representation DTD
 *
 *  @author Stefano Fornari @ Funambol
 *  @see    Credential
 *
 *  @version $Id: AuthenticationChallenge.java,v 1.4 2003/11/02 15:07:44 stefano_fornari Exp $
 *
 */
public final class AuthenticationChallenge
{
    private String type;
    private String format;
    private NextNonce nextNonce;

    /**
     * Constructs a new AuthenticationChallange object.
     *
     * @param type The authentication type (see Constants) - NOT NULL
     * @param format The encoding format (suzh as b64) - NOT NULL
     * @param nextNonce next nonce value - NULL
     *
     */
    public AuthenticationChallenge( final String    type     ,
                                    final String    format   ,
                                    final NextNonce nextNonce) {
        if (type == null) {
            throw new NullPointerException("The authentication type cannot be null");
        }
        if (format == null) {
            throw new NullPointerException("The authentication format cannot be null");
        }
        
        this.type      = type     ;
        this.format    = format   ;
        this.nextNonce = nextNonce;
    }

    /**
     * Constructs a new AuthenticationChallenge object from its XML representation.
     *
     * @throws RepresentationException in case of errors
     */
    AuthenticationChallenge(final Element chal)
    throws RepresentationException {
        this(
            Util.extractType(Util.extractElement(chal, "Meta", true), true),
            Util.extractFormat(Util.extractElement(chal, "Meta", true), true),
            Util.extractNextNonce(Util.extractElement(chal, "Meta", true), false)
        );
    }

    /**
     * Returns the nextNonce property or null
     *
     *  @return the nextNonce property or null
     */
    public NextNonce getNextNonce() {
        return nextNonce;
    }

    /**
     * Returns the authentication type
     *
     * @return authentication type. 
     *
     */
    public String getType() {
        return type;
    }

    /**
     * Returns the authentication format
     *
     * @return format the authentication format
     *
     */
    public String getFormat() {
        return format;
    }
    
    /**
     * Creates a basic authentication challange.
     * This will have type = Constants.AUTH_TYPE_BASIC and 
     * format = Constants.FORMAT_B64
     *
     * @return the newly created AuthenticationChallange
     */
    public static AuthenticationChallenge getBasicAuthenticationChallenge() {
        return new AuthenticationChallenge(
            Constants.AUTH_TYPE_BASIC, 
            Constants.FORMAT_B64,
            null
        );
    }


    /**
     * Returns the XML representation of this object
     *
     * @return an XML representation of this object
     *
     */
    public String toXML() {
        StringBuffer sbXML = new StringBuffer();

        sbXML.append("<Chal><Meta>\n")
             .append("<Type xmlns='syncml:metinf'>").append(type).append("</Type>\n")
             .append("<Format xmlns='syncml:metinf'>").append(format).append("</Format>\n");
        
        if (nextNonce != null) {
            sbXML.append("<NextNonce xmlns='syncml:metinf'>")
                 .append(nextNonce)
                 .append("</NextNonce>\n");
        }
        sbXML.append("</Meta></Chal>\n");
        
        return sbXML.toString();
    }


}
