/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */


package sync4j.framework.core;

import sync4j.framework.core.Constants;
import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to &ltCred&gt element in SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @see    AuthenticationChallenge
 *
 *  @version  $Id: Credential.java,v 1.9 2003/12/15 16:55:09 luigiafassina Exp $
 *
 */
public final class Credential
implements java.io.Serializable {
    private Authentication auth;
    
    /**
     *
     *   @param auth must be non-null. This parameter will usually be an
     *               instance of {@link BasicAuthentication} or {@link MD5Authentication}
     *
     *
     */
    public Credential(final Authentication auth) {
        if (auth == null) {
            throw new NullPointerException("auth parameter is null");
        }
        
        this.auth = auth;
    }

    // todo (?) : add a constructor with this signature
    //        public Credential(final Meta m, final Data d)
    
    /**
     *
     *  @param eCred &ltCred&gt element. Must be non-null.
     *
     */
    Credential(final Element eCred)
    throws RepresentationException {
        // todo : verify that eCred is a <Cred>
        this(Util.extractElement(eCred, "Meta", false),
        Util.extractData(eCred, true));
    }
    
    /**
     *
     *  @param eMeta meta information. May be null.
     *  @param d May be null. (TODO: is this right?)
     *
     */
    Credential(final Element eMeta, Data d)
    throws RepresentationException {
        String format, type;
        //
        // note: the first param (Element eMeta) may be null
        //
        if (eMeta == null) {
            format = null;
            type = null;
        }
        else {
            format = Util.extractFormat(eMeta, false);
            type = Util.extractType(eMeta, false);
        }
        auth = createAuthentication(d.getValue(), type);
    }
    
    /**
     *
     *  @return may return null
     *
     */
    public String getType() {
        return auth.getType();
    }
    
    /**
     *
     *  @return may return null
     *
     */
    public String getFormat() {
        return auth.getFormat();
    }
    
    /**
     *
     *  @return may return null (TODO: is this right?)
     *
     */
    // todo : should getData return a Data object?
    public String getData() {
        return auth.getData();
    }
    
    /**
     * Return the username stored in this credential
     *
     * @return the username stored in this credential
     */
    public String getUsername() {
        if (auth instanceof BasicAuthentication) {
            return ((BasicAuthentication)auth).getUsername();
        } else if (auth instanceof ClearAuthentication) {
            return ((ClearAuthentication)auth).getUsername();
        }
        
        return null;
    }
    
    /**
     *
     *  @return an XML representation of this object
     *
     */
    String toXML() {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<Cred>\n");
        if ((this.getType() != null)
        || (this.getFormat() != null)) {
            sbXML.append("<Meta>\n");
            if (this.getType() != null) {
                sbXML.append("<Type>");
                sbXML.append(this.getType());
                sbXML.append("</Type>\n");
            }
            if (this.getFormat() != null) {
                sbXML.append("<Format>");
                sbXML.append(this.getFormat());
                sbXML.append("</Format>\n");
            }
            sbXML.append("</Meta>\n");
        }
        sbXML.append("<Data>");
        sbXML.append(this.getData());
        sbXML.append("</Data>\n");
        sbXML.append("</Cred>\n");
        return sbXML.toString();
    }
    
    /**
     * Creates and returns a "guest" credential.
     *
     * TO DO: MD5 authentication support
     *
     * @return the newly created credential
     */
    public static Credential getGuestCredential() {
        try {
            String guest = "guest:guest";
            
            return new Credential(
                       createAuthentication(guest, Constants.AUTH_TYPE_CLEAR)
                   );
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }
    
    /**
     * Create and return the Authentication object corresponding to the given
     * type and data.
     *
     * @param type the type of the required Authentication object. One of the 
              values AUTH_TYPES_XXX defined in sync4j.framework.core.Constants
     * @param data thedata to be interpreted based on the type
     *
     * @return the corresponding Authentication object.
     */
    public static Authentication createAuthentication(String data,
                                                      String type) {
        if (Constants.AUTH_SUPPORTED_TYPES.indexOf(type) < 0) {
            return new BasicAuthentication(data, false);
        }
        
        if (Constants.AUTH_TYPE_BASIC.equals(type)) {
            return new BasicAuthentication(data, false);
        } else if (Constants.AUTH_TYPE_CLEAR.equals(type)) {
            return new ClearAuthentication(data);
        }
        
        //
        // 
        return null;
    }
}
