/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */


package sync4j.framework.core;

import sync4j.framework.core.RepresentationException;
import sync4j.framework.core.xml.Element;

/**
 *
 *  corresponds to &ltDataStore&gt element in SyncML devinf DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: DataStore.java,v 1.6 2003/12/15 16:55:09 luigiafassina Exp $
 */
public final class DataStore {
    private SourceRef m_sourceRef;
    private String m_strDisplayName;
    private long m_lMaxGuidSize;
    private ContentTypeInfo m_preferredReceiveContentTypeInfo;
    private ContentTypeInfo[] m_supportedReceiveContentTypeInfo;
    private ContentTypeInfo m_preferredTransmitContentTypeInfo;
    private ContentTypeInfo[] m_supportedTransmitContentTypeInfo;
    private DataStoreMemoryInfo m_dataStoreMemoryInfo;
    private SyncCapabilities m_syncCapabilities;
    
    /**
     *
     * @param sourceRef Must be non-null.
     * @param strDisplayName May be null.
     * @param lMaxGuidSize maximum GUID size. Set to -1 if the Maximum GUID size is unknown or unspecified.
     *             Otherwise, this parameter should be a positive number.
     * @param preferredReceiveContentTypeInfo Must be non-null.
     * @param supportedReceiveContentTypeInfo Must be non-null. Array elements must be non-null.
     * @param preferredTransmitContentTypeInfo Must be non-null.
     * @param supportedTransmitContentTypeInfo Must be non-null. Array elements must be non-null.
     * @param dsmi datastore memory info. May be null.
     * @param syncCap Must be non-null.
     *
     */
    public DataStore(
    final SourceRef sourceRef,
    final String strDisplayName /* may be null */,
    final long lMaxGuidSize,
    final ContentTypeInfo preferredReceiveContentTypeInfo,
    final ContentTypeInfo[] supportedReceiveContentTypeInfo,
    final ContentTypeInfo preferredTransmitContentTypeInfo,
    final ContentTypeInfo[] supportedTransmitContentTypeInfo,
    final DataStoreMemoryInfo dsmi /* may be null */,
    final SyncCapabilities syncCap /* may be null */) {
        if (sourceRef == null) {
            throw new NullPointerException(
            "SourceRef parameter is null");
        }
        
        if ((lMaxGuidSize == 0)
        || (lMaxGuidSize < -1)) {
            throw new IllegalArgumentException(
            "illegal parameter value: "
            + "lMaxGuidSize = "
            + lMaxGuidSize);
        }
        
        if (preferredReceiveContentTypeInfo == null) {
            throw new NullPointerException("preferredReceiveContentTypeInfo parameter");
        }
        
        if (preferredTransmitContentTypeInfo == null) {
            throw new NullPointerException("preferredTransmitContentTypeInfo parameter");
        }
        
        if (syncCap == null) {
            throw new NullPointerException("syncCap parameter");
        }
        
        m_sourceRef = sourceRef;
        
        m_strDisplayName = strDisplayName;
        
        m_lMaxGuidSize = lMaxGuidSize;
        
        m_preferredReceiveContentTypeInfo = preferredReceiveContentTypeInfo;
        
        // todo : check to see if the parameter is null - ?
        m_supportedReceiveContentTypeInfo = (ContentTypeInfo[]) supportedReceiveContentTypeInfo.clone();
        
        for (int i = 0; i < m_supportedReceiveContentTypeInfo.length; i++) {
            if (m_supportedReceiveContentTypeInfo[i] == null) {
                throw new NullPointerException(
                "supportedReceiveContentTypeInfo array contains a null value at "
                + i);
            }
        }
        
        m_preferredTransmitContentTypeInfo = preferredTransmitContentTypeInfo;
        
        // todo : check to see if the parameter is null - ?
        m_supportedTransmitContentTypeInfo = (ContentTypeInfo[]) supportedTransmitContentTypeInfo.clone();
        
        for (int i = 0; i < m_supportedTransmitContentTypeInfo.length; i++) {
            if (m_supportedTransmitContentTypeInfo[i] == null) {
                throw new NullPointerException(
                "supportedTransmitContentTypeInfo array contains a null value at "
                + i);
            }
        }
        
        m_dataStoreMemoryInfo = dsmi;
        
        m_syncCapabilities = syncCap;
        
    }
    
    /**
     *
     *  @param eDataStore element
     *
     */
    DataStore(final Element eDataStore)
    throws RepresentationException {
        // todo : verify that eDataStore is <DataStore>
        this(Util.extractSourceRef(eDataStore, true),
        Util.extractDisplayName(eDataStore, false),
        Util.extractMaxGUIDSize(eDataStore, false),
        Util.extractRxPref(eDataStore, true),
        Util.extractRxs(eDataStore, false),
        Util.extractTxPref(eDataStore, true),
        Util.extractTxs(eDataStore, false),
        Util.extractDSMem(eDataStore, false),
        Util.extractSyncCap(eDataStore, true));
    }
    
    /**
     *
     *  @return source ref. The return value is guaranteed to be non-null.
     *
     */
    public SourceRef getSourceRef() {
        return m_sourceRef;
    }
    
    /**
     *
     *  @return display name. May be null.
     *
     */
    public String getDisplayName() {
        return m_strDisplayName;
    }
    
    /**
     *
     * @return value of -1 if the Maximum GUID size is unspecified / unknown. Otherwise,
     *           this method returns a positive number.
     *
     */
    public long getMaxGuidSize() {
        return m_lMaxGuidSize;
    }
    
    /**
     *
     *  method corresponds to &ltRx-Pref&gt element
     *
     * @return content type information.
     *
     */
    public ContentTypeInfo getPreferredReceiveContentTypeInfo() {
        return m_preferredReceiveContentTypeInfo;
    }
    
    /**
     *
     *   method corresponds to &ltRx&gt element
     *
     *  @return array of Content type information. Return value is guaranteed to be non-null.
     *
     */
    public ContentTypeInfo[] getSupportedReceiveContentTypeInfo() {
        return m_supportedReceiveContentTypeInfo;
    }
    
    /**
     *
     *  method corresponds to &ltTx-Pref&gt element
     *
     * @return content type information
     *
     *
     */
    public ContentTypeInfo getPreferredTransmitContentTypeInfo() {
        return m_preferredTransmitContentTypeInfo;
    }
    
    /**
     *
     *  method corresponds to &ltTx&gt element
     *
     * @return array of content type information. Return value is guaranteed to be non-null.
     *
     */
    public ContentTypeInfo[] getSupportedTransmitContentTypeInfo() {
        return m_supportedTransmitContentTypeInfo;
    }
    
    /**
     *
     * @return datastore memory information. May return null.
     *
     */
    public DataStoreMemoryInfo getDataStoreMemoryInfo() {
        return m_dataStoreMemoryInfo;
    }
    
    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    public SyncCapabilities getSyncCapabilities() {
        return m_syncCapabilities;
    }
    
    /**
     *
     * @return an XML representation of this object
     *
     */
    String toXML() {
        StringBuffer sbXML = new StringBuffer();
        sbXML.append("<DataStore>");
        sbXML.append(this.getSourceRef().toXML());
        if ((m_strDisplayName != null) && (m_strDisplayName.trim().length() > 0)) {
            sbXML.append("<DisplayName>");
            sbXML.append(this.getDisplayName());
            sbXML.append("</DisplayName>\n");
        }
        if (m_lMaxGuidSize > 0) {
            sbXML.append("<MaxGUIDSize>");
            sbXML.append(this.getMaxGuidSize());
            sbXML.append("</MaxGUIDSize>");
        }
        sbXML.append("<Rx-Pref>");
        sbXML.append(this.getPreferredReceiveContentTypeInfo().toXML());
        sbXML.append("</Rx-Pref>");
        sbXML.append("<Tx-Pref>");
        sbXML.append(this.getPreferredTransmitContentTypeInfo().toXML());
        sbXML.append("</Tx-Pref>");
        for (int i = 0; i < m_supportedReceiveContentTypeInfo.length; i++) {
            sbXML.append("<Rx>");
            sbXML.append(m_supportedReceiveContentTypeInfo[i].toXML());
            sbXML.append("</Rx>");
        }        
        for (int i = 0; i < m_supportedTransmitContentTypeInfo.length; i++) {
            sbXML.append("<Tx>");
            sbXML.append(m_supportedTransmitContentTypeInfo[i].toXML());
            sbXML.append("</Tx>");
        }
        if (getDataStoreMemoryInfo() != null) {
            sbXML.append(this.getDataStoreMemoryInfo().toXML());
        }
        sbXML.append(this.getSyncCapabilities().toXML());
        sbXML.append("</DataStore>\n");
        return sbXML.toString();
    }
}
