/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.RepresentationException;
import sync4j.framework.core.xml.Element;

/**
 *
 *  corresponds to the &ltDSMem&gt element in the SyncML devinf DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: DataStoreMemoryInfo.java,v 1.4 2003/12/15 16:55:09 luigiafassina Exp $
 */
public final class DataStoreMemoryInfo
{
    private boolean m_bUsesSharedMemory;
    private long m_lMaximumMemorySizeInBytes;
    private long m_lMaximumDataStoreItems;

    /**
     *
     *  @param bUsesSharedMemory <b>true</b> if the datastore uses shared memory
     *  @param lMaximumMemorySizeInBytes can be a positive number, zero, or -1.
     *                                -1 indicates that the parameter is unspecified.
     *  @param lMaximumDataStoreItems can be a positive number, zero, or -1.
     *                                -1 indicates that the parameter is unspecified.
     *
     *  @throws java.lang.IllegalArgumentException
     *
     */
    public DataStoreMemoryInfo(
            final boolean bUsesSharedMemory /* <SharedMem> */,
            final long lMaximumMemorySizeInBytes /* <MaxMem> */,
            final long lMaximumDataStoreItems /* <MaxID> */)
    {
        if (lMaximumMemorySizeInBytes < -1)
        {
            throw new IllegalArgumentException(
                    "bad value for lMaximumMeorySizeInBytes parameter: "
                    + lMaximumMemorySizeInBytes);
        }
        if (lMaximumDataStoreItems < -1)
        {
            throw new IllegalArgumentException(
                    "bad value for lMaximumDataStoreItems parameter: "
                    + lMaximumDataStoreItems);
        }
        m_lMaximumMemorySizeInBytes = lMaximumMemorySizeInBytes;
        m_lMaximumDataStoreItems = lMaximumDataStoreItems;
        m_bUsesSharedMemory = bUsesSharedMemory;
    }

    /**
     *
     * @param eDSMem &ltDSMem&gt element
     *
     */
    DataStoreMemoryInfo(final Element eDSMem)
            throws RepresentationException
    {
        // todo : verify that eDSMem is <DSMem> - ?
        this(Util.extractSharedMem(eDSMem, false),
                Util.extractMaxMem(eDSMem, false),
                Util.extractMaxID(eDSMem, false));
    }

    /**
     *
     *  @return <b>true</b> if the datastore uses shared memory
     *
     */
    public boolean usesSharedMemory()
    {
        return m_bUsesSharedMemory;
    }

    /**
     *
     *  @return a return value of -1 indicates that the property value is unspecified
     *
     */
    public long getMaximumMemorySizeInBytes()
    {
        return m_lMaximumMemorySizeInBytes;
    }

    /**
     *
     *  @return a return value of -1 indicates that the property value is unspecified
     *
     */
    public long getMaximumDataStoreItems()
    {
        return m_lMaximumDataStoreItems;
    }

    /**
     *
     *  @return an XML representation of this object
     *
     */
    String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<DSMem>\n");
        if (this.usesSharedMemory())
        {
            sbXML.append("<SharedMem/>\n");
        }
        if (this.getMaximumMemorySizeInBytes() >= 0)
        {
            sbXML.append("<MaxMem>");
            sbXML.append(this.getMaximumMemorySizeInBytes());
            sbXML.append("</MaxMem>\n");
        }
        if (this.getMaximumDataStoreItems() >= 0)
        {
            sbXML.append("<MaxID>");
            sbXML.append(this.getMaximumDataStoreItems());
            sbXML.append("</MaxID>\n");
        }
        sbXML.append("</DSMem>\n");
        return sbXML.toString();
    }

}
