/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltDelete&gt element in the SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: DeleteCommand.java,v 1.6 2003/12/15 16:55:09 luigiafassina Exp $
 *
 */
public final class DeleteCommand extends ModificationCommand
{
    public static final String COMMAND_NAME = "Delete";

    private boolean m_bNoResponse;
    private boolean m_bArchiveData;
    private boolean m_bIsSoftDelete;
    private Credential m_cred;

    /**
     *
     * @param cmdid command identifier. Must be non-null.
     * @param bNoResponse <b>true</b> if no response is required.
     * @param bArchiveData <b>true</b> if the deleted data should be archived.
     * @param bIsSoftDelete <b>true</b> if this is a "soft delete". If set to <b>false</b>,
     *               then this delete command is a "hard delete". This is how the
     *               SyncML 1.0.1 specification defines the terms:
     *               <blockquote>
     *  The SyncML Delete command provides the capability for a SyncML request to delete data
     *  from the recipient's data store. Two forms of deletion are supported. Normally, when a
     *  Delete command is specified, it conveys a request to completely delete the specified data
     *  from the recipient's data store. The deleted data SHOULD no longer be associated with the
     *  originator's synchronization data. This is the semantics of a "Hard Delete". In addition,
     *  SyncML provides support for a "Soft Delete" command. The "Soft Delete" provides the
     *  capability to request the deletion of data from the recipient but NOT to remove it from the
     *  recipient's synchronization data. The rationale for a "Soft Delete" is based on the possibility
     *  of limited storage resources in a mobile device. The data is deleted to free-up storage for
     *  other, higher priority data. Conceptually, the deleted data remains in the set of
     *  synchronized data for the recipient of the "Soft Delete" command.
     *              </blockquote>
     *
     * @param cred authentication credentials. May be null.
     * @param m meta information. May be null.
     * @param items Must be non-null. Also, array elements must be non-null.
     *
     */
    public DeleteCommand(
            final CommandIdentifier cmdid,
            final boolean bNoResponse,
            final boolean bArchiveData,
            final boolean bIsSoftDelete,
            final Credential cred,
            final Meta m,
            final Item[] items)
    {
        super(cmdid, m, items);

        m_bNoResponse = bNoResponse;
        m_bArchiveData = bArchiveData;
        m_bIsSoftDelete = bIsSoftDelete;
        m_cred = cred;

        if (items.length < 1)
        {
            throw new IllegalArgumentException(
                    "Item[] array length is < 1");
        }
    }

    /**
     *
     * @param eDelete &ltDelete&gt element
     *
     */
    DeleteCommand(final Element eDelete)
            throws RepresentationException
    {
        // todo : verify that eDelete is <Delete>
        this(Util.extractCommandId(eDelete, true),
                Util.extractNoResp(eDelete, false),
                Util.extractArchive(eDelete, false),
                Util.extractSftDel(eDelete, false),
                Util.extractCred(eDelete, false),
                Util.extractMeta(eDelete, false),
                Util.extractItems(eDelete, true));

    }

    /**
     *
     *  @return <b>true</b> if no response is required.
     *
     */
    public boolean getNoResponse()
    {
        return m_bNoResponse;
    }

    /**
     *
     *
     * @return <b>true</b> if the deleted data should be archived. TODO : re-verify / clarify the meaning of
     *                  this method.
     *
     */
    public boolean getArchiveData()
    {
        return m_bArchiveData;
    }

    /**
     *
     *  @return <b>true</b> if this is a "Soft delete". Otherwise, this is a "hard delete".
     *
     */
    public boolean isSoftDelete()
    {
        return m_bIsSoftDelete;
    }

    /**
     *
     * @return authentication credentials. May be null.
     *
     */
    public Credential getCredential()
    {
        return m_cred;
    }

    /**
     *
     *  @return an XML representation of this object
     *
     */
    String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<Delete>");
        sbXML.append(this.getCommandIdentifier().toXML());
        if (this.getNoResponse())
        {
            sbXML.append("<NoResp/>\n");
        }
        if (this.getArchiveData())
        {
            sbXML.append("<Archive/>\n");
        }
        if (this.isSoftDelete())
        {
            sbXML.append("<SftDel/>\n");
        }
        if (this.getCredential() != null)
        {
            sbXML.append(this.getCredential().toXML());
        }
        if (this.getMeta() != null)
        {
            sbXML.append(this.getMeta().toXML());
        }
        for (int i = 0; i < items.length; i++)
        {
            sbXML.append(items[i].toXML());
        }
        sbXML.append("</Delete>\n");
        return sbXML.toString();
    }

    public String getName() {
        return COMMAND_NAME;
    }
}
