/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.RepresentationException;
import sync4j.framework.core.xml.Element;

/**
 *
 *  coresponds to the &ltDevInf&gt element in the SyncML devinf DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: DeviceInfo.java,v 1.4 2003/07/09 11:36:25 stefano_fornari Exp $
 */
public final class DeviceInfo
{
    private DTDVersion m_dtdversion;
    private String m_strManufacturer;
    private String m_strModelName;
    private String m_strOEM;
    private String m_strFirmwareVersion;
    private String m_strSoftwareVersion;
    private String m_strHardwareVersion;
    private String m_strDeviceIdentifier;
    private String m_strDeviceType;
    private DataStore[] m_datastores;
    private ContentTypeCapability[] m_contentTypeCapabilities;
    private Extension[] m_extensions;

    /**
     *
     * @param verdtd must be non-null.
     * @param strManufacturer device manufacturer. May be null.
     * @param strModelName device model. May be null.
     * @param strOEM device OEM. May be null.
     * @param strFirmwareVersion device firmware version. May be null.
     * @param strSoftwareVersion device software version. May be null.
     * @param strHardwareVersion device hardware version. May be null.
     * @param strDeviceIdentifier Must be non-null. Also, string length must be greater than zero.
     * @param strDeviceType Must be non-null.
     * @param datastores Must be non-null. Array length must be greater than zero. Array elements
     *                   must be non-null.
     * @param contentTypeCapabilities Must be non-null. Array length may be zero. Array elements
     *                   must be non-null.
     * @param extensions Must be non-null. Array length may be zero. Array elements
     *                   must be non-null.
     *
     * @throws java.lang.IllegalArgumentException
     *
     */
    public DeviceInfo(
            final DTDVersion verdtd /* <VerDTD> */,
            final String strManufacturer /* <Man> */,
            final String strModelName /* <Mod> */,
            final String strOEM /* <OEM> */,
            final String strFirmwareVersion /* <FwV> */,
            final String strSoftwareVersion /* <SwV> */,
            final String strHardwareVersion /* <HwV> */,
            final String strDeviceIdentifier /* <DevID> */,
            final String strDeviceType /* <DevTyp> */,
            final DataStore[] datastores,
            final ContentTypeCapability[] contentTypeCapabilities,
            final Extension[] extensions)
    {
        if (verdtd == null)
        {
            throw new NullPointerException(
                    "verdtd parameter is null");
        }

        if (strDeviceIdentifier == null)
        {
            throw new NullPointerException(
                    "strDeviceIdentifier parameter is null");
        }
        else if (strDeviceIdentifier.length() == 0)
        {
            throw new IllegalArgumentException(
                    "strDeviceIdentifier parameter length is zero");
        }

        if (strDeviceType == null)
        {
            throw new NullPointerException(
                    "strDeviceType parameter is null");
        }

        if (contentTypeCapabilities == null)
        {
            throw new NullPointerException(
                    "contentTypeCapabilities array is null");
        }

        if (extensions == null)
        {
            throw new NullPointerException(
                    "extensions array is null");
        }

        m_dtdversion = verdtd;
        m_strManufacturer = strManufacturer;
        m_strModelName = strModelName;
        m_strOEM = strOEM;
        m_strFirmwareVersion = strFirmwareVersion;
        m_strSoftwareVersion = strSoftwareVersion;
        m_strHardwareVersion = strHardwareVersion;
        m_strDeviceIdentifier = strDeviceIdentifier;
        m_strDeviceType = strDeviceType;

        m_datastores = (datastores != null) 
                     ? (DataStore[]) datastores.clone()
                     : new DataStore[0]
                     ;

        for (int i = 0; i < m_datastores.length; i++)
        {
            if (m_datastores[i] == null)
            {
                throw new NullPointerException(
                        "null value in DataStore[] at index " + i);
            }
        }


        m_contentTypeCapabilities = (ContentTypeCapability[]) contentTypeCapabilities.clone();

        for (int i = 0; i < m_contentTypeCapabilities.length; i++)
        {
            if (m_contentTypeCapabilities[i] == null)
            {
                throw new NullPointerException(
                        "null value in ContentTypeCapability[] at index " + i);
            }
        }

        m_extensions = (Extension[]) extensions.clone();

        for (int i = 0; i < m_extensions.length; i++)
        {
            if (m_extensions[i] == null)
            {
                throw new NullPointerException(
                        "null value in Extension[] at index " + i);
            }
        }
    }

    /**
     *
     *  @param strXML must be non-null
     *
     *  @throws RepresentationException
     *
     */
    public DeviceInfo(final String strXML)
            throws RepresentationException    {
        this(Util.toElement(strXML, "DevInf"));
    }


    /**
     *
     * @param eDevInf &ltDevInf&gt element. Must be non-null.
     *
     */
    DeviceInfo(final Element eDevInf)
            throws RepresentationException
    {
        // todo : verify that eDevInf is <DevInf> - ?
        this(Util.extractDTDVersion(eDevInf, true),
                Util.extractManufacturer(eDevInf, false),
                Util.extractModelName(eDevInf, false),
                Util.extractOEM(eDevInf, false),
                Util.extractFirmwareVersion(eDevInf, false),
                Util.extractSoftwareVersion(eDevInf, false),
                Util.extractHardwareVersion(eDevInf, false),
                Util.extractDevID(eDevInf, true),
                Util.extractDevTyp(eDevInf, true),
                Util.extractDataStores(eDevInf, true),
                Util.extractContentTypeCapabilities(eDevInf, false),
                Util.extractExtensions(eDevInf, false));
    }


    /**
     *
     * @return Device Info DTD version
     *
     */
    public DTDVersion getDTDVersion()
    {
        return m_dtdversion;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getManufacturer()
    {
        return m_strManufacturer;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getModelName()
    {
        return m_strModelName;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getOEM()
    {
        return m_strOEM;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getFirmwareVersion()
    {
        return m_strFirmwareVersion;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getSoftwareVersion()
    {
        return m_strSoftwareVersion;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getHardwareVersion()
    {
        return m_strHardwareVersion;
    }

    /**
     *
     * @return device identifier. Guaranteed to be non-null.
     *
     */
    public String getDeviceIdentifier()
    {
        return m_strDeviceIdentifier;
    }

    /**
     *
     * @return device type. Guaranteed to be non-null.
     *
     */
    public String getDeviceType()
    {
        return m_strDeviceType;
    }

    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    public DataStore[] getDataStores()
    {
        return (DataStore[]) m_datastores.clone();
    }

    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    public ContentTypeCapability[] getContentTypeCapabilities()
    {
        return (ContentTypeCapability[]) m_contentTypeCapabilities.clone();
    }

    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    public Extension[] getExtensions()
    {
        return (Extension[]) m_extensions.clone();
    }


    /**
     *
     * @return an XML representation of this object
     *
     */
    public String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<DevInf xmlns='" + Constants.NAMESPACE_DEVINF.getURI() + "'>\n");
        sbXML.append(this.getDTDVersion().toXML());
        if (this.getManufacturer() != null)
        {
            sbXML.append("<Man>");
            sbXML.append(this.getManufacturer());
            sbXML.append("</Man>\n");
        }
        if (this.getOEM() != null)
        {
            sbXML.append("<OEM>");
            sbXML.append(this.getOEM());
            sbXML.append("</OEM>\n");
        }
        if (this.getFirmwareVersion() != null)
        {
            sbXML.append("<FwV>");
            sbXML.append(this.getFirmwareVersion());
            sbXML.append("</FwV>\n");
        }
        if (this.getSoftwareVersion() != null)
        {
            sbXML.append("<SwV>");
            sbXML.append(this.getSoftwareVersion());
            sbXML.append("</SwV>\n");
        }
        if (this.getHardwareVersion() != null)
        {
            sbXML.append("<HwV>");
            sbXML.append(this.getHardwareVersion());
            sbXML.append("</HwV>\n");
        }
        sbXML.append("<DevID>");
        sbXML.append(this.getDeviceIdentifier());
        sbXML.append("</DevID>\n");
        sbXML.append("<DevTyp>");
        sbXML.append(this.getDeviceType());
        sbXML.append("</DevTyp>\n");
        for (int i = 0; i < m_datastores.length; i++)
        {
            sbXML.append(m_datastores[i].toXML());
        }
        for (int i = 0; i < m_contentTypeCapabilities.length; i++)
        {
            sbXML.append(m_contentTypeCapabilities[i].toXML());
        }
        for (int i = 0; i < m_extensions.length; i++)
        {
            sbXML.append(m_extensions[i].toXML());
        }
        sbXML.append("</DevInf>\n");
        return sbXML.toString();
    }

}
