/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltMap&gt tag in the SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: MapCommand.java,v 1.4 2003/02/25 10:59:07 stefano_fornari Exp $
 *
 */
public final class MapCommand extends AbstractCommand
{
    public static String COMMAND_NAME = "Map";

    private Target m_target;
    private Source m_source;
    private Credential m_cred;
    private Meta m_meta;
    private MapItem[] m_mapItems;

    /**
     *
     * @param cmdid command identifier. Must be non-null.
     * @param t must be non-null.
     * @param s must be non-null.
     * @param cred authentication credential. May be null.
     * @param m meta data. May be null.
     * @param mapItems must be non-null. Array length must be greater than zero.
     *                  Array elements must be non-null.
     *
     */
    public MapCommand(
            final CommandIdentifier cmdid,
            final Target t,
            final Source s,
            final Credential cred,
            final Meta m,
            final MapItem[] mapItems)
    {
        super(cmdid);

        m_target = t;
        m_source = s;
        m_cred = cred;
        m_meta = m;
        m_mapItems = (MapItem[]) mapItems.clone();

        if (m_target == null)
        {
            throw new IllegalArgumentException(
                    "Target is null");
        }

        if (m_source == null)
        {
            throw new IllegalArgumentException(
                    "Source is null");
        }

        if (m_mapItems.length < 1)
        {
            throw new IllegalArgumentException(
                    "MapItem array length should be greater than zero, actual length = "
                    + m_mapItems.length);
        }

        for (int i = 0; i < m_mapItems.length; i++)
        {
            if (m_mapItems[i] == null)
            {
                throw new NullPointerException(
                        "MapItem[] array contains a null value at index "
                        + i);
            }
        }

    }


    /**
     *
     *  @param eMap must be non-null.
     *
     */
    MapCommand(final Element eMap)
            throws RepresentationException
    {
        // todo : verify that eMap is <Map>
        this(Util.extractCommandId(eMap, true),
                Util.extractTarget(eMap, true),
                Util.extractSource(eMap, true),
                Util.extractCred(eMap, false),
                Util.extractMeta(eMap, false),
                Util.extractMapItems(eMap, true));
    }


    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    public Target getTarget()
    {
        return m_target;
    }

    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    public Source getSource()
    {
        return m_source;
    }

    /**
     *
     * @return The return value may be null.
     *
     */
    public Credential getCredential()
    {
        return m_cred;
    }

    /**
     *
     * @return The return value may be null.
     *
     */
    public Meta getMeta()
    {
        return m_meta;
    }

    /**
     *
     *  @return The return value is guaranteed to be non-null.
     *
     */
    public MapItem[] getMapItems()
    {
        return (MapItem[]) m_mapItems.clone();
    }

    /**
     *
     * @return an XML representation of this object
     *
     */
    public String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<Map>\n");
        sbXML.append(this.getCommandIdentifier().toXML());
        sbXML.append(this.getTarget().toXML());
        sbXML.append(this.getSource().toXML());
        if (this.getCredential() != null)
        {
            sbXML.append(this.getCredential().toXML());
        }
        if (this.getMeta() != null)
        {
            sbXML.append(this.getMeta().toXML());
        }
        for (int i = 0; i < m_mapItems.length; i++)
        {
            sbXML.append(m_mapItems[i].toXML());
        }
        sbXML.append("</Map>\n");
        return sbXML.toString();
    }

    public String getName() {
        return MapCommand.COMMAND_NAME;
    }
}
