/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.*;
import sync4j.framework.core.xml.*;

/**
 * 
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: MetInfUtil.java,v 1.5 2003/12/15 16:55:09 luigiafassina Exp $
 */
public class MetInfUtil {
    
    // --------------------------------------------------------------- Constants
    
    public static final String METINF_FORMAT       = "Format"    ;
    public static final String METINF_TYPE         = "Type"      ;
    public static final String METINF_MARK         = "Mark"      ;
    public static final String METINF_SIZE         = "Size"      ;
    public static final String METINF_ANCHOR       = "Anchor"    ;
    public static final String METINF_VERSION      = "Version"   ;
    public static final String METINF_NEXT_NONCE   = "NextNonce" ;
    public static final String METINF_MAX_MSG_SIZE = "MaxMsgSize";
    public static final String METINF_MAX_OBJ_SIZE = "MaxObjSize";
    public static final String METINF_EMI          = "EMI"       ;
    public static final String METINF_MEM          = "Mem"       ;
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Extract and return the proper MetaContent based on the specific element
     * inside the &lt;Meta&gt;. The following mapping is followed:
     *
     * <ul>
     *   <li>&lt;Format&gt; =&ght; StringMetaContent
     *   <li>&lt;Type&gt; =&ght; StringMetaContent
     *   <li>&lt;Mark&gt; =&ght; StringMetaContent
     *   <li>&lt;Size&gt; =&ght; IntegerMetaContent
     *   <li>&lt;Anchor&gt; =&ght; AnchorMetaContent
     *   <li>&lt;Version&gt; =&ght; StringMetaContent
     *   <li>&lt;NextNonce&gt; =&ght; StringMetaContent
     *   <li>&lt;MaxMsgSize&gt; =&ght; IntegerMetaContent
     *   <li>&lt;MaxObjSize&gt; =&ght; IntegerMetaContent
     *   <li>&lt;EMI&gt; =&ght; EMIMetaContent
     *   <li>&lt;Mem&gt; =&ght; MemMetaContent
     * </ul>
     *
     * @param element the <i>Element</i> representing the &lt;Meta&gt; element
     * 
     * @return the corresponding <i>MetaContent</i>
     *
     * @throws RepresentationException
     */
    public static MetaContent extractMetInf(final Element element) 
    throws RepresentationException {
        String  name = element.getName();
        if (  
             METINF_FORMAT.equals    (name) || 
             METINF_TYPE.equals      (name) ||
             METINF_MARK.equals      (name) ||
             METINF_VERSION.equals   (name) ||
             METINF_NEXT_NONCE.equals(name)
        ) {
            return new StringMetaContent(element.getText(), name);
        } else if (
                    METINF_SIZE.equals        (name) || 
                    METINF_MAX_MSG_SIZE.equals(name) || 
                    METINF_MAX_OBJ_SIZE.equals(name)
                  ) {

            Integer i = null;
            try {
                i = Integer.valueOf(element.getText());
            } catch (NumberFormatException e) {
                throw new RepresentationException("Expected an integer, found: " + element.getText());
            }

            return new IntegerMetaContent(i, name);
        } else if (METINF_ANCHOR.equals(name)) {
            return new AnchorMetaContent(element);
        } else if (METINF_MEM.equals(name)) {
            return new MemMetaContent(element);
        } else if (METINF_EMI.equals(name)) {
            throw new RepresentationException("<Meta> child <" + name + "> not implemented yet");
        }
        
        //
        // You should not be here!!!
        //

        throw new RepresentationException("Unknown <Meta> child: <" + name + ">");
    }

}
