/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.xml.Element;
import sync4j.framework.core.RepresentationException;

/**
 *
 *  corresponds to &ltMetInf&gt tag in SyncML metinfo DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *  @version $Id: MetaInfo.java,v 1.3 2003/02/25 10:59:07 stefano_fornari Exp $
 *
 */
public final class MetaInfo
{
    private String m_strEncodingFormat;
    private String m_strType;
    private String m_strMark;
    private SyncAnchor m_anchor;
    private String m_strDataVersion;
    private NextNonce m_nextNonce;
    private long m_lMaxMsgSize;
    private long m_lDataSizeInBytes;
    private ExperimentalMetaInfo[] m_aEMI;
    private MemoryInfo m_memInfo;

    /**
     *
     * @param strEncodingFormat encoding format. May be null
     * @param strType usually a MIME type. May be null
     * @param strMark mark. May be null  @see Mark
     * @param lDataSizeInBytes  data size in bytes. Must be &gt= 0
     * @param anchor  Anchor.  May be null
     * @param strDataVersion data version. May be null
     * @param nonce Next nonce value.  May be null
     * @param lMaxMsgSize maximum message size in bytes. Must be &gt= 1
     * @param aEMI experimental meta info.  Must be non-null. Array length must be &gt= 0. Array
     *              elements must be non-null.
     * @param memInfo  memory information. May be null.
     *
     * @throws java.lang.IllegalArgumentException
     *
     */
    public MetaInfo(
            final String strEncodingFormat /* <Format> element */,
            final String strType /* <Type> element */,
            final String strMark /* <Mark> element */,
            final long lDataSizeInBytes /* <Size> */,
            final SyncAnchor anchor /* <Anchor> element */,
            final String strDataVersion /* <Version> element */,
            final NextNonce nonce /* <NextNonce> element */,
            final long lMaxMsgSize /* <MaxMsgSize> */,
            final ExperimentalMetaInfo[] aEMI /* <EMI> element */,
            final MemoryInfo memInfo /* <Mem> element */)
    {

        // todo : do I need any additional parameter validation code?

        m_strEncodingFormat = strEncodingFormat;
        m_strType = strType;
        m_strMark = strMark;
        m_anchor = anchor;

        m_lDataSizeInBytes = lDataSizeInBytes;

        if ((lDataSizeInBytes < 0)
                && (lDataSizeInBytes != -1))
        {
            throw new IllegalArgumentException(
                    "lDataSizeInBytes parameter is < 0");
        }

        m_strDataVersion = strDataVersion;
        m_nextNonce = nonce;

        m_lMaxMsgSize = lMaxMsgSize;

        if ((lMaxMsgSize < 1)
                && (lMaxMsgSize != -1))
        {
            throw new IllegalArgumentException(
                    "lMaxMsgSize parameter is < 1");
        }

        if (aEMI == null)
        {
            throw new NullPointerException("ExperimentalMetaInfo[] parameter is null");
        }

        m_aEMI = (ExperimentalMetaInfo[]) aEMI.clone();
        for (int i = 0; i < m_aEMI.length; i++)
        {
            if (m_aEMI[i] == null)
            {
                throw new NullPointerException(
                        "ExperimentalMetaInfo[] contains a null "
                        + "element at index " + i);
            }
        }

        m_memInfo = memInfo;

    }

    /**
     *
     * @param strXML must be non-null
     *
     */
    public MetaInfo(final String strXML)
            throws RepresentationException
    {
        this(Util.toElement(strXML, "MetInf"));
    }

    /**
     *
     * @param eMetInf must be non-null
     *
     */
    MetaInfo(final Element eMetInf)
            throws RepresentationException
    {
        // todo : verify that eMetInf is <MetInf> - ?
        this(Util.extractFormat(eMetInf, false),
                Util.extractType(eMetInf, false),
                Util.extractMark(eMetInf, false),
                Util.extractSize(eMetInf, false),
                Util.extractAnchor(eMetInf, false),
                Util.extractVersion(eMetInf, false),
                Util.extractNextNonce(eMetInf, false),
                Util.extractMaxMsgSize(eMetInf, false),
                Util.extractEMIs(eMetInf, false),
                Util.extractMem(eMetInf, false));
    }

    /**
     *
     *   The return value corresponds to the &ltSize&gt
     *   element in the MetInf DTD.
     *
     *   @return -1 if the size was not specified
     *
     *
     */
    public long getDataSizeInBytes()
    {
        return m_lDataSizeInBytes;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getEncodingFormat()
    {
        return m_strEncodingFormat;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getType()
    {
        return m_strType;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getMark()
    {
        return m_strMark;
    }

    /**
     *
     *  @return may return null
     *
     */
    public SyncAnchor getSyncAnchor()
    {
        return m_anchor;
    }

    /**
     *
     *  @return may return null
     *
     */
    public String getDataVersion()
    {
        return m_strDataVersion;
    }

    /**
     *
     *  @return may return null
     *
     */
    public NextNonce getNextNonce()
    {
        return m_nextNonce;
    }

    /**
     *
     *   The return value corresponds to the &ltMaxMsgSize&gt
     *   element in the MetInf DTD.
     *
     *   @return -1 if the size was not specified
     *
     *
     */
    public long getMaxMessageSize()
    {
        return m_lMaxMsgSize;
    }

    /**
     *
     * @return an array of ExperimentalMetaInfo objects.
     *      The return value is guaranteed to be non-null.
     *      The array length may be zero.
     *      The array elements are guaranteed to be non-null.
     *
     */
    public ExperimentalMetaInfo[] getExperimentalMetaInfo()
    {
        return (ExperimentalMetaInfo[]) m_aEMI.clone();
    }

    /**
     *
     *  The return value of this method corresponds to
     *  the &ltMem&gt element in the MetInf DTD.
     *
     *  @return may return null
     *
     */
    public MemoryInfo getMemoryInfo()
    {
        return m_memInfo;
    }

    /**
     *
     *
     * @return an XML representation of this object
     *
     */
    // todo : do we really want to declare this "public" - ?
    public String toXML()
    {
        StringBuffer sbXML = new StringBuffer();
        sbXML.append("<MetInf>\n");
        if (this.getEncodingFormat() != null)
        {
            sbXML.append("<Format>");
            sbXML.append(this.getEncodingFormat());
            sbXML.append("</Format>\n");
        }
        if (this.getType() != null)
        {
            sbXML.append("<Type>");
            sbXML.append(this.getType());
            sbXML.append("</Type>\n");
        }
        if (this.getMark() != null)
        {
            sbXML.append("<Mark>");
            sbXML.append(this.getMark());
            sbXML.append("</Mark>\n");
        }
        if (this.getDataSizeInBytes() != -1)
        {
            sbXML.append("<Size>");
            sbXML.append(this.getDataSizeInBytes());
            sbXML.append("</Size>\n");
        }
        if (this.getSyncAnchor() != null)
        {
            sbXML.append(this.getSyncAnchor().toXML());
        }
        if (this.getDataVersion() != null)
        {
            sbXML.append("<Version>");
            sbXML.append(this.getDataVersion());
            sbXML.append("</Version>\n");
        }
        if (this.getNextNonce() != null)
        {
            sbXML.append(this.getNextNonce().toXML());
        }
        if (this.getMaxMessageSize() != -1)
        {
            sbXML.append("<MaxMsgSize>");
            sbXML.append(this.getMaxMessageSize());
            sbXML.append("</MaxMsgSize>\n");
        }
        for (int i = 0; i < m_aEMI.length; i++)
        {
            sbXML.append(m_aEMI[i].toXML());
        }
        if (this.getMemoryInfo() != null)
        {
            sbXML.append(this.getMemoryInfo().toXML());
        }
        sbXML.append("</MetInf>\n");
        return sbXML.toString();
    }

}
