/*
 Copyright (c) 2001-2002-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltSearch&gt tag in the SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: SearchCommand.java,v 1.5 2003/12/15 16:55:09 luigiafassina Exp $
 *
 */
public final class SearchCommand extends AbstractCommand {
    public static String COMMAND_NAME = "Search";
    
    private boolean m_bNoResponse;
    private boolean m_bNoResults;
    private Credential m_cred;
    private Target m_target;
    private Source[] m_sources;
    private String m_strLanguage;
    private Meta m_meta;
    private Data m_data;
    
    /**
     *
     *  @param cmdid command identifier. Must be non-null
     *  @param bNoResponse <b>true</b> if no response is required
     *  @param bNoResults <b>true</b> if no results are required
     *  @param cred  authentication credentials. May be null.
     *  @param t  may be null
     *  @param sources Must be non-null. Array must contain 1 or more Source's. The
     *              array elements must be non-null.
     *  @param strLanguage language. May be null
     *  @param m  meta data. Must be non-null
     *  @param d  must contain the search grammar. Must be non-null
     *
     *  @throws java.lang.IllegalArgumentException
     *
     */
    public SearchCommand(
    final CommandIdentifier cmdid,
    final boolean bNoResponse,
    final boolean bNoResults,
    final Credential cred,
    final Target t,
    final Source[] sources,
    final String strLanguage,
    final Meta m,
    final Data d /* todo : perhaps change this parameter to "SearchGrammar" or "ISearchGrammar" */ ) {
        super(cmdid);
        
        m_bNoResponse = bNoResponse;
        m_bNoResults = bNoResults;
        m_cred = cred;
        m_target = t;
        
        if (sources == null) {
            throw new NullPointerException(
            "Source[] parameter is null");
        }
        
        m_sources = (Source[]) sources.clone();
        m_strLanguage = strLanguage;
        m_meta = m;
        m_data = d;
        
        if (m_sources.length < 1) {
            throw new IllegalArgumentException(
            "Source[] array length is < 1");
        }
        
        for (int i = 0; i < m_sources.length; i++) {
            if (m_sources[i] == null) {
                throw new NullPointerException(
                "Source[] array contains a null value at " + i);
            }
        }
        
        if (m_meta == null) {
            throw new NullPointerException(
            "Meta parameter is null");
        }
        
        if (m_data == null) {
            throw new NullPointerException(
            "Data parameter is null");
        }
    }
    
    /**
     *
     * @param eSearch must be non-null.
     *
     */
    SearchCommand(final Element eSearch)
    throws RepresentationException {
        // todo : verify that eSearch is <Search> - ?
        this(Util.extractCommandId(eSearch, true),
        Util.extractNoResp(eSearch, false),
        Util.extractNoResults(eSearch, false),
        Util.extractCred(eSearch, false),
        Util.extractTarget(eSearch, false),
        Util.extractSources(eSearch, true),
        Util.extractLang(eSearch, false),
        Util.extractMeta(eSearch, true),
        Util.extractData(eSearch, true));
    }
    
    /**
     *
     * @return <b>true</b> if no response is required
     *
     */
    public boolean getNoResponse() {
        return m_bNoResponse;
    }
    
    
    /**
     *
     * @return <b>true</b> if no results are required
     *
     */
    public boolean getNoResults() {
        return m_bNoResults;
    }
    
    /**
     *
     *  @return may return null
     *
     *
     */
    public Credential getCredential() {
        return m_cred;
    }
    
    /**
     *
     *  @return may return null
     *
     *
     */
    public Target getTarget() {
        return m_target;
    }
    
    /**
     *
     * @return The return value is guaranteed to be non-null.
     *      The array elements are guaranteed to be non-null.
     *
     *
     */
    public Source[] getSources() {
        return (Source[]) m_sources.clone();
    }
    
    /**
     *
     *  @return may return null
     *
     */
    public String getLanguage() {
        return m_strLanguage;
    }
    
    /**
     *
     * @return may return null
     *
     */
    public Meta getMeta() {
        return m_meta;
    }
    
    /**
     *
     * @return The return value is guaranteed to be non-null.
     *
     */
    // todo : perhaps change this method to "getSearchGrammar"
    public Data getData() {
        return m_data;
    }
    
    /**
     *
     * @return an XML representation of this object
     *
     */
    String toXML() {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<Search>");
        sbXML.append(this.getCommandIdentifier().toXML());
        if (this.getNoResponse()) {
            sbXML.append("<NoResp/>\n");
        }
        if (this.getNoResults()) {
            sbXML.append("<NoResults/>\n");
        }
        if (this.getCredential() != null) {
            sbXML.append(this.getCredential().toXML());
        }
        if (this.getTarget() != null) {
            sbXML.append(this.getTarget().toXML());
        }
        for (int i = 0; i < m_sources.length; i++) {
            sbXML.append(m_sources[i].toXML());
        }
        if (this.getLanguage() != null) {
            sbXML.append("<Lang>");
            sbXML.append(this.getLanguage());
            sbXML.append("</Lang>\n");
        }
        sbXML.append(this.getMeta().toXML());
        sbXML.append(this.getData().toXML());
        sbXML.append("</Search>\n");
        return sbXML.toString();
    }
    public String getName() {
        return SearchCommand.COMMAND_NAME;
    }
    
    
    
}
