/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import java.util.*;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltSyncBody&gt element in the
 *  SyncML represent DTD
 *
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *  @see    SyncHeader
 *
 *
 *  @version $Id: SyncBody.java,v 1.6 2003/12/15 16:55:09 luigiafassina Exp $
 */
public final class SyncBody
{
    private static final String ALLOWED_COMMANDS = ":Alert:Add:Atomic:Copy:Delete:Exec:Get:Map:Put:Replace:Results:Search:Sequence:Status:Sync:";
    
    private AbstractCommand[] m_commands;
    private boolean m_bFinal;

    /**
     *
     *  @param commands This parameter must be non-null. Also, the array
     *                  length must be greater than zero. The array elements
     *                  must be non-null. The array elements must be an instance
     *                  of one of these classes: {@link AlertCommand},  {@link AtomicCommand},
     *                  {@link CopyCommand}, {@link ExecCommand}, {@link GetCommand}, {@link MapCommand},
     *                  {@link PutCommand}, {@link ResultsCommand}, {@link SearchCommand}, {@link SequenceCommand},
     *                  {@link StatusCommand}, {@link SyncCommand}
     *  @param bFinal if this parameter is true, it means that this is the final message that is being sent
     *
     *  @throws java.lang.IllegalArgumentException
     *
     */
    public SyncBody(
            final AbstractCommand[] commands,
            final boolean bFinal)
    {
        if (commands == null)
        {
            throw new NullPointerException(
                    "commands array parameter is null");
        }

        m_commands = (AbstractCommand[]) commands.clone();

        if (m_commands.length < 1)
        {
            throw new IllegalArgumentException(
                    "AbstractCommand[] array length is less than 1");
        }
        for (int i = 0; i < m_commands.length; i++)
        {
            AbstractCommand c = m_commands[i];
            if (c == null)
            {
                throw new NullPointerException(
                        "AbstractCommand array contains a null value, index = "
                        + i);
            }
            
            if (ALLOWED_COMMANDS.indexOf(c.getName()) < 0)
            {
                throw new IllegalArgumentException(
                        "command array contains the wrong command: "
                        + c.getName());
            }
        }
        m_bFinal = bFinal;
    }

    /**
     *
     * @param syncBodyElement must be non-null
     *
     */
    SyncBody(final Element syncBodyElement)
            throws RepresentationException
    {
        // todo : add code to verify that syncBodyElement is <SyncBody>
        this(Util.extractCommands(syncBodyElement, true),
                Util.extractFinal(syncBodyElement, false));
    }

    /**
     *
     *  @return the return value is guaranteed to be non-null. Also,
     *          the elements of the array are guaranteed to be non-null.
     *
     */
    public AbstractCommand[] getCommands()
    {
        return (AbstractCommand[]) m_commands.clone();
    }
    
    /**
     * Mark this body as final
     */
    public void setFinal() {
        m_bFinal = true;
    }

    /**
     *
     *  @return The return value is true if this is the final message being sent.
     *
     */
    public boolean isFinal()
    {
        return m_bFinal;
    }

    /**
     *
     *  @return an XML representation of this object
     *
     */
    String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<SyncBody>\n");
        for (int i = 0; i < m_commands.length; i++)
        {
            sbXML.append(m_commands[i].toXML());
            sbXML.append("\n");
        }
        if (this.isFinal())
        {
            sbXML.append("<Final/>\n");
        }
        sbXML.append("</SyncBody>\n");
        return sbXML.toString();
    }

}
