/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltSync&gt command in the SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: SyncCommand.java,v 1.3 2003/02/25 10:59:08 stefano_fornari Exp $
 *
 */
public final class SyncCommand extends AbstractCommand {
    public static String COMMAND_NAME = "Sync";
    
    private boolean m_bNoResponse;
    private Credential m_cred;
    private Target m_target;
    private Source m_source;
    private Meta m_meta;
    private AbstractCommand[] m_commands;
    
    /**
     *
     * @param cmdid must be non-null
     * @param bNoResponse <b>true</b> if no response is required
     * @param cred authentication credential. May be null.
     * @param t may be null
     * @param s may be null
     * @param m may be null
     * @param commands Must be non-null. Array length may be zero.
     *          The array's elements must be non-null.
     *          The array's elements must be of one of
     *          the following types:
     *          {@link AddCommand},
     *          {@link AtomicCommand},
     *          {@link CopyCommand},
     *          {@link DeleteCommand},
     *          {@link ReplaceCommand},
     *          {@link SequenceCommand}
     *
     * @throws java.lang.IllegalArgumentException
     *
     */
    public SyncCommand(
    final CommandIdentifier cmdid,
    final boolean bNoResponse,
    final Credential cred /* optional param */,
    final Target t /* optional param */,
    final Source s /* optional param */,
    final Meta m /* optional param */,
    final AbstractCommand[] commands /* zero or more elements */) {
        super(cmdid);
        
        if (commands == null) {
            throw new NullPointerException(
            "commands array is null");
        }
        
        m_bNoResponse = bNoResponse;
        m_cred = cred;
        m_target = t;
        m_source = s;
        m_meta = m;
        m_commands = (AbstractCommand[]) commands.clone();
        
        // validate the contents of the m_commands array
        for (int i = 0; i < m_commands.length; i++) {
            if (m_commands[i] == null) {
                throw new NullPointerException(
                "null element in AbstractCommand array, index = "
                + i);
            }
            else if ((!(m_commands[i] instanceof AddCommand))
            && (!(m_commands[i] instanceof AtomicCommand))
            && (!(m_commands[i] instanceof CopyCommand))
            && (!(m_commands[i] instanceof DeleteCommand))
            && (!(m_commands[i] instanceof ReplaceCommand))
            && (!(m_commands[i] instanceof SequenceCommand))) {
                throw new IllegalArgumentException(
                "illegal command in array: "
                + m_commands[i].getClass().toString());
            }
        }
    }
    
    /**
     *
     * @param eSync must be non-null
     *
     */
    SyncCommand(final Element eSync)
    throws RepresentationException {
        // todo : verify that eSync is <Sync> - ?
        this(Util.extractCommandId(eSync, true),
        Util.extractNoResp(eSync, false),
        Util.extractCred(eSync, false),
        Util.extractTarget(eSync, false),
        Util.extractSource(eSync, false),
        Util.extractMeta(eSync, false),
        Util.extractAbstractCommands(eSync, false));
    }
    
    /**
     *
     * @return <b>true</b> if no response is required
     *
     */
    public boolean getNoResponse() {
        return m_bNoResponse;
    }
    
    /**
     *
     * @return authentication credential object or null
     *
     */
    public Credential getCredential() {
        return m_cred;
    }
    
    /**
     *
     * @return may return null
     *
     */
    public Target getTarget() {
        return m_target;
    }
    
    /**
     *
     * @return may return null
     *
     */
    public Source getSource() {
        return m_source;
    }
    
    /**
     *
     * @return may return null
     *
     */
    public Meta getMeta() {
        return m_meta;
    }
    
    /**
     *
     * @return The return value is guaranteed to be non-null.
     *          The array's elements are guaranteed to be non-null.
     *
     */
    public AbstractCommand[] getCommands() {
        return (AbstractCommand[]) m_commands.clone();
    }
    
    String toXML() {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<Sync>");
        sbXML.append(this.getCommandIdentifier().toXML());
        if (this.getNoResponse() == true) {
            sbXML.append("<NoResp/>\n");
        }
        if (this.getCredential() != null) {
            sbXML.append(this.getCredential().toXML());
        }
        if (this.getTarget() != null) {
            sbXML.append(this.getTarget().toXML());
        }
        if (this.getSource() != null) {
            sbXML.append(this.getSource().toXML());
        }
        if (this.getMeta() != null) {
            sbXML.append(this.getMeta().toXML());
        }
        for (int i = 0; i < m_commands.length; i++) {
            sbXML.append(m_commands[i].toXML());
        }
        sbXML.append("</Sync>\n");
        return sbXML.toString();
    }
    
    public String getName() {
        return SyncCommand.COMMAND_NAME;
    }
    
}
