/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.core;

import sync4j.framework.core.xml.*;

/**
 *
 *  corresponds to the &ltSyncHdr&gt element in the
 *  SyncML represent DTD
 *
 *  @author Sean C. Sullivan
 *  @author Stefano Fornari @ Funambol
 *
 *  @see    SyncBody
 *
 *  @version $Id: SyncHeader.java,v 1.6 2003/12/15 16:55:09 luigiafassina Exp $
 */
public final class SyncHeader
{
    private DTDVersion m_verdtd;
    private ProtocolVersion m_verproto;
    private SyncSessionIdentifier m_ssid;
    private String m_strMsgID;
    private Target m_target;
    private Source m_src;
    private String m_responseURI;
    private boolean m_bNoResponse;
    private Credential m_cred;
    private Meta m_meta;

    /**
     *
     *  @param verdtd SyncML DTD version
     *  @param verproto SyncML protocol version
     *  @param ssid sync session identifier
     *  @param strMsgID message ID
     *  @param target target URI
     *  @param src source URI
     *  @param responseURI may be null.
     *  @param bNoResponse true if no response is required
     *  @param cred credentials. May be null.
     *  @param meta may be null.
     *
     */
    public SyncHeader(
            final DTDVersion verdtd,
            final ProtocolVersion verproto,
            final SyncSessionIdentifier ssid,
            final String strMsgID,
            final Target target,
            final Source src,
            final String responseURI /* may be null */,
            final boolean bNoResponse,
            final Credential cred /* may be null */,
            final Meta meta /* may be null */)
    {
        if (verdtd == null)
        {
            throw new NullPointerException(
                    "verdtd parameter is null");
        }

        if (verproto == null)
        {
            throw new NullPointerException(
                    "verproto parameter is null");
        }

        if (ssid == null)
        {
            throw new NullPointerException(
                    "ssid parameter is null");
        }

        if (strMsgID == null)
        {
            throw new NullPointerException(
                    "strMsgID parameter is null");
        }
        else if (strMsgID.length() == 0)
        {
            throw new IllegalArgumentException(
                    "strMsgID parameter:  string length is zero");
        }

        if (target == null)
        {
            throw new NullPointerException(
                    "target parameter is null");
        }

        if (src == null)
        {
            throw new NullPointerException(
                    "src parameter is null");
        }

        m_verdtd = verdtd;
        m_verproto = verproto;
        m_ssid = ssid;
        m_strMsgID = strMsgID;
        m_target = target;
        m_src = src;
        m_responseURI = responseURI;
        m_bNoResponse = bNoResponse;
        m_cred = cred;
        m_meta = meta;
    }

    /**
     *
     *  @param syncHeaderElement element &ltSyncHdr&gt. Parameter must be non-null.
     *
     */
    SyncHeader(final Element syncHeaderElement)
            throws RepresentationException
    {
        // todo : verify that syncHeaderElement is <SyncHdr>
        this(Util.extractDTDVersion(syncHeaderElement, true),
                Util.extractVerProto(syncHeaderElement, true),
                Util.extractSessionID(syncHeaderElement, true),
                Util.extractMessageID(syncHeaderElement, true),
                Util.extractTarget(syncHeaderElement, true),
                Util.extractSource(syncHeaderElement, true),
                Util.extractRespURI(syncHeaderElement, false),
                Util.extractNoResp(syncHeaderElement, false),
                Util.extractCred(syncHeaderElement, false),
                Util.extractMeta(syncHeaderElement, false));
    }

    /**
     *
     *  @return DTD version
     *
     */
    public DTDVersion getDTDVersion()
    {
        return m_verdtd;
    }

    /**
     *
     *  @return protocol version
     *
     */
    public ProtocolVersion getProtocolVersion()
    {
        return m_verproto;
    }

    /**
     *
     *  @return sync session identifier
     *
     */
    public SyncSessionIdentifier getSyncSessionIdentifier()
    {
        return m_ssid;
    }

    /**
     *
     *  @return message ID
     *
     */
    public String getMessageID()
    {
        return m_strMsgID;
    }

    /**
     *
     *  @return target
     *
     */
    public Target getTarget()
    {
        return m_target;
    }

    /**
     *
     *  @return source
     *
     */
    public Source getSource()
    {
        return m_src;
    }

    /**
     *
     *
     *  @return the response URI; may return null
     *
     *
     */
    public String getResponseURI()
    {
        return m_responseURI;
    }
    
    /**
     * Sets the response URI.
     *
     * @param uri the new response URI; NOT NULL
     */
    public void setResponseURI(String uri) {
        m_responseURI = uri;
    }

    /**
     *
     *  @return true if no response is required.
     *
     */
    public boolean getNoResponse()
    {
        return m_bNoResponse;
    }

    /**
     *
     *
     *  @return may return null
     *
     *
     */
    public Credential getCredential()
    {
        return m_cred;
    }

    /**
     *
     *
     *  @return may return null
     *
     *
     */
    public Meta getMeta()
    {
        return m_meta;
    }


    /**
     *
     *  @return a String that contains an XML representation of this object
     *
     */
    String toXML()
    {
        StringBuffer sbXML;
        sbXML = new StringBuffer();
        sbXML.append("<SyncHdr>\n");
        sbXML.append(m_verdtd.toXML());
        sbXML.append(m_verproto.toXML());
        sbXML.append("<SessionID>");
        sbXML.append(m_ssid.toString());
        sbXML.append("</SessionID>\n");
        sbXML.append("<MsgID>");
        sbXML.append(m_strMsgID);
        sbXML.append("</MsgID>\n");
        sbXML.append(m_target.toXML());
        sbXML.append(m_src.toXML());
        if (m_responseURI != null)
        {
            sbXML.append("<RespURI>");
            sbXML.append(m_responseURI);
            sbXML.append("</RespURI>\n");
        }
        if (this.getNoResponse())
        {
            sbXML.append("<NoResp/>\n");
        }
        if (this.getCredential() != null)
        {
            sbXML.append(this.getCredential().toXML());
        }
        if (this.getMeta() != null)
        {
            sbXML.append(this.getMeta().toXML());
        }
        sbXML.append("</SyncHdr>\n");
        return sbXML.toString();
    }

}
