/*
 *  Copyright (c) 2001-2002 sync4j project
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions, and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions, and the disclaimer that follows
 *  these conditions in the documentation and/or other materials
 *  provided with the distribution.
 *  3. The name "sync4j" must not be used to endorse or promote products
 *  derived from this software without prior written permission.
 *  4. Products derived from this software may not be called "sync4j", nor
 *  may "sync4j" appear in their name, without prior written permission.
 *  In addition, we request (but do not require) that you include in the
 *  end-user documentation provided with the redistribution and/or in the
 *  software itself an acknowledgement equivalent to the following:
 *  "This product includes software developed by the
 *  sync4j project."
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package sync4j.framework.core.xml.nanoxml;

import nanoxml.kXMLParseException;
import nanoxml.kXMLElement;

import sync4j.framework.core.xml.IXmlOperation;
import sync4j.framework.core.xml.ParseException;

import java.util.*;

/**
 * This class adapts to the NanoXML DOM API.
 *
 * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @version   $Id: XmlOperationImpl.java,v 1.11 2003/12/27 21:25:36 stefano_fornari Exp $
 * @since     SyncML 1.1
 * @see       nanoxml.kXMLElement
 */
public final class XmlOperationImpl implements IXmlOperation {
    /**
     * Constructor for the XmlOperationImpl object
     */
    public XmlOperationImpl() { }


    /**
     * Description of the Method
     *
     * @param xmlFragment                 Description of the Parameter
     * @param strElementName              Description of the Parameter
     * @return                            Description of the Return Value
     * @exception ParseException  Description of the Exception
     */
    public sync4j.framework.core.xml.Element toElement(
                                             String xmlFragment, String strElementName)
         throws ParseException {
        if (xmlFragment == null) {
            throw new NullPointerException("XML string is null");
        }
        // bugfix dbu, 18.3.2003: per default kXML is case insensitive.
        // this leads to attributes being converted to upper case if parsed xml
        // is transformed to string again.
        // the internal default constructor makes the same call, except for the third parameter (IgnoreCase)
        kXMLElement e = new kXMLElement(new Hashtable(), false, false);
        try {
            e.parseString(xmlFragment);
        } catch (kXMLParseException ke) {
            throw new ParseException(ke.toString());
        }

        if (e == null) {
            throw new ParseException("root element is null");
        } else if (strElementName != null) {
            String tagName = e.getTagName();
            if (tagName == null || !tagName.equals(strElementName)) {
                throw new ParseException(
                    "expected <"
                     + strElementName
                     + "> but found <"
                     + tagName
                     + ">");
            }
        }

        return new ElementImpl(e);
    }


    /**
     * <p>
     *
     * Wraps the <code>kXMLElement</code> object into the adapter type, <code>ElementImpl</code>
     * . The reason is that the iteration/ enumeration always returns a list
     * of the DOM-specific objects, and we meant to hide the DOM-specific
     * class. </p> <p>
     *
     * This method is analogous to the type casting, <code>(kXMLElement)i.next()</code>
     * .
     *
     * @param obj  the object to be casted.
     * @return     an instance of the wrapper, <code>ElementImpl</code>
     */
    public sync4j.framework.core.xml.Element castToElement(Object obj) {
        return new ElementImpl((kXMLElement) obj);
    }


    /**
     * Wraps for its counterpart with the parent's namespace.
     *
     * @param eParent                     Description of the Parameter
     * @param strChildName                Description of the Parameter
     * @param bElementIsRequired          Description of the Parameter
     * @return                            Description of the Return Value
     * @exception ParseException  Description of the Exception
     */
    public sync4j.framework.core.xml.Element extractElement(
        final Object eParent,
              final String strChildName,
              final boolean bElementIsRequired) throws ParseException {
        return extractElement(
            eParent,
            strChildName,
            bElementIsRequired,
            new NamespaceImpl(""));
    }


    /**
     * @param ns                          Expects an instance of
     *      NamespaceImpl. Ignored.
     * @param eParent                     Description of the Parameter
     * @param strChildName                Description of the Parameter
     * @param bElementIsRequired          Description of the Parameter
     * @return                            Description of the Return Value
     * @exception ParseException  Description of the Exception
     */
    public sync4j.framework.core.xml.Element extractElement(
        final Object eParent,
              final String strChildName,
              final boolean bElementIsRequired,
              final Object ns)
         throws ParseException {
        ElementImpl parent = (ElementImpl) eParent;
        ElementImpl e = parent.getChild(
            strChildName,
            "");

        if (e == null) {
            if (bElementIsRequired) {
                throw new ParseException(
                    "parent "
                     + parent.getElement()
                     + " is missing child: <"
                     + strChildName
                     + ">");
            } else {
                return null;
            }
        }

        return e;
    }


    /**
     * Wraps its counterpart with the parent's namespace.
     *
     * @param eParent                     Description of the Parameter
     * @param strChildName                Description of the Parameter
     * @param bElementIsRequired          Description of the Parameter
     * @return                            Description of the Return Value
     * @exception ParseException  Description of the Exception
     */
    public String extractElementText(
        final Object eParent,
              final String strChildName,
              final boolean bElementIsRequired)
         throws ParseException {
        return extractElementText(eParent,
            strChildName,
            bElementIsRequired,
            "");
    }


    /**
     * @param ns                          Expects an instance of NamespaceImpl
     * @param eParent                     Description of the Parameter
     * @param strChildName                Description of the Parameter
     * @param bElementIsRequired          Description of the Parameter
     * @return                            Description of the Return Value
     * @exception ParseException  Description of the Exception
     */
    public String extractElementText(
        final Object eParent,
              final String strChildName,
              final boolean bElementIsRequired,
              final Object ns)
         throws ParseException {
        sync4j.framework.core.xml.Element e = extractElement(
            eParent,
            strChildName,
            bElementIsRequired,
            ns);
        if (e == null) {
            if (bElementIsRequired) {
                // todo : throw an exception here
            } else {
                return null;
            }
        }

        kXMLElement ee = ((ElementImpl) e).getElement();
        String strText = ee.getContents();
        if (strText == null) {
            strText = "";
        }
        return strText;
    }
    
    public String toString(Object element) {
        kXMLElement ee = ((ElementImpl) element).getElement();
        return ee.toString();
    }

    /**
     * No concept of namespace in nanoxml... What to do?
     *
     * @param uri  Description of the Parameter
     * @return     The namespace value
     */
    public sync4j.framework.core.xml.Namespace getNamespace(String uri) {
        return new NamespaceImpl(uri);
    }
}

