/*
 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.database;

import java.security.Principal;

import org.apache.commons.lang.builder.ToStringBuilder;

import sync4j.framework.core.*;

/**
 * This class represents a synchronizable database.
 *
 * @author  Stefano Fornari @ Funambol
 * @version $Id: Database.java,v 1.4 2003/04/08 20:41:14 stefano_fornari Exp $
 */
public class Database {
    
    // -------------------------------------------------------------- Properties
    
    /**
     * The database name 
     */
    private String name = null;
    
    public String getName() {
        return this.name;
    }
    
    /**
     * The database type (e.g. text/x-vcard) 
     */
    private String type = null;
    
    public String getType() {
        return this.type;
    }
    
    public void setType(String type) {
        this.type = type;
    }
    
    /** 
     * The target of the database
     */
    private Target target = null;
    
    public Target getTarget() {
        return this.target;
    }
    
    public void setTarget(Target target) {
        this.target = target;
    }
    
    /** 
     * The source of the database
     */
    private Source source = null;
    
    public Source getSource() {
        return this.source;
    }
    
    public void setSource(Source source) {
        this.source = source;
    }
    
    /**
     * Client synchronization anchor
     */
    private SyncAnchor anchor = null;
    
    public SyncAnchor getAnchor() {
        return anchor;
    }
    
    public void setAnchor(SyncAnchor anchor) {
        this.anchor = anchor;
    }
    
    /**
     * Server synchronization anchor
     */
    private SyncAnchor serverAnchor = null;
    
    public SyncAnchor getServerAnchor() {
        return serverAnchor;
    }
    
    public void setServerAnchor(SyncAnchor serverAnchor) {
        this.serverAnchor = serverAnchor;
    }
    
    /**
     * Synchronization method
     */
    private int method = AlertCode.SLOW;
    
    public int getMethod() {
        return this.method;
    }
    
    public void setMethod(int method) {
        this.method = method;
    }
    
    /**
     * Database status
     */
    private int statusCode = StatusCode.OK;
    
    public int getStatusCode() {
        return statusCode;
    }
    
    public void setStatusCode(int statusCode) {
        this.statusCode = statusCode;
    }
    
    /**
     * The alert command used to request the sychronization of this database
     */
    private AlertCommand alertCommand = null;
    
    public AlertCommand getAlertCommand() {
        return this.alertCommand;
    }
    
    public void setAlertCommand(AlertCommand alertCommand) {
        this.alertCommand = alertCommand;
    }
    
    /**
     * The associated principal
     */
    private Principal principal = null;
    
    /** Getter for property principal.
     * @return Value of property principal.
     *
     */
    public Principal getPrincipal() {
        return principal;
    }
    
    /** Setter for property principal.
     * @param principal New value of property principal.
     *
     */
    public void setPrincipal(Principal principal) {
        this.principal = principal;
    }
            
    // -------------------------------------------------------------------------
    
    /**
     * The items of the database to be added, copied, deleted, executed, replaced
     */
    private Item[] addItems       = null;
    private Item[] copyItems      = null;
    private Item[] deleteItems    = null;
    private Item[] execItems      = null;
    private Item[] replaceItems   = null;
    private Item[] existingItems  = null;
    
    public void setAddItems(Item[] items) {
        this.addItems = items;
    }
    
    public Item[] getAddItems() {
        return this.addItems;
    }
    
    public void setCopyItems(Item[] items) {
        this.copyItems = items;
    }
    
    public Item[] getCopyItems() {
        return this.copyItems;
    }
    
    public void setDeleteItems(Item[] items) {
        this.deleteItems = items;
    }
    
    public Item[] getDeleteItems() {
        return this.deleteItems;
    }
    
    public void setExecItems(Item[] items) {
        this.execItems = items;
    }
    
    public Item[] getExecItems() {
        return this.execItems;
    }
    
    public void setReplaceItems(Item[] items) {
        this.replaceItems = items;
    }
    
    public Item[] getReplaceItems() {
        return this.replaceItems;
    }
    
    public void setExistingItems(Item[] items) {
        this.existingItems = items;
    }
    
    public Item[] getExistingItems() {
        return this.existingItems;
    }
    
    // ------------------------------------------------------------ Constructors
    
    /** 
     * Creates a new instance of Database 
     *
     * @param name database idenfier - NOT NULL
     * @param name database type (e.g. text/x-vcard) - NULL
     * @param target database target - NULL
     * @param source database source - NULL
     * @param anchor database last/next anchor tags NULL
     * @param principal the associated Principal. NULL means no associated principal
     */
    public Database(String     name     ,
                    String     type     ,
                    Target     target   ,
                    Source     source   ,
                    SyncAnchor anchor   ,
                    Principal  principal) {
        if (name == null) {
            throw new NullPointerException("name cannot be null");
        }
        
        this.name      = name  ;
        this.type      = type  ;
        this.target    = target;
        this.source    = source;
        this.anchor    = anchor;
        this.principal = principal;
    }
    
    /**
     * Creates a new instance of Database 
     *
     * @param name database idenfier - NOT NULL
     */
    public Database(String name) {
        this(name, null, null, null, null, null);
    }
    
    // ---------------------------------------------------------- Public methods
    
    public String getLast() {
        return (anchor != null) ? anchor.getLast() : null;
    }
    
    public String getNext() {
        return (anchor != null) ? anchor.getNext() : null;
    }
    
    public String toString() {
        ToStringBuilder sb = new ToStringBuilder(this);
        
        sb.append("name",       name      ).
           append("type",       type      ).
           append("statusCode", statusCode).
           append("target",     target    ).
           append("source",     source    ).
           append("anchor",     anchor    ).
           append("principal",  principal );
        
        return sb.toString();
    }    
}
