/*
 
 $Id: Format.java,v 1.5 2003/12/23 12:50:07 luigiafassina Exp $
 
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */

package sync4j.framework.dm;

import sync4j.framework.core.RepresentationException;
import sync4j.framework.core.xml.Element;
import sync4j.framework.core.Util;

import java.util.Enumeration;

/**
 * This class corresponds to  &ltFormat&gt and &ltDFFormat&gt elements
 * of the SyncML DM Description Framework DTD.
 *
 * See <a href="http://java.sun.com/docs/books/effective/toc.html">p114</a>
 * on "Item 21: Replace Enum Constructs with Classes"
 *
 * Content model: (b64 | bool | chr | int | node | null | xml)>
 *
 * @author  <a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @since   SyncML 1.1
 */
public final class Format extends BaseElement
{
    public static final String B64 = "b64";
    public static final String BOOL = "bool";
    public static final String CHR = "chr";
    public static final String INT = "int";
    public static final String NODE = "node";
    public static final String NULL = "null";
    public static final String XML = "xml";
    
    private final String m_format;
    private final String m_tag;
    
    /**
     *	Constructs an instance of Format representing &ltFormat&gt or
     *	&ltDFFormat&gt, depending on the tag.
     *
     *	@param f    the format string
     *	@param tag  either &ltFormat&gt or &ltDFFormat&gt
     */
    public Format(String f, String tag)
    {
        // validate one of the formats...
        m_format = f;
        
        m_tag = tag;
    }
    
    public String getFormat()
    {
    	return m_format;
    }
    
    public String getTag()
    {
	    return m_tag;
    }
    
    /**
     * Extracts contents in &ltFormat&gt or &ltDFFormat&gt.  Notice that
     * the child will be an element without content (e.g. &lt64/&gt)
     *
     * @param tag   can be "Format" or "DFFormat"
     * @return a instance of Format.
     * @throws RepresentationException if the element has an invalid form
     */
    public static Format extractFrom(Element eParent, String tag)
	    throws RepresentationException
    {
        // will throw RepresentationException if e == null!
        Util.extractElement(eParent, tag, true);
        
        // we do not know which child, but one thing for sure:
        // there is only one child.  So we break after the first iteration.
        try 
        {
            String childName = null;
            for (Enumeration enum = eParent.getChildren(); 
                    enum.hasMoreElements();) 
            {
                childName = Util.castToElement(enum.nextElement()).getName();
                break;
            }
            return new Format(childName, tag);
        }
        catch (NullPointerException npe)
        {
            throw new RepresentationException(npe.getMessage());
        }
}
    
    /**
     * Returns an XML representation of the tag.  Uses string literals
     * so that the code will always look from the string pool instead
     * of dynamically creating the tag by reflection and wasting resources.
     *
     * @return	an XML representation of the tag.
     */
    public String toXML()
    {
        StringBuffer body = new StringBuffer();
        
        body.append("<").append(m_tag).append(">")
            .append(m_format)
            .append("</").append(m_tag).append(">\n");
        
        return body.toString();
    }
    
}
