/*
 *  $Id: MgmtTree.java,v 1.7 2003/12/23 12:50:07 luigiafassina Exp $
 *  Copyright (c) 2001-2002 sync4j project
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions, and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions, and the disclaimer that follows
 *  these conditions in the documentation and/or other materials
 *  provided with the distribution.
 *  3. The name "sync4j" must not be used to endorse or promote products
 *  derived from this software without prior written permission.
 *  4. Products derived from this software may not be called "sync4j", nor
 *  may "sync4j" appear in their name, without prior written permission.
 *  In addition, we request (but do not require) that you include in the
 *  end-user documentation provided with the redistribution and/or in the
 *  software itself an acknowledgement equivalent to the following:
 *  "This product includes software developed by the
 *  sync4j project."
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package sync4j.framework.dm;

import sync4j.framework.core.xml.Element;
import sync4j.framework.core.*;

/**
 * This class is the root element of the DTD of the SyncML DM Device
 * Description Framework. <code>MgmtTree</code> instance will have its state
 * changed when more devices are added, replaced or deleted from the device
 * management tree root.
 *
 * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @since     SyncML 1.1
 */
public class MgmtTree extends UriTree {
    /**
     * The manufacturer of the device.
     */
    private final String m_manufacturer;

    /**
     * The Version DTD
     */
    private final DTDVersion m_dtdVersion;

    /**
     * The model number of the device.
     */
    private final String m_modelNumber;

    public String getManufacturer() {
        return m_manufacturer;
    }

    
    public DTDVersion getDtdVersion() {
        return m_dtdVersion;
    }

    
    public String getModelNumber() {
        return m_modelNumber;
    }
    

    /**
     * Constructs a Device Management Tree from an XML String.
     *
     * @throws Sync4jException
     */
    public MgmtTree(String strXML) throws Sync4jException {
        this(Util.toElement(strXML, "MgmtTree"));
    }


    /**
     * Constructs a MgmtTree from a JDOM element.
     *
     * @param root                                 the root of the jdom tree,
     *      not the management tree
     * @exception Sync4jException 
     */
    MgmtTree(Element root) throws Sync4jException {
        this(
            new DTDVersion(Util.extractElementText(root, "VerDTD", true)),
            Util.extractElementText(root, "Man", false),
            Util.extractElementText(root, "Mod", false),
            extractNode(root));
    }


    /** 
     * Constructs a device managment tree based on the content model (VerDTD,
     * Man?, Mod?, Node+). The constructor performs requirement check per DTD
     * definition/spec.
     *
     * @param version
     * @param manufacturer
     * @param modelNumber
     * @param n 
     *
     * @throws Sync4jException
     */
    public MgmtTree(DTDVersion version,
                    String manufacturer,
                    String modelNumber,
                    Node n) throws Sync4jException {
        super(n);

        if (!version.getValue().equals("1.1")) {
            throw new Sync4jException(
                version.getValue() + "is not supported.");
        }

        m_dtdVersion = version;
        m_manufacturer = manufacturer;
        m_modelNumber = modelNumber;
    }


    /**
     * Converts the element into an XML string.
     *
     * @return   an XML representation of the tag
     */
    public String toXML() {
        StringBuffer body = new StringBuffer();

        body.append("<MgmtTree>");

        body.append(m_dtdVersion.toXML());

        if (m_manufacturer != null) {
            body.append("<Man>").append(m_manufacturer).append("</Man>\n");
        }

        if (m_modelNumber != null) {
            body.append("<Mod>").append(m_modelNumber).append("</Mod>\n");
        }

        body.append(m_root.toXML());

        body.append("</MgmtTree>");

        return body.toString();
    }


    /**
     * Extracts a &ltNode&gt from its JDOM parent element.
     *
     * @param eParent                  its JDOM parent
     * @return                         Description of the Return Value
     * @throws RepresentationException
     */
    private static Node extractNode(Element eParent) throws Sync4jException {
        Element e = Util.extractElement(eParent, "Node", true);
        if (e == null) {
            throw new Sync4jException("Missing Node");
        }
        return new Node(e);
    }


    /**
     * For DMAcc
     *
     * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
     */
    public static class AddrType {
        public static final int HTTP = 1;
        public static final int WSP = 2;
        public static final int OBEX = 3;
    }


    /**
     * For bearer
     *
     * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
     */
    public static class Bearer {
        public static final int OBEX = 1;
        public static final int GSM_USSD = 2;
    }


    /**
     * For NAP AddrType
     *
     * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
     */
    public static class NAPAddrType {
        public static final int IPV4 = 1;
        public static final int IPV6 = 2;
    }
}

