/*
 *  $Id: Node.java,v 1.5 2003/12/23 12:50:07 luigiafassina Exp $
 *  Copyright (c) 2001-2002 sync4j project
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions, and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions, and the disclaimer that follows
 *  these conditions in the documentation and/or other materials
 *  provided with the distribution.
 *  3. The name "sync4j" must not be used to endorse or promote products
 *  derived from this software without prior written permission.
 *  4. Products derived from this software may not be called "sync4j", nor
 *  may "sync4j" appear in their name, without prior written permission.
 *  In addition, we request (but do not require) that you include in the
 *  end-user documentation provided with the redistribution and/or in the
 *  software itself an acknowledgement equivalent to the following:
 *  "This product includes software developed by the
 *  sync4j project."
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package sync4j.framework.dm;

import sync4j.framework.core.xml.Element;
import sync4j.framework.core.*;

import java.util.Enumeration;
import java.util.Hashtable;

/**
 * This class corresponds to &ltNode&gt element of the SyncML Device
 * Management Tree and Description Framework DTD. <p>
 *
 * Node is the representation of an instantiated object. Note that this
 * element is recursive and that a Node with a Value tag must always terminate
 * the recursion. It is possible for a Node to omit both the next recursive
 * Node and a Value, this means that the hierarchy of Nodes continues
 * elsewhere. This can be used to increase readability of very deep trees.
 * <!ELEMENT Node (NodeName, Path?, RTProperties?, DFProperties, (Node* |
 * Value?))>
 *
 * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @since     SyncML 1.1
 */
public class Node extends UriNode {
    // NodeName, Path and Node* are in UriNode

    // 6.7.1.2
    /**
     * Description of the Field
     */
    public final static String DEFAULT_ROOT_ACL = "Add=*&Replace=*";

    /**
     * Run time properties
     *
     * @see   RTProperties
     */
    private RTProperties m_rtProperties;

    /**
     * Description framework properties
     *
     * @see   DFProperties
     */
    private DFProperties m_dfProperties;

    /**
     * Value of this node, if this node does not have a subtree
     *
     * @see   UriNode#m_subTree
     */
    private String m_value;


    /**
     * Constructs a node from other sub-elements, in its Java representation.
     *
     * @param nodeName      the name of node; must be present
     * @param path          the path
     * @param rtProperties  the runtime properties
     * @param dfProperties  the device framework properties; must be present
     * @param children      the nodes
     * @param value         a value for this node if no children
     */
    public Node(String nodeName,
                String path,
                RTProperties rtProperties,
                DFProperties dfProperties,
                Hashtable children,
                String value) {
        super(nodeName, path, children);
        // could have enforced 6.7.1.2, forcing the root ACL to be
        // DEFAULT_ROOT_ACL
        m_rtProperties = rtProperties;
        if (dfProperties == null) {
            throw new IllegalArgumentException(
                "DFProperties is a REQUIRED element.");
        }
        m_dfProperties = dfProperties;
        m_value = value;
    }


    /**
     * Constructs a node from a JDOM tree element.
     *
     * @param e                           the tree element
     * @exception RepresentationException  Description of the Exception
     * @throws RepresentationException if the JDOM
     *      element does not have a valid markup (XML) sub document.
     * @see                               MgmtTree#MgmtTree(Element)
     */
    public Node(Element e) throws RepresentationException {
        this(
            extractNodeName(e),
            extractPath(e),
            RTProperties.extractFrom(e),
            DFProperties.extractFrom(e),
            extractSubNode(e),
            Util.extractElementText(e, "Value", false));
    }


    /**
     * @return   an instance of RTProperties, the run-time properties.
     */
    public RTProperties getRTProperties() {
        return m_rtProperties;
    }


    /**
     * @return   an instance of DFProperties, the device format properties.
     */
    public DFProperties getDFProperties() {
        return m_dfProperties;
    }


    /**
     * @return   the value
     */
    public String getValue() {
        return m_value;
    }


    /**
     * Returns an XML representation of the &ltNode&gt element. Since the
     * SyncML documents are usually used internally instead of being displayed
     * for human reading, indentation is not introduced while formatting the
     * XML document.
     *
     * @return   the &ltNode&gt, as an XML document.
     */
    public String toXML() {
        StringBuffer body = new StringBuffer();

        body.append("<Node>");

        body.append("<NodeName>")
            .append(getNodeName())
            .append("</NodeName>\n");

        // path is an optional element; if it is null, we don't want
        // <Path>null</Path>
        if (m_path != null) {
            body.append("<Path>").append(m_path.trim()).append("</Path>\n");
        }

        // run-time properties is an optional element
        if (m_rtProperties != null) {
            body.append(m_rtProperties.toXML());
        }

        body.append(m_dfProperties.toXML());

        if (m_value != null) {
            body.append("<Value>").append(m_value.trim()).append("</Value>\n");
        } else {
            // "embedded" approach per 7.4.1.5;
            // will refactor to include node-wise approach per 7.4.1.7
            Hashtable subTree = getSubTree();
            if (subTree != null) {
                for (Enumeration enum = subTree.elements(); enum.hasMoreElements(); ) {
                    body.append(((Node) enum.nextElement()).toXML());
                }
            }
        }

        body.append("</Node>\n");

        return body.toString();
    }


    /**
     * Extracts &ltNodeName&gt according to 7.4.1.6. &ltNodeName&gt can be an
     * empty tag as an input, but it must have a value according to the DTD.
     *
     * @param e                           The JDOM element
     * @return                            a node name
     * @exception RepresentationException  Description of the Exception
     */
    private static String extractNodeName(Element e)
         throws RepresentationException {
        String nodeName = Util.extractElementText(e, "NodeName", false);
        if (nodeName.equals("")) {
            nodeName = "x";
        }
        return nodeName;
    }


    /**
     * Extracts &ltPath&gt according to 7.4.1.7. If the path of this element
     * has not been specified, the element's parent will be queried for the
     * path. This method will be called recursively all the way to the top,
     * i.e. the root.
     *
     * @param e                        the element
     * @return                         the path of the element
     * @throws RepresentationException  if there's something wrong with markup.
     */
    private static String extractPath(Element e)
         throws RepresentationException {
        String path = null;
        StringBuffer sb = new StringBuffer();

        path = Util.extractElementText(e, "Path", false);
        // 7.4.1.7 recursive lookup
        if (path == null) {
            Element parent = e.getParent();
            path = extractPath(parent);
            String parentName = extractNodeName(parent);

            if (path.equals("./")) {
                sb.append(path).append(parentName);
            } else {
                sb.append(path).append(PATH_SEPARATOR).append(parentName);
            }
            parentName = null;
        } else {
            sb.append(path);
        }

        path = null;
        return sb.toString();
    }


    /**
     * Extracts the sub node/subtree from a &ltNode&gt element, which is
     * represented in a JDOM element. <p>
     *
     * If there is no subtree, this node is a leaf object. A <code>null</code>
     * will be returned. <p>
     *
     * On the other hand, it will iteratively search for all subnodes to put
     * them into a subtree. The subtree will be returned. <p>
     *
     * Note: Currently this extraction does not support input in the format of
     * 7.4.1.7.
     *
     * @param eParent                     this node, or the parent of the sub
     *      node/subtree to be extracted
     * @return                            the subtree, which is represented in
     *      a <code>Hashtable</code>.
     * @exception RepresentationException  Description of the Exception
     * @throws RepresentationException if the XML
     *      document in <code>eParent</code> is malformed.
     * @see                               UriNode#createSubTree()
     */
    private static Hashtable extractSubNode(Element eParent)
         throws RepresentationException {
        Hashtable subTree = null;

        // it may have multiple nodes in this subtree
        for (Enumeration enum = eParent.getChildren("Node");
            enum.hasMoreElements(); ) {
            Element child = Util.castToElement(enum.nextElement());
            // get the node name of this child node
            // i.e. <NodeName> of this child <Node>
            String nodeName = child.getChildText("NodeName");
            // create a new subtree only if it's the first time
            if (subTree == null) {
                subTree = Node.createSubTree();
            }
            // put subnodes into the subTree
            subTree.put(nodeName, new Node(child));
        }
        return subTree;
    }
}

