/*
 
 $Id: Occurrence.java,v 1.4 2003/12/23 12:50:07 luigiafassina Exp $
 
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */

package sync4j.framework.dm;

import sync4j.framework.core.xml.Element;
import sync4j.framework.core.*;

import java.util.Enumeration;

/**
 * This class corresponds to  &ltOccurrence&gt element
 * of the SyncML DM Description Framework DTD.
 *
 * See <a href="http://java.sun.com/docs/books/effective/toc.html">p114</a>
 * on "Item 21: Replace Enum Constructs with Classes"
 *
 * Content model: (One | ZeroOrOne | ZeroOrMore
 * | OneOrMore | ZeroOrN | OneOrN )>
 *
 * @author	<a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @since	SyncML 1.1
 */
public final class Occurrence extends BaseElement
{
    public static final String ONE = "One";
    public static final String ZEROORONE = "ZeroOrOne";
    public static final String ZEROORMORE = "ZeroOrMore";
    public static final String ONE0RMORE = "OneOrMore";
    public static final String ZEROORN = "ZeroOrN";
    public static final String ONEORN = "OneOrN";
    
    /**
     * For <!ELEMENT ZeroOrN (#PCDATA)> and <!ELEMENT OneOrN (#PCDATA)>
     */
    private final String m_n;
    private final String m_occurrence;
    
    /**
     *	Constructs an instance to represent &ltOccurrence&gt.
     *
     *	@param o	the occurrence string
     */
    public Occurrence(String o, String n)
	    throws IllegalArgumentException
    {
        // 7.4.3.18 & 7.4.3.20
        if (n != null)
        {
            int i = Integer.parseInt(n);
            if (i < 2 || i > 65536)
            {
            throw new IllegalArgumentException(
                "Value for " + o + "should be between 2 and 65536");
            }
        }
        
        m_occurrence = o;
        m_n = n;
    }
    
    /**
     *
     * @return the occurrence string.
     */
    public String getOccurrence()
    {
	    return m_occurrence;
    }
    
    /**
     *
     * @return the number of occurrence, if it exists, otherwise null.
     */
    public String getN()
    {
	    return m_n;
    }
    
    /**
     * Extracts &ltOccurrence&gt from its JDOM parent element.  
     * The special case is that if the occurrence is either:
     *
     *	    &lt!ELEMENT ZeroOrN (#PCDATA)&gt
     * or
     *	    &lt!ELEMENT OneOrN (#PCDATA)&gt
     *
     * their content has to be extracted also.
     *
     * @param eParent the JDOM eParent element
     * @return an instance of <code>Occurrence</code>
     * @throws RepresentationException
     */
    public static Occurrence extractFrom(Element eParent)
        throws RepresentationException
    {
        Element e = Util.extractElement(eParent, "Occurrence", false);
        try
        {
            if (e != null)
            {
                Element child = null;
                String childName = null;
                // always only one children; if not, error!
                for (Enumeration enum = eParent.getChildren();
                    enum.hasMoreElements();)
                {
                    child = Util.castToElement(enum.nextElement());
                    childName = child.getName();
                    break;
                }
                String n = null;
                if (childName != null)
                {
                    if (childName.equals(ZEROORN) || childName.equals(ONEORN))
                    {
                        n = child.getText();
                        return new Occurrence(childName, n);
                    }
                    else
                    {
                        return new Occurrence(childName, null);
                    }
                }
                else
                {
                    throw new RepresentationException("<Occurrence> has no child");
                }
            }
            // 7.4.3.5
            return new Occurrence(ONE, null);
        }
        catch (IllegalArgumentException iae)
        {
            throw new RepresentationException(iae.getMessage());
        }
    }

    /**
     * Converts the tag content to an XML doc.
     * 
     * @see Occurrence#extractFrom(Element)
     */
    public String toXML()
    {
        StringBuffer body = new StringBuffer();
        
        body.append("<Occurrence>").append("<").append(m_occurrence);
        
        if ((m_occurrence.equals(ZEROORN)) ||
        (m_occurrence.equals(ONEORN)))
        {
            body.append(">").append(m_n)
            .append("</")
            .append(m_occurrence)
            .append(">\n");
        }
        else
        {
            body.append("/>");
        }
        
        body.append("</Occurrence>\n");
        
        return body.toString();
    }
    
}
