/*
 *  $Id: UriNode.java,v 1.2 2003/12/15 16:55:41 luigiafassina Exp $
 *  Copyright (c) 2001-2002 sync4j project
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions, and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions, and the disclaimer that follows
 *  these conditions in the documentation and/or other materials
 *  provided with the distribution.
 *  3. The name "sync4j" must not be used to endorse or promote products
 *  derived from this software without prior written permission.
 *  4. Products derived from this software may not be called "sync4j", nor
 *  may "sync4j" appear in their name, without prior written permission.
 *  In addition, we request (but do not require) that you include in the
 *  end-user documentation provided with the redistribution and/or in the
 *  software itself an acknowledgement equivalent to the following:
 *  "This product includes software developed by the
 *  sync4j project."
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package sync4j.framework.dm;

import sync4j.framework.core.*;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * This class implements a node in a UriTree. The node keeps its sub tree in a
 * <code>Hashtable</code> collection. The method <code>createSubTree</code>
 * defines the instance The state of <code>UriNode</code> changes as it may
 * have more/less children. The mutable data members in this class are
 * synchronized for concurrent access.
 *
 * @author    <a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @see       Node
 * @since     SyncML 1.1
 */
public class UriNode extends BaseElement {
    /**
     * Description of the Field
     */
    public final static String ROOT_PATH = ".";
    /**
     * Description of the Field
     */
    public final static String PATH_SEPARATOR = "/";

    /**
     * Name of the node. It should not have any path information.
     */
    protected String m_nodeName;

    /**
     * Path of the node. This implies where its parent is.
     */
    protected final String m_path;

    /**
     * The children of the node.
     */
    protected Hashtable m_subTree;


    /**
     * Constructs a node.
     *
     * @param path      the node's URI, without its name.
     * @param children  the node's children; it can be <tt>null</tt> .
     * @param nodeName  Description of the Parameter
     */
    public UriNode(final String nodeName, final String path, Hashtable children) {
        if (nodeName == null) {
            throw new NullPointerException("NodeName is null");
        }
        m_nodeName = nodeName;
        m_path = path;
        m_subTree = children;
    }


    /**
     * @return   the full name of this node
     */
    public String getAbsoluteName() {
        return getPath() + getNodeName();
    }


    /**
     * @return   name of the node
     */
    public synchronized String getNodeName() {
        return m_nodeName.trim();
    }


    /**
     * Sets the node name.
     *
     * @param n  the new node name to be set
     */
    public synchronized void setNodeName(String n) {
        m_nodeName = n;
    }


    /**
     * @return   the path in URI format.
     */
    public String getPath() {
        return m_path;
    }


    /**
     * @return   the children of the node, also known as the interior objects.
     */
    public synchronized Hashtable getSubTree() {
        return m_subTree;
    }


    /**
     * Inserts a node under the current node.
     *
     * @param n                      the node to be added
     * @throws NullPointerException  if n is <tt>node</tt>
     */
    synchronized void insert(UriNode n) {
        if (isLeaf()) {
            m_subTree = createSubTree();
        }

        m_subTree.put(n.getNodeName(), n);
    }


    /**
     * Removes the target node and its children.
     *
     * @param target                 the target node to be removed.
     * @throws NullPointerException  if target is <tt>null</tt>
     * @see                          UriTree#remove
     */
    synchronized void remove(UriNode target) {
        // remove its name-value from the subTree Map first
        m_subTree.remove(target.getNodeName());
    }


    /**
     * Returns a node object based on the URI specified. This method has
     * default access because API user should really be using <code>UriTree.resolveName</code>
     *
     * @param uri  the path to be examined.
     * @return     the last node reachable according to the uri; <tt>null</tt>
     *      if that node does not exists.
     * @see        UriTree#resolveName(String)
     */
    synchronized UriNode resolveName(String uri) {
        UriNode curUriNode = this;

        // short-cut
        if (ROOT_PATH.equals(uri)) {
            return curUriNode;
        }

        // supports lookup of either way of addressing in 6.3
        String fullPath = (uri.charAt(0) == '.') ? uri.substring(1) : uri;
        Enumeration st = tokenize(fullPath, PATH_SEPARATOR);

        Hashtable curBranch = curUriNode.getSubTree();
        // parsing the path
        if (st.hasMoreElements()) {
            String curToken = (String) st.nextElement();
            String temp = curUriNode.getNodeName();
            if (curToken.equals(temp)) {
                if (curUriNode.isLeaf() && st.hasMoreElements()) {
                    // curUriNode is a leaf object, and
                    // the path has not been fully parsed
                    return null;
                }
                while (st.hasMoreElements()) {
                    curToken = (String) st.nextElement();
                    curUriNode = (UriNode) curBranch.get(curToken);
                    if (curUriNode != null) {
                        curBranch = curUriNode.getSubTree();
                    } else {
                        break;
                    }
                }
            }
        } else {
            curUriNode = null;
        }

        fullPath = null;
        return curUriNode;
    }


    /**
     * Determines whether it's a leaf node.
     *
     * @return   true if it's a leaf node, false if it has interior objects.
     */
    public synchronized boolean isLeaf() {
        return (m_subTree == null);
    }


    /**
     * Determines whether this node is a root node. A root node will have its
     * path as ".". DeviceInfo and DeviceDetail objects may use this tree, too
     * (see DM Standard Object spec). Therefore, the root node can be:
     * ./SyncML ./DevInfo ./DevDetail
     *
     * @return   <tt>true</tt> this node fits the root node criteria.
     */
    protected boolean isRootNode() {
        return m_path.equals(ROOT_PATH);
    }


    /**
     * Gets the size, which is number of nodes under this node and itself.
     *
     * @return   number of nodes
     */
    synchronized int getSize() {
        int retVal = 1;

        if (!isLeaf()) {
            for (Enumeration enum = m_subTree.elements();
                enum.hasMoreElements(); ) {
                retVal += ((UriNode) enum.nextElement()).getSize();
            }
        }
        return retVal;
    }

    // default for Hashtable is 11.  We'll choose 5 instead.
    final static int INIT_CAP = 5;


    /**
     * Creates a subTree. It uses <code>Hashtable</code> for faster retrieval
     * of node according to its nodeName.
     *
     * @return   a subTree instance.
     */
    protected static Hashtable createSubTree() {
        return new Hashtable(INIT_CAP);
    }


    /**
     * Description of the Method
     *
     * @return   Description of the Return Value
     */
    protected String toXML() {
        // should never call here
        return null;
    }


    /**
     * Replaces <code>StringTokenizer</code> in J2SE. Done for J2ME.
     *
     * @param str    Description of the Parameter
     * @param delim  Description of the Parameter
     * @return       Description of the Return Value
     */
    private Enumeration tokenize(final String str, final String delim) {
        Vector tokens = new Vector();
        char[] aChar = delim.toCharArray();
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < str.length(); i++) {
            char currChar = str.charAt(i);
            if (currChar != aChar[0]) {
                buf.append(currChar);
            } else {
                if (buf.length() > 0) {
                    tokens.addElement(buf.toString());
                    buf = new StringBuffer();
                }
            }
        }
        if (buf.length() > 0) {
            tokens.addElement(buf.toString());
        }

        return tokens.elements();
    }
}

