/*
 
 $Id: UriTree.java,v 1.1 2002/08/09 15:56:14 stefano_fornari Exp $
 
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 
 */

package sync4j.framework.dm;

import sync4j.framework.core.*;
import java.util.*;

/**
 * This class implements a modifiable URI tree.
 *
 * <code>UriTree</code> is thread-safe; it is synchronized at root level.
 * Most of the operations in this class goes through its own root node.
 * Since the operations in the root node is synchronized, the client
 * do not have to synchronize the instance of <code>UriTree</code>.
 *
 * @author  <a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @see	UriNode
 * @since   SyncML 1.1
 */
public class UriTree extends BaseElement
{
    /**
     * root of the tree
     */
    protected UriNode m_root;
    
    /**
     * Constructs a tree object without a root node.
     */
    public UriTree()
    {
	m_root = null;
    }
    
    protected UriTree(UriNode n)
    {
	m_root = n;
    }
    
    /**
     *
     * @return	the root node of this tree.
     */
    public UriNode getRoot()
    {
	return m_root;
    }
    
    /**
     * Returns number of nodes in this tree.
     * NULL tree node does not count as one node.
     * @return	number of nodes in this tree.
     */
    public int size()
    {
	return m_root.getSize();
    }
    
    /**
     *	Adds a node to this tree.  Duplicate node will be replaced.
     *	Note: business logic/checks pertaining to DM framework is done
     *	outside of this add().
     *
     *	@param n    the node added into this tree
     *	@return	<code>true</code> if the node is added.
     *	@throws	NullPointerException if the object is <tt>null</tt>.
     */
    public boolean add(UriNode n)
    {
	boolean retVal = false;
	
	if (m_root == null)
	{
	    if (!n.getPath().equals(UriNode.ROOT_PATH))
	    {
		// cannot add!
	    } 
	    else
	    {
		m_root = n;
		retVal = true;
	    }
	} 
	else
	{
	    UriNode parent = m_root.resolveName(n.getPath());
	    if (parent != null)
	    {
		parent.insert(n);
		retVal = true;
	    }
	    else
	    {
		// can't find the path!
	    }
	}
	
	return retVal;
    }
    
    /**
     * Removes a node from this tree.  After calling <code>remove</code>,
     * the client code has to set the parameter object to null to ensure
     * memory obtained by the parameter object is freed for garbage
     * collection.
     *
     * @param n the node to be removed
     * @return	<code>true</code> if the node is removed.
     * @throws	NullPointerException if the object is <tt>null</tt>
     */
    public boolean remove(UriNode n)
    {
	boolean retVal = false;
	
	if (m_root != null)
	{
	    UriNode parent = m_root.resolveName(n.getPath());
	    if (parent != null)
	    {
		parent.remove(n);
		retVal = true;
	    }
	}
	
	return retVal;
    }
    
    /**
     * Resolves the URI from the root.  For example, if the URI is
     * "./Vendor/RingSignals/", the node returned will be at RingSignals.
     * It is an entry point to <code>resolveName</code> in
     * <code>UriNode</code>.
     *
     * @param uri	the URI of the node in question
     * @return the node in the tree; <tt>null</tt> if not found.
     * @see UriNode#resolveName(String)
     */
    public UriNode resolveName(String uri)
    {
	return m_root.resolveName(uri);
    }
        
    protected String toXML()
    {
	// should never call here
	return null;
    }
}

