/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 */

package sync4j.framework.engine.source;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Properties;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.security.Principal;
import java.sql.Timestamp;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;

import sync4j.framework.tools.Base64;
import sync4j.framework.tools.IOTools;

/**
 *
 * @author Stefano Fornari @ Funambol
 *
 * LOG NAME: sync4j.source
 *
 * verison $Id: FileSystemSyncSource.java,v 1.29 2003/12/15 16:56:17 luigiafassina Exp $
 */
public class FileSystemSyncSource
extends AbstractSyncSource
implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    public static final String DATABASE_FILE_NAME    = "sync.db";
    public static final String DATABASE_HEADER       = "FileSystemSyncSource file database";

    public static final String LOG_NAME = "sync4j.framework.engine";

    // -------------------------------------------------------------- Properties

    /**
     * The directory where files are stored (the default is the current
     * directory) - read/write
     */
    private String sourceDirectory = ".";

    public void setSourceDirectory(String sourceDirectory) {
        this.sourceDirectory = sourceDirectory;
    }

    public String getSourceDirectory() {
        return this.sourceDirectory;
    }

    /**
     * Should the content be encoded?
     */
    private boolean encode = true;

    /** Getter for property encode.
     * @return Value of property encode.
     *
     */
    public boolean isEncode() {
        return encode;
    }

    /** Setter for property encode.
     * @param encode New value of property encode.
     *
     */
    public void setEncode(boolean encode) {
        this.encode = encode;
    }

    // ------------------------------------------------------------ Constructors

    public FileSystemSyncSource() {}

    public FileSystemSyncSource(String name, String sourceDirectory) {
        super(name);
        this.sourceDirectory = sourceDirectory;
        setSourceURI(name);
    }

    /**
     * The name defaults to the directory name
     */
    public FileSystemSyncSource(String sourceDirectory) {
        this(new File(sourceDirectory).getName(), sourceDirectory);
    }

    // ---------------------------------------------------------- Public Methods

    /*
    * @see SyncSource
    */

    public SyncItem setSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {
        try {
            String fileName = syncItem.getKey().getKeyAsString();
            byte[] fileContent =
                (byte[])syncItem.getPropertyValue(SyncItem.PROPERTY_BINARY_CONTENT);

            if (fileContent == null) {
                fileContent = new byte[0];
            }
            
            File f = new File(sourceDirectory, fileName);
            if (encode) {
                IOTools.writeFile(Base64.decode(fileContent), f);
            } else {
                IOTools.writeFile(fileContent, f);
            }            
           
            Timestamp t = 
                (Timestamp)syncItem.getPropertyValue(SyncItem.PROPERTY_TIMESTAMP);
                        
            f.setLastModified(t.getTime());

            setState(principal, fileName, SyncItemState.UPDATED, t);

            SyncItem newSyncItem =
                new SyncItemImpl(this, fileName, SyncItemState.NEW);

            newSyncItem.setProperties(syncItem.getProperties());

            return newSyncItem;
        } catch (IOException e) {
            throw new SyncSourceException( "Error setting the item "
                                          + syncItem
                                          , e
                                          );
        }
    }

    /*
    * @see SyncSource
    */

    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();
        ArrayList ret              = new ArrayList();

        for (int i=0; i<syncItems.length; ++i) {
            try {
                ret.add(setSyncItem(principal, syncItems[i]));
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException( "Error setting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }

        return (SyncItem[])ret.toArray(new SyncItem[] {});
    }

    /*
    * @see SyncSource
    */
    public SyncItem getSyncItemFromId(Principal principal, SyncItemKey syncItemKey)
    throws SyncSourceException{
        SyncItem[] syncItems =
            getSyncItemsFromIds(principal, new SyncItemKey[] {syncItemKey});

        if ((syncItems == null) || (syncItems.length == 0)) {
            return null;
        }

        return syncItems[0];
    }

    /*
    * @see SyncSource
    */
    public SyncItem[] getSyncItemsFromIds(Principal principal, SyncItemKey[] syncItemKeys)
    throws SyncSourceException{
        Properties syncDatabase = updateSyncDatabase(principal);

        ArrayList syncItems = new ArrayList();

        SyncItem  syncItem  = null;
        String    fileName  = null, stateString = null;
        char      fileState = ' ';
        for (int i=0; ((syncItemKeys != null) && (i<syncItemKeys.length)); ++i) {
            fileName = syncItemKeys[i].getKeyAsString();
            stateString = syncDatabase.getProperty(fileName);
            if (stateString != null) {
                fileState = stateFromStateString(stateString);
                syncItem = new SyncItemImpl(this, fileName, fileState);

                try{
                    if (encode) {
                        syncItem.setProperty(
                            new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                             Base64.encode(IOTools.readFileBytes(new File(sourceDirectory, fileName))))
                        );
                    } else {
                        syncItem.setProperty(
                            new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                             IOTools.readFileBytes(new File(sourceDirectory, fileName)))
                        );
                    }
                } catch (IOException e) {
                    throw new SyncSourceException("Error reading " + fileName, e);
                }

                syncItems.add(syncItem);
            }
        }

        return (SyncItem[]) syncItems.toArray(new SyncItem[syncItems.size()]);
    }

    /*
    * @see SyncSource
    */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /*
    * @see SyncSource
    */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since    )
    throws SyncSourceException {
        return filterSyncItems(principal, since, String.valueOf(SyncItemState.NEW));
    }

    /*
    * @see SyncSource
    */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /*
    * @see SyncSource
    */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since    )
    throws SyncSourceException {
        return filterSyncItems(principal, since, String.valueOf(SyncItemState.DELETED));
    }

    /**
    * @see SyncSource
    */
    public SyncItem[] getUpdatedSyncItems(Principal principal,
                                          Timestamp since    ) throws SyncSourceException {
        return filterSyncItems(principal, since, String.valueOf(SyncItemState.UPDATED));
    }

    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since    ) throws SyncSourceException {
        return null;
    }

    /*
    * @see SyncSource
    */
    public void removeSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {
        String fileName = syncItem.getKey().getKeyAsString();

        new File(sourceDirectory, fileName).delete();

        //set item to state DELETE
        Timestamp t = (Timestamp)syncItem.getPropertyValue(SyncItem.PROPERTY_TIMESTAMP);
        setState(principal, fileName, SyncItemState.DELETED, t);

    }

    /*
    * @see SyncSource
    */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {
        ArrayList syncItemsInError = new ArrayList();

        for (int i=0; i<syncItems.length; ++i) {
            try {
                removeSyncItem(principal, syncItems[i]);
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException( "Error deleting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }
    }

    /**
    * @see SyncSource
    */
    public SyncItem[] getAllSyncItems(Principal principal)
    throws SyncSourceException {
        return filterSyncItems(principal, null, "UNS");
    }

    /**
     * The same as <i>getSyncItemFromId()</i>
     * @see SyncSource
     *
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem)
    throws SyncSourceException {
        return getSyncItemFromId(principal, syncItem.getKey());
    }

    /**
     * The same as <i>getSyncItemFromIds()</i>
     *
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {
        ArrayList items = new ArrayList();
        for (int i = 0; i<syncItems.length; ++i){
            items.add(getSyncItemFromId(principal, syncItems[i].getKey()));
        }

        return (SyncItem[])items.toArray(new SyncItem[items.size()]);
    }

    /**
     * A string representation of this object
     */
    public String toString() {
        StringBuffer sb = new StringBuffer(super.toString());

        sb.append(" - { ");
        sb.append("name: ").append(name).append(", ");
        sb.append("sourceDirectory: ").append(sourceDirectory).append(" }");

        return sb.toString();
    }


    // --------------------------------------------------------- Private methods

    /**
     * The synchronization database is stored in a file whose name is given
     * by the value of the constant DATABASE_FILE_NAME prefixed by the user name
     * and the device id. For example:
     * <pre>
     *   guest.Sync4jTest.sync.db
     * </pre>
     * The database is a property file where each entry has the following
     * format:
     * <pre>
     * [filename]=[state][lastmodified_timestamp]
     * </pre>
     * For example:
     * <blockquote>
     * readme.txt=U98928743098094
     * </blockquote>
     * <p>
     * updateSyncDatabase works as follows:
     * <pre>
     * 1.    Read the existing database (if it exists)
     * 2.    Scan the source directory getting all files in the directory
     * 3.    For each file f in the source directory
     * 3.1.    If f is already in the database
     * 3.1.1.    If f has been modified after the lastmodified_timestamp stored into the database
     * 3.1.1.1.    Set the state of the file to UPDATE and store the new lastmodified_timestamp
     * 3.2.    Else
     * 3.2.1.    Add f to the database setting its state to NEW and store the lastmodified_timestamp
     * 3.    End For each
     * 4.    For each file f in the database
     * 4.1.    If f does not exist in the source directory
     * 4.1.1.    Set the state to DELETED
     * 5.    End For each
     * </pre>
     * <p>
     * At the end of the process, the updated database is saved and than returned.
     *
     * @return the updated databse. In case of error, an error message is traced
     *         and an empty Properties object is returned.
     */
    private Properties updateSyncDatabase(Principal principal) {

        Properties syncDatabase = new Properties();

        try {
            File fileSyncDatabase = getDatabaseFile(principal);

            //
            // Reads the existing database
            //
            if (fileSyncDatabase.exists()) {
                FileInputStream fis = new FileInputStream(fileSyncDatabase);
                syncDatabase.load(fis);
                fis.close();
            }

            //
            // Get the list of files in the source directory
            //
            ArrayList existingFiles = getExistingFiles();

            //
            // Get the list of the file in the databsae
            //
            Enumeration databaseFiles = syncDatabase.propertyNames();

            String state    = null,
                   fileName = null;

            long lastModified;

            for (Iterator i = existingFiles.iterator(); i.hasNext();) {
                fileName = (String)i.next();
                lastModified = new File(sourceDirectory, fileName).lastModified();

                state = syncDatabase.getProperty(fileName);

                if (state != null) {
                    //
                    // The file is already in the database
                    //
                    if (lastModified > lastModifiedFromStateString(state)) {
                        state = buildStateString(SyncItemState.UPDATED, lastModified);
                        syncDatabase.setProperty(fileName, state);
                    }
                } else {
                    //
                    // The file is not in the database
                    //
                    state = buildStateString(SyncItemState.NEW, lastModified);
                    syncDatabase.setProperty(fileName, state);
                }
            }  // next i

            for (; databaseFiles.hasMoreElements();) {
                fileName = (String)databaseFiles.nextElement();

                if (!existingFiles.contains(fileName)) {
                    state = buildStateString(SyncItemState.DELETED, System.currentTimeMillis());
                    syncDatabase.setProperty(fileName, state);
                }
            }  // next

            //
            // Save & return
            //
            FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
            syncDatabase.store(fos, DATABASE_HEADER);
            fos.close();
        } catch (IOException e) {
            Logger log = Sync4jLogger.getLogger("source");
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error updating sync database: " + e.getMessage());
            }
            Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "updateSyncDatabase", e);
        }

        return syncDatabase;
    }

    private File getDatabaseFile(Principal principal) {
        return new File(sourceDirectory, DATABASE_FILE_NAME);
    }

    private String buildStateString(char state, long lastModified) {
        return state + String.valueOf(lastModified);
    }

    private long lastModifiedFromStateString(String state) {
        return Long.parseLong(state.substring(1));
    }

    private char stateFromStateString(String state) {
        if ((state == null) || (state.length() == 0)) return SyncItemState.UNKNOWN;

        return state.charAt(0);
    }

    private ArrayList getExistingFiles() throws IOException {
        ArrayList ret = new ArrayList();

        String[] files = new File(sourceDirectory).list();

        if (files != null) {
            for (int i = 0; i<files.length; ++i) {
                if (!files[i].equals(DATABASE_FILE_NAME)) {
                    ret.add(files[i]);
                }
            }  // next i
        }

        return ret;
    }

    private void setState(Principal principal, 
                          String    file     ,
                          char      state    ,
                          Timestamp t        ) {
        try {
            Properties syncDatabase = new Properties();

            File fileSyncDatabase = getDatabaseFile(principal);

            //
            // Reads the existing database
            //
            if (fileSyncDatabase.exists()) {
                FileInputStream fis = new FileInputStream(fileSyncDatabase);
                syncDatabase.load(fis);
                fis.close();
            }

            syncDatabase.setProperty(file, buildStateString(state, t.getTime()));

            FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
            syncDatabase.store(fos, DATABASE_HEADER);
            fos.close();
        } catch (IOException e) {
            Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "setState", e);
        }
    }

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, last sync timestamp and state (see
     * SyncItemState). If state is equals to UNKNOWN all items are returned.<br>
     * Note that the current implementation ignores the principal: data do not
     * depend on users.
     *
     * @param principal principal. null means any
     * @param since     last sync timestamp. null neans since ever
     * @param states    A string containing the the requested states (e.g. "UN ")
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal principal,
                                       Timestamp since    ,
                                       String    states   )
    throws SyncSourceException {
        Properties syncDatabase = updateSyncDatabase(principal);

        ArrayList syncItems = new ArrayList();

        long fileTimestamp,
             sinceTimestamp = (since == null) ? -1 : since.getTime();

        SyncItem  syncItem      = null;
        String    fileName      = null;
        String    stateString   = null;
        char      fileState           ;
        for (Enumeration e = syncDatabase.keys(); e.hasMoreElements(); ) {
            fileName  = (String)e.nextElement();
            stateString = (String)syncDatabase.get(fileName);
            fileState = stateFromStateString(stateString);
            if (states.indexOf(fileState) >= 0) {
                fileTimestamp = lastModifiedFromStateString(stateString);
                //
                // We check if the file has changed after the since timestamp
                // with a resolution of one second. This is because some dbms
                // (such as MySQL) have a timestamp resolution of max 1 second.
                //
                if ((fileTimestamp - sinceTimestamp) > 1000 ) {
                    syncItem = new SyncItemImpl(this, fileName, fileState);

                    if (fileState != SyncItemState.DELETED) {
                        try {
                            if (encode) {
                                syncItem.setProperty(
                                    new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT ,
                                                     Base64.encode(IOTools.readFileBytes(new File(sourceDirectory, fileName))))
                                );
                            } else {
                                syncItem.setProperty(
                                    new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                                     IOTools.readFileBytes(new File(sourceDirectory, fileName)))
                                );
                            }
                        } catch (IOException ioe) {
                            throw new SyncSourceException( "Error reading "
                                                         + fileName
                                                         , ioe
                                                         );
                        }
                    }
                    syncItems.add(syncItem);
                }
            }
        }  // next e

        return (SyncItem[])syncItems.toArray(new SyncItem[0]);
    }
}
