/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*/
package sync4j.framework.engine.source;

import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;

import java.security.Principal;
import java.sql.Timestamp;

/**
 * A <i>SyncSource</i> is responsible for the storing and retrieving of 
 * <i>SyncItem</i> objects. It is used also for getting newly created or removed
 * <i>SyncItem</i>s.<br>
 * Note that the <i>SyncSource</i> inteface doesn't make any assumption about the
 * underlying data source. Each concrete implementation will use the storage its
 * specific database.
 *
 * @author Stefano Fornari @ Funambol
 *
 * @version $Id: SyncSource.java,v 1.15 2003/12/22 11:58:41 stefano_fornari Exp $
 */
public interface SyncSource {
    
    // ---------------------------------------------------------- Public Methods
    
    /**
     * Returns the name of the source
     *
     * @return the name of the source
     */
    public String getName();
    
    /**
     * Returns the source URI
     *
     * @return the absolute URI of the source
     */
    public String getSourceURI();
    
    /**
     * Returns the type of the source (i.e. text/x-vcard)
     *
     * @return the type of the source
     */
    public String getType();

    /**
     * Returns the type info of the content handled by this source
     *
     * @return the type info of the content handled by this source
     */
    public SyncSourceInfo getInfo();
    
    /**
     * Called before any other synchronization method. To interrupt the sync
     * process, throw a SyncSourceException. 
     *
     * @param syncMode the synchronization mode being performed
     *
     * @throws SyncSourceException to interrupt the process with an error
     */
    void beginSync(int syncMode) throws SyncSourceException;
    
    /**
     * Called after the modifications have been applied and just before commit
     * the source synchronization. To interrupt the committing, throw a 
     * SyncSourceException.
     *
     * @throws SyncSourceException to interrupt the process with an error
     */
    void endSync() throws SyncSourceException;

    /**
     * @param principal the principal for which the data has to be considered. 
     *                  Null means all principals
     * @param since consider the changes since this point in time. Null means 
     *              all items regardless when they was changed.
     *
     * @return an array of keys containing the <i>SyncItem</i>'s key of the updated
     *         items after the last synchronizazion. It MUST NOT return null for
     *         no keys, but instad an empty array.
     */
    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since    )
    throws SyncSourceException ;

    /**
     * @param principal the principal for which the data has to be considered
     *                  Null means all principals
     * @param since consider the changes since this point in time. Null means 
     *              all items regardless when they was changed.
     *
     * @return an array of keys containing the <i>SyncItem</I>'s key of the updated
     *         items after the last synchronizazion. It MUST NOT return null for
     *         no keys, but instad an empty array.
     */
    public SyncItem[] getUpdatedSyncItems(Principal principal,
                                          Timestamp since    ) 
    throws SyncSourceException ;

    /**
     * Replaces an existing <i>SyncItem</i> or adds a new <i>SyncItem</i> if it 
     * does not exist. The item is also returned giving the opportunity to the
     * source to modify its content and return the updated item (i.e. updating
     * the id to the GUID).
     *
     * @param principal the entity that wants to do the operation
     * @param syncInstance  the bean to replace/add
     *
     * @return the inserted/updated item
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem setSyncItem(Principal principal, SyncItem syncInstance) 
    throws SyncSourceException ;

    /**
     * Replaces existing SyncItems or adds new SyncItems if they do not 
     * exist.
     *
     * @param principal the entity that wants to do the operation
     * @param syncItems  the beans to be replaced/added
     *
     * @return the inserted/updated 
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems) 
    throws SyncSourceException ;

    /**
     * Removes a SyncItem given its key.
     *
     * @param principal the entity that wants to do the operation
     * @param syncItem  the item to remove
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public void removeSyncItem(Principal principal, SyncItem syncItem) 
    throws SyncSourceException ;

    /**
     * Removes the given <i>SyncItem</i>s.
     *
     * @param principal the entity that wants to do the operation
     * @param syncItems  the items to remove
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems) 
    throws SyncSourceException;
    
    /**
     * @param principal the principal for which the data has to be considered
     *                  Null means all principals
     * @param since consider the changes since this point in . Null means 
     *              all items regardless when they was changed.
     *
     * @return an array of keys containing the <i>SyncItem</i>'s key of the newly 
     *         created items after the last synchronizazion. If there are no new
     *         items an empty array MUST BE returned. 
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since    ) 
    throws SyncSourceException ;

    /**
     * @param principal the principal for which the data has to be considered
     *                  Null means all principals
     * @param since consider the changes since this point in time. Null means 
     *              all items regardless when they was changed.
     *
     * @return an array of keys containing the <i>SyncItem</i>'s key of the newly 
     *         created items after the last synchronizazion. If there are no new
     *         items an empty array is returned.
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since    ) 
    throws SyncSourceException ;

    /**
     * @param principal the principal for which the data has to be considered
     *                  Null means all principals
     * @param since consider the changes since this point in time. Null means 
     *              all items regardless when they was changed.
     *
     * @return an array of keys containing the <i>SyncItem</i>'s key of the deleted
     *         items after the last synchronizazion. If there are no deleted
     *         items an empty array is returned.
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since    ) 
    throws SyncSourceException ;

    /**
     * @param principal the principal for which the data has to be considered
     *                  Null means all principals
     * @param since consider the changes since this point in time. Null means 
     *              all items regardless when they was changed.
     *
     * @return an array of keys containing the <i>SyncItem</i>'s key of the deleted
     *         items after the last synchronizazion. If there are no deleted
     *         items an empty array is returned.
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since    )
    throws SyncSourceException ;

    /**
     * @param principal the principal for which the data has to be considered
     *                  Null means all principals
     *
     * @return an array of all <i>SyncItem</i>s stored in this source. If there 
     *         are no items an empty array is returned.
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem[] getAllSyncItems(Principal principal)
    throws SyncSourceException;

    /**
     * @return an array containing the <i>SyncItem</i>s corresponding to the given
     *         array of keys. Unexisting keys must be silently ignored.
     *
     * @param principal look for data belonging to this principal, or null for 
     *        not user-related searching
     * @param syncItemKeys  the keys of the SyncItems to return
     *
     * @throws SyncSourceException in case of error (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem[] getSyncItemsFromIds(Principal principal, SyncItemKey[] syncItemKeys) throws SyncSourceException;
    
    /**
     * @return return the <i>SyncItem</i>s corresponding to the given
     *         key. If no item is found, null is returned
     *
     * @param principal look for data belonging to this principal, or null for 
     *        not user-related searching
     * @param syncItemKey  the key of the SyncItem to return
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem getSyncItemFromId(Principal principal, SyncItemKey syncItemKey) throws SyncSourceException;
    
    /**
     * @return an item from a twin item. Each source implementation is free to 
     *         interpret this as it likes (i.e.: comparing all fields).
     *
     * @param principal look for data belonging to this principal, or null for 
     *        not user-related searching
     * @param syncItem the twin item
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem) throws SyncSourceException;
    
    /**
     * @return an array of items from corresponding twin items. Each source 
     *         implementation is free to interpret this as it likes (i.e.: 
     *         comparing all fields).
     * @param principal look for data belonging to this principal, or null for 
     *        not user-related searching 
     * @param syncItems the twin items
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal, SyncItem[] syncItems) throws SyncSourceException;
}
