/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.protocol;

import java.util.List;
import java.util.ArrayList;

import sync4j.framework.core.*;
import sync4j.framework.database.Database;

import sync4j.framework.protocol.ProtocolUtil;
import sync4j.framework.protocol.v11.ClientModificationsRequirements;

/**
 * Represents a Client Modification package of the SyncML protocol.
 *
 *  The class is designed to be used in two times. First a <i>ClientModification</i>
 *  is created and checked for validity and compliancy with the protocol. Than
 *  <i>getResponse()</i> can be used to get a response message for the given
 *  request. During the request validation process some information about the
 *  request message are cached into instance variables and used in <i>getResponse()</i>.<br>
 *  Example:
 *    TO DO
 *
 * @author  Stefano Fornari @ Funambol
 * @author  Chintan
 *
 * @version $Id: ClientModifications.java,v 1.17 2003/12/15 16:57:20 luigiafassina Exp $
 */
public class ClientModifications 
extends SyncPackage 
implements Flags {
    
    // ------------------------------------------------------------ Constructors
       
    /**
      *
      *  @param syncHeader the header of the syncronization packet
      *  @param syncBody   the body of the syncronization packet
      *
      *  @throws Sync4jException
      */
    public ClientModifications(final SyncHeader syncHeader, 
                               final SyncBody   syncBody  )
    throws Sync4jException {
        super(syncHeader, syncBody);
        checkRequirements();
    }
    
	/**
	 *
	 *	@param syncHeader the header of the synchronization packet
	 *	@param syncBody   the body   of the synchronization packet
	 *  @param syncdb     the array of databases to be synchronized
	 *
	 *  @throws Sync4jException
	 */

    public ClientModifications(final      SyncHeader syncHeader,
    						   final      SyncBody   syncBody  ,
    						   Database[] syncDb               )
    throws Sync4jException{
		super(syncHeader, syncBody);
		databases = syncDb;
		checkRequirements();
	}
    
    // -------------------------------------------------------------- Properties

    /**
     * Has the server sent its capabilities and is expecting a response? 
     * If yes, <i>serverCapabilitiesCmdId</i> is set to the id of the Put command
     * sent by the server. If not, <i>serverCapabilitiesCmdId</i> is empty.
     */
    private CommandIdentifier serverCapabilitiesCmdId = null;
    
    public CommandIdentifier getServerCapabilitiesCmdId() {
        return this.serverCapabilitiesCmdId;
    }
    
    public void setServerCapabilitiesCmdId(CommandIdentifier serverCapabilitiesCmdId) {
        this.serverCapabilitiesCmdId = serverCapabilitiesCmdId;
    }
    
    /**
     * Has the server requested client capabilities? 
     * If yes, <i>clientCapabilitiesCmdId</i> is set to the id of the Get command
     * sent by the server. If not, <i>clientCapabilitiesCmdId</i> is empty.
     */
    private CommandIdentifier clientCapabilitiesCmdId = null;
    
    public CommandIdentifier getClientCapabilitiesCmdId() {
        return this.clientCapabilitiesCmdId;
    }
    
    public void setClientCapabilitiesCmdId(CommandIdentifier clientCapabilitiesCmdId) {
        this.clientCapabilitiesCmdId = clientCapabilitiesCmdId;
    }
    
    /**
     * The results command in response to the request of client capabilities
     */
    private ResultsCommand clientCapabilitiesResults = null;
    
    public ResultsCommand getClientCapabilitiesResults() {
        return this.clientCapabilitiesResults;
    }
    
    /**
     * The status command in response to the sending of server capabilities
     */
    private StatusCommand serverCapabilitiesStatus = null;
    
    public StatusCommand getServerCapabilitiesStatus() {
        return this.serverCapabilitiesStatus;
    }
    
    /**
     * The client Sync command identifier. It is used when a response is required.
     */
    private CommandIdentifier clientSyncCmdId = null;
    
    public CommandIdentifier getClientSyncCmdId() {
        return this.clientSyncCmdId;
    }
    
    public void setClientSyncCmdId(CommandIdentifier clientSyncCmdId) {
        this.clientSyncCmdId = clientSyncCmdId;
    }
    
    /**
     * The modification commands the server wants to sent to the client.
     */
    private AbstractCommand[] serverModifications = null;
    
    public AbstractCommand[] getServerModifications() {
        return this.serverModifications;
    }
    
    public void setServerModifications(AbstractCommand[] serverModifications) {
        this.serverModifications = serverModifications;
    }
    
    /**
     * The status to be returned for the client sync command.
     */
    private StatusCommand[] clientModificationsStatus = null;
    
    public StatusCommand[] getClientModificationsStatus() {
        return this.clientModificationsStatus;
    }
    
    public void setClientModificationsStatus(StatusCommand[] clientModificationsStatus) {
        this.clientModificationsStatus = clientModificationsStatus;
    }
    
    /**
     * Caches the commands sent by the client. It is set during the
     * checking of the requirements.
     */
    private AbstractCommand[] clientCommands = null;

    public AbstractCommand[] getClientCommands() {
        return clientCommands;
    }
        
    /**
     * Caches the SyncCommand sent by the client. It is set during the checking
     * of requirements.
     */
    private SyncCommand[] clientSyncCommands = null;
    
    public SyncCommand[] getClientSyncCommands() {
        return this.clientSyncCommands;
    }
   
    /**
     * Databases that the server wants to synchronize. 
     */
    private Database[] databases = null;

    public void setDatabases(Database[] databases) {
        this.databases = databases;
    }

    public Database[] getDatabases() {
        return this.databases;
    }
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Checks that all requirements regarding the header of the initialization
     * packet are respected.
     *
     * @throws ProtocolException
     */
    public void checkHeaderRequirements() throws ProtocolException {
        ClientModificationsRequirements.checkDTDVersion     (syncHeader.getDTDVersion()           );
        ClientModificationsRequirements.checkProtocolVersion(syncHeader.getProtocolVersion()      );
        ClientModificationsRequirements.checkSessionId      (syncHeader.getSyncSessionIdentifier());
        ClientModificationsRequirements.checkMessageId      (syncHeader.getMessageID()            );
        ClientModificationsRequirements.checkTarget         (syncHeader.getTarget()               );
        ClientModificationsRequirements.checkSource         (syncHeader.getSource()               );
    }
    
    /**
     * Checks that all requirements regarding the body of the initialization
     * packet are respected.
     *
     * NOTE: bullet 2 pag 34 is not clear. Ignored for now.
     *
     * @throws ProtocolException
     */
    public void checkBodyRequirements() throws ProtocolException {
        clientCommands = syncBody.getCommands(); // NOTE: initializes the clientCommands property
        
        //
        // If the server sent the device information to the client and requested
        // a response, serverCapabilitiesCmdId contains the command id of the
        // request command. A Status command with the same cmd id reference 
        // must exist.
        //
        checkServerCapabilitiesStatus();
        
        //
        // If the server requested the device capabilities of the client, 
        // clientCapabilitiesCmdId contains the command id of the Get command. 
        // A Results command with the same cmd id reference must exist.
        //
        checkClientCapabilitiesResult();

        //
        // The Sync command must exists
        //
        checkSyncCommand();
    }
    
    // ----------------------------------------------------------- getResponse()
    
    /**
     * Constructs a proper response message.<p>
     * The sync package to the client has the following purposes:
     * <ul>
     *  <li>To inform the client about the results of sync analysis.
     *  <li>To inform about all data modifications, which have happened in the 
     *      server since the previous time when the server has sent the 
     *      modifications to the client.
     * </ul>
     *
     * @return the response message
     *
     * @throws ProtocolException in case of error or inconsistency
     */
    public Message getResponse() throws ProtocolException {
        ArrayList commandList = new ArrayList();
        
        if (idGenerator == null) {
            throw new NullPointerException("The id generator is null. Please set a value for idGenerator");
        }
        
        //
        // Constructs all required response commands.
        //
        // NOTE: if NoResp is specified in the header element, than no 
        //       response commands must be returned regardless NoResp is 
        //       specified or not in subsequent commands
        //
        if (syncHeader.getNoResponse() == false) {

            TargetRef[] targetRefs = new TargetRef[] { new TargetRef(syncHeader.getTarget().getURI()) };
            SourceRef[] sourceRefs = new SourceRef[] { new SourceRef(syncHeader.getSource().getURI()) };

            StatusCommand statusCommand = new StatusCommand(
                                              idGenerator.next()               ,
                                              syncHeader.getMessageID()        ,
                                              "0" /* command ref */            ,
                                              "SyncHdr" /* see SyncML specs */ ,
                                              targetRefs                       ,
                                              sourceRefs                       ,
                                              null /* credential */            ,
                                              null /* challenge */             ,
                                              new Data(StatusCode.OK)          , 
                                              new Item[0]
                                          );

            commandList.add(statusCommand);
        
            //
            // 2. The Status element MUST be included in SyncBody if requested by 
            //    the client. It is now used to indicate the general status of 
            //    the sync analysis and the status information related to data 
            //    items sent by the client (e.g., a conflict has happened.). 
            //
            for (int i=0; (  isFlag(FLAG_SYNC_STATUS_REQUIRED)
                          && (clientModificationsStatus != null) 
                          && (i<clientModificationsStatus.length) ); ++i) {
                commandList.add(clientModificationsStatus[i]);
            }
        }  
        
        //
        // 3. The Sync element MUST be included in SyncBody, if earlier there
        // were no occurred errors, which could prevent the server to process 
        // the sync analysis and to send its modifications back to the client.
        //
        for (int i=0; ((serverModifications != null) && (i<serverModifications.length)); ++i) {
            commandList.add(serverModifications[i]);
        }
        
        //
        // Constructs return message
        //
        Target target = new Target(syncHeader.getSource().getURI(),
                                   syncHeader.getSource().getLocationName());
        Source source = new Source(syncHeader.getTarget().getURI(),
                                   syncHeader.getTarget().getLocationName());
        SyncHeader responseHeader = new SyncHeader (
                                        getDTDVersion()                      ,
                                        getProtocolVersion()                 ,
                                        syncHeader.getSyncSessionIdentifier()                     ,
                                        syncHeader.getMessageID()                                 ,
                                        target                                                    ,
                                        source                                                    ,
                                        null  /* response URI */                                  ,
                                        false                                                     ,
                                        null /* credentials */                                    ,
                                        null /* meta data */                                 
                                    );
        

		AbstractCommand[] commands 	= 	null;
        int size = commandList.size();
        if (size == 0) {
			 commands = new AbstractCommand[1];
		} else {
			commands = new AbstractCommand[size];
		}
        for (int i=0; i < size; i++) {
            commands[i] = (AbstractCommand)commandList.get(i);
        }

        SyncBody responseBody = new SyncBody(
                                    commands,
                                    isFlag(FLAG_FINAL_MESSAGE) /* final */
                                );

        try {
            return new Message(responseHeader, responseBody);
        } catch (RepresentationException e) {
            //
            // It should never happen !!!!
            //
            throw new ProtocolException("Unexpected error", e);
        }
    }
    
    /**
     * Create a Status command for the Sync sent by the client.
     *
     * <b>NOTE</b>: the protocol does not specify any information about the format
     * and the content of this message. By now a dummy status command is created
     * and returned.
     *
     * @return a StatusCommand object
     */
    public StatusCommand createSyncStatusCommand() {
        return new StatusCommand( idGenerator.next()                       ,
                                  "0" /* message id; TO DO */              , 
                                  clientSyncCmdId.getValue()               ,
                                  SyncCommand.COMMAND_NAME                 ,
                                  (TargetRef[])null /* target refs */      ,
                                  (SourceRef[])null /* source refs */      ,
                                  null /* credential */                    ,
                                  null /* chal */                          ,
                                  null /* Data */                          ,
                                  null /* items */                         );
                                  
    }
    
    /**
     * For the Sync element, there are the following requirements.
     * <ul>
     *   <li> CmdID is required.
     *   <li> The response can be required for the Sync command. (See the Caching of Map Item,
     *        Chapter 2.3.1)
     *   <li> Target is used to specify the target database.
     *   <li> Source is used to specify the source database.
     * </ul>
     *
     * 5. If there is any modification in the server after the previous sync, 
     * there are following requirements for the operational elements (e.g., 
     * Replace, Delete, and Add 4 ) within the Sync element.
     * <ul>
     *   <li> CmdID is required.
     *   <li> The response can be required for these operations.
     *   <li> Source MUST be used to define the temporary GUID (See Definitions) 
     *        of the data item in the server if the operation is an addition. 
     *        If the operation is not an addition, Source MUST NOT be included.
     *   <li> Target MUST be used to define the LUID (See Definitions) of the 
     *        data item if the operation is not an addition. If the operation is
     *        an addition, Target MUST NOT be included.
     *   <li> The Data element inside Item is used to include the data itself if
     *        the operation is not a seletion.
     *   <li> The Type element of the MetaInf DTD MUST be included in the Meta 
     *        element to indicate the type of the data item (E.g., MIME type). 
     *        The Meta element inside an operation or inside an item can be used.
     * </ul>
     *
     * @param db the database to be synchronized
     *
     * @return a Sync command
     *
     * @throws ProtocolException
     */
    public SyncCommand createSyncCommand(Database db) 
    throws ProtocolException {
        CommandIdentifier syncId = idGenerator.next();
        
        AbstractCommand[] commands = null;

		// if db.getMethod is One_Way_Sync_CLIENT
		// no synccommand from Server to Client

		if(db.getMethod() != AlertCode.ONE_WAY_FROM_CLIENT){
			commands = prepareCommands(db);
		}
        
        return new SyncCommand(
            syncId                             ,
            isFlag(FLAG_SYNC_RESPONSE_REQUIRED),
            null                               , /* credentials */
            db.getTarget()                     ,
            db.getSource()                     ,
            null                               , /* Meta        */
            commands                           
        );
            
    }
    
    /**
     * Returns an array of synchronization commands (Add, Copy, Delete, Exec, 
     * Replace) based on the content of the given database.
     *
     * @param db the database to be synchronized
     *
     * @return an array of AbstractCommand
     */
    public AbstractCommand[] prepareCommands(Database db) {
        ArrayList commands = new ArrayList();
        
        Meta meta = new Meta(new StringMetaContent(db.getType(), "Type"));
        
        Item[] items = null;  // reused many times
        
        //
        // Add
        //
        items = db.getAddItems();
        if (items != null) {
            commands.add(
                new AddCommand(
                    idGenerator.next()                          ,
                    isFlag(FLAG_MODIFICATIONS_RESPONSE_REQUIRED),
                    null /* credentials */                      ,
                    meta                                        ,
                    items                                       )
            );
        }
        
        //
        // Copy
        //
        items = db.getCopyItems();
        if (items != null) {
            commands.add(
                new CopyCommand(
                    idGenerator.next()                          ,
                    isFlag(FLAG_MODIFICATIONS_RESPONSE_REQUIRED),
                    null /* credentials */                      ,
                    meta                                        ,
                    items                                       )
            );
        }
        
        //
        // Delete
        //
        items = db.getDeleteItems();
        if (items != null) {
            commands.add(
                new DeleteCommand(
                    idGenerator.next()                          ,
                    isFlag(FLAG_MODIFICATIONS_RESPONSE_REQUIRED),
                    isFlag(FLAG_ARCHIVE_DATA)                   ,
                    isFlag(FLAG_SOFT_DELETE)                    ,
                    null /* credentials */                      ,
                    meta                                        ,
                    items                                       )
            );
        }
        
        //
        // Exec
        //
        items = db.getExecItems();
        
        for (int i=0; ((items != null) && (i<items.length)); ++i) {
            commands.add(
                new ExecCommand(
                    idGenerator.next()                          ,
                    isFlag(FLAG_MODIFICATIONS_RESPONSE_REQUIRED),
                    null /* credentials */                      ,
                    items[i]                                    )
            );
        }
        
        //
        // Replace
        //
        items = db.getReplaceItems();
        if (items != null) {
            commands.add(
                new ReplaceCommand(
                    idGenerator.next()                          ,
                    isFlag(FLAG_MODIFICATIONS_RESPONSE_REQUIRED),
                    null /* credentials */                      ,
                    meta                                        ,
                    items                                       )
            );
        }

        int size = commands.size();
        AbstractCommand [] aCommands = new AbstractCommand[size];
        for (int i=0; i < size; i++) {
            aCommands[i] = (AbstractCommand)commands.get(i);
        }
        return aCommands;
    }
    
    // --------------------------------------------------------- Private methods
    
    /**
     * Checks if the requested status for server capabilities has been specified.
     * <p>
     *
     * @throws ProtocolException
     */
    private void checkServerCapabilitiesStatus() 
    throws ProtocolException {
        //
        // If serverCapabilitiesCmdId is null no serverCapabilities status is required
        //
        if (serverCapabilitiesCmdId == null) return;
        
        List list = ProtocolUtil.filterCommands(clientCommands         ,
                                                StatusCommand.class    ,
                                                serverCapabilitiesCmdId);
        
        if (list.size() == 0) {
            Object[] args = new Object[] { serverCapabilitiesCmdId.getValue() };
            throw new ProtocolException(ClientModificationsRequirements.ERRMSG_MISSING_STATUS_COMMAND, args);
        }
        
        serverCapabilitiesStatus = (StatusCommand)list.get(0);
    }
    
    /**
     * Checks if the result for client capabilities has been given.
     * <p>
     *
     * @throws ProtocolException
     */
    private void checkClientCapabilitiesResult() 
    throws ProtocolException {
        //
        // If clientCapabilitiesCmdId is null no client capabilities were required
        //
        if (clientCapabilitiesCmdId == null) return;
        
        List list = ProtocolUtil.filterCommands(clientCommands         ,
                                                ResultsCommand.class   ,
                                                clientCapabilitiesCmdId);
        
        if (list.size() == 0) {
            Object[] args = new Object[] { clientCapabilitiesCmdId.getValue() };
            throw new ProtocolException(ClientModificationsRequirements.ERRMSG_MISSING_RESULTS_COMMAND, args);
        }
        
        ResultsCommand results = (ResultsCommand)list.get(0);
        
        ClientModificationsRequirements.checkCapabilities(
            results, 
            ClientModificationsRequirements.CLIENT_CAPABILITIES
        );
        
        clientCapabilitiesResults = results;
    }
    
    /**
     * Checks the Sync command.
     * <p>Filters out the Sync Messages from client to Server with
     * ONE_WAY_SYNC_SERVER
     *
     * @throws ProtocolException
     */
    private void checkSyncCommand() 
    throws ProtocolException {
        List list = ProtocolUtil.filterCommands(clientCommands   ,
                                                SyncCommand.class);
        
        if (list.size() == 0) {
            clientSyncCommands = new SyncCommand[0];
            return;
        }

        clientSyncCommands = (SyncCommand[])list.toArray(new SyncCommand[list.size()]);
    }
    
}
