/*

 Copyright (c) 2001 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*/

package sync4j.framework.protocol;

import java.util.ArrayList;

import sync4j.framework.core.*;
import sync4j.framework.database.Database;
import sync4j.framework.protocol.v11.*;
import sync4j.framework.core.xml.Namespace;

/**
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: ProtocolUtil.java,v 1.5 2003/12/15 16:57:20 luigiafassina Exp $
 *
 */
public class ProtocolUtil {

    /** Filters the given commands based on the given class
     *
     * @param commands commands to be filtered
     * @param filterClass selector
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterCommands(AbstractCommand[] commands, Class filterClass) {
        ArrayList filteredCommands = new ArrayList();

        for (int i=0; i<commands.length; ++i) {
            if (filterClass.isInstance(commands[i])) {
                filteredCommands.add(commands[i]);
            }
        }

        return filteredCommands;
    }

    /**
     * Filters the given commands based on the given command id. A command is
     * selected if it a ResponseCommand (only ResponseCommand subclasses have
     * got the reference command id) and its command id ref matches <i>cmdId</i>.
     *
     * @param commands commands to be filtered
     * @param cmdId selector
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterCommands(AbstractCommand[] commands,
                                      CommandIdentifier cmdId   ) {
        ArrayList filteredCommands = new ArrayList();

        for (int i=0; i<commands.length; ++i) {
            if (  (commands[i] instanceof ResponseCommand)
               && (((ResponseCommand)commands[1]).getCommandIdentifier().equals(cmdId))) {
                filteredCommands.add(commands[i]);
            }
        }

        return filteredCommands;
    }

    /**
     * Filters the given commands based on the given command type and command id.
     * It combines <i>filterCommands(commands, filterClass)</i> and
     * <i>filterCommands(commands, cmdId)</i> returning only the commands
     * that respect both requirements.
     *
     * @param commands commands to be filtered
     * @param filterClass class type selector
     * @param cmdId selector
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterCommands(AbstractCommand[] commands   ,
                                      Class             filterClass,
                                      CommandIdentifier cmdId      ) {
        //
        // Since filtering on command identifier seems more selective,
        // filterCommands(commands, cmdId) is called first and than
        // filterCommands(..., filterClass) is called with the returned values.
        //
        ArrayList list = filterCommands(commands, cmdId);
        int size = list.size();
        AbstractCommand [] aCommands = new AbstractCommand[size];
        for (int i=0; i < size; i++) {
            aCommands[i] = (AbstractCommand)list.get(i);
        }
        return filterCommands(
                    /* not compatible with j2me
                    (AbstractCommand[])list.toArray(new AbstractCommand[0]),
                    */
                    aCommands,
                   filterClass
               );
    }

    /**
     * Extracts a DeviceInfo object from a PutCommand command.
     * <p>
     * <b>IMPORTANT</b>: <i>put</i> is not checked for validity, the caller must
     * assure it is a valid put command containig device capabilities.
     *
     * @param put the PutCommand object
     *
     * @return a DeviceInfo object containing the device information extracted
     *         from the given command.
     *
     * @throws ProtocolException
     */
    public static DeviceInfo getDeviceInfo(PutCommand put)
    throws ProtocolException {
        try {
            return new DeviceInfo(put.getItems()[0].getData().getValue());
        } catch (Exception e) {
            throw new ProtocolException(e.getMessage());
        }
    }

    /**
     * Returns the SyncML xml tag for a "type" element (e.g
     * <Type xmlns='syncml:metinf'>application/vnd.syncml-devinf+xml</Type>).
     *
     * @param type  the type value
     * @param ns the type namespace
     *
     * @return the xml form of the type element
     *
     */
    public static String getTypeElement(Namespace ns, String type) {
        return "<Type xmlns='"
             + ns.getURI()
             + "'>"
             + type
             + "</Type>";
    }

    /**
     * Creates and returns and AlertCommand for the synchronization of the
     * given database.
     *
     * @param id the command id - NULL
     * @param noResponse
     * @param credential - NULL
     * @param db the database to be synchronized - NOT NULL
     *
     * @return the AlertCommand object
     */
    public static AlertCommand createAlertCommand(CommandIdentifier id        ,
                                                  boolean           noResponse ,
                                                  Credential        credential,
                                                  Database          db        ) {
        Item[] items = new Item[1];

        SyncAnchor serverAnchor = db.getServerAnchor();
        AnchorMetaContent anchor = new AnchorMetaContent(
                                       serverAnchor.getLast(),
                                       serverAnchor.getNext()
                                   );
        items[0] = new Item(db.getTarget() /* target */ ,
                            db.getSource()              ,
                            new Meta(anchor)            ,
                            null           /* data */   );

        return new AlertCommand(
                   id            ,
                   noResponse     ,
                   credential    ,
                   db.getMethod(),
                   items
               );
    }

    /**
     * Translates a Target object to a Source object
     *
     * @param target the target object - NULL
     *
     * @return a Source object with the same URI and local name of <i>target</i>
     */
    public static Source target2Source(Target target) {
        if (target == null) return null;

        return new Source(target.getURI(), target.getLocationName());
    }

    /**
     * Translates a Source object to a Target object
     *
     * @param source the source object - NULL
     *
     * @return a Target object with the same URI and local name of <i>source</i>
     */
    public static Target source2Target(Source source) {
        if (source == null) return null;

        return new Target(source.getURI(), source.getLocationName());
    }
    
    /** 
     * Extracts the target and source refs from an array of items
     *
     * @param items the items to inspect. If null targetRefs and sourceRefs
     *              remain unchanged
     * @param targetRefs a reference to an array that will contain the references
     *                   to the items' targets
     * @param sourceRefs a reference to an array that will contain the references
     *                   to the items' sources
     *
     */
    public static void extractRefs(Item[] items          , 
                                   TargetRef[] targetRefs,
                                   SourceRef[] sourceRefs) {
        if (items == null) {
            return;
        }
        
        Target t = null;
        Source s = null;
        for (int i=0; i<items.length; ++i) {
            t = items[i].getTarget();
            s = items[i].getSource();
            
            targetRefs[i] = (t != null) ? new TargetRef(t) : null;
            sourceRefs[i] = (s != null) ? new SourceRef(s) : null;
        }
    }
    
}
