/*

 $Id: SimpleIdGenerator.java,v 1.4 2002/12/18 21:25:37 stefano_fornari Exp $

 Copyright (c) 2001 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.protocol;

import sync4j.framework.protocol.IdGenerator;

/**
 * This class works as a simple id generator that after being initialized with a
 * numeric value, increments the counter each time the next method is called.
 * <p>
 * SimpleIdGenerator is thread-safe
 *
 * @author  Stefano Fornari @ funambol
 */
public final class SimpleIdGenerator
        implements IdGenerator, java.io.Serializable {

  // -------------------------------------------------------------- Properties

  /**
   * The counter
   */
  private long counter = 0;

  public long getCounter()
  {
    return counter;
  }

  /**
   * Standard setter method
   */
  public void setCounter(long counter)
  {
    this.counter = counter;
  }

  /**
   * The units the counter must be incremented each time next is called
   */
  private int increment = 1;

  public int getIncrement()
  {
    return increment;
  }

  /**
   * Standard setter method
   */
  public void setIncrement(int increment)
  {
    this.increment = increment;
  }

  // ------------------------------------------------------------ Constructors

  /**
   * Creates a new instance of SimpleIdGenerator
   *
   * @param counter the starting value
   * @param increment the increment
   */
  public SimpleIdGenerator(long counter, int increment)
  {
    this.counter = counter;
    this.increment = increment;
  }

  /**
   * Like this(counter,1)
   *
   * @param counter the starting value
   */
  public SimpleIdGenerator(int counter)
  {
    this(counter, 1);
  }

  /**
   * Reset the generator to 0.
   */
  public void reset()
  {
    this.counter = 0;
  }

  /**
   * Like this(0,1)
   */
  public SimpleIdGenerator()
  {
    this(0, 1);
  }

  /**
   * Returns the next value of the counter (incrementing the counter by the
   * increment)
   *
   * @return the next generated value
   */
  public synchronized String next()
  {
    counter += increment;

    return String.valueOf(counter);
  }

  /**
   * Returns the last generated id (which is the current id).
   *
   * @return the last generated id
   */
  public synchronized String current()
  {
    return String.valueOf(counter);
  }
}
