/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.protocol;

import java.util.ArrayList;
import java.util.Hashtable;
import java.security.Principal;

import sync4j.framework.core.*;
import sync4j.framework.database.Database;

import sync4j.framework.protocol.SyncPackage;
import sync4j.framework.protocol.ProtocolUtil;
import sync4j.framework.protocol.v11.InitializationRequirements;

import sync4j.framework.security.Sync4jPrincipal;

/**
 * Represents the Initialization package of the SyncML protocol.
 * 
 * Example:
 * <pre>
 *   SyncInitialization syncInit = new SyncInitialization(header, body);
 *   ... do something ...
 *   syncInit.setServerCapabilities(serverCapabilities);
 *   syncInit.setAuthorizedStatusCode(StatusCode.AUTHENTICATION_ACCEPTED);
 *   syncInit.setClientCapabilitiesStatusCode(StatusCode.OK);
 *   ... other initializations ...
 *   Message response = syncInit.getResponse();
 * </pre>
 *
 * @author Stefano Fornari @ Funambol
 *
 * @version $Id: SyncInitialization.java,v 1.27 2003/12/30 16:29:28 luigiafassina Exp $
 *
 * @see SyncPackage
 */
public class SyncInitialization
extends SyncPackage
{
    /**
     * Contains the request for server capabilities sent by the client
     * (null means capabilities not required)
     */
    private GetCommand serverCapabilitiesRequest = null;

    /**
     * The device capabilities sent by the client.
     */
    private PutCommand clientCapabilities = null;

    public DeviceInfo getClientDeviceInfo() throws ProtocolException {
        return ProtocolUtil.getDeviceInfo(this.clientCapabilities);
    }

    /**
     * Caches the commands sent by the client.It is set during the
     * checking of the requirements.
     */
    private AbstractCommand[] clientCommands = null;

    public AbstractCommand[] getClientCommands() {
        return clientCommands;
    }

    /**
     * Caches the alert command sent by the client. It is set during the
     * checking of the requirements.
     */
    private AlertCommand[] clientAlerts = null;

    public AlertCommand[] getClientAlerts() {
        return clientAlerts;
    }

    /**
     * Client Capabilities status code
     */
    private int clientCapabilitiesStatusCode = -1;

    public int getClientCapabilitiesStatusCode() {
        return this.clientCapabilitiesStatusCode;
    }

    public void setClientCapabilitiesStatusCode(int clientCapabilitiesStatusCode) {
        this.clientCapabilitiesStatusCode = clientCapabilitiesStatusCode;
    }

    /**
     * Server capabilities
     */
    private DeviceInfo serverCapabilities = null;

    public void setServerCapabilities(DeviceInfo capabilities) {
        this.serverCapabilities = capabilities;
    }

    public DeviceInfo getServerCapabilities() {
        return this.serverCapabilities;
    }

    /**
     * Databases that the server wants to synchronized. They can be differnt
     * from the databases the client has requested to be synchronized.
     */
    private Database[] databases = null;

    public void setDatabases(Database[] databases) {
        this.databases = databases;
    }

    public Database[] getDatabases() {
        return this.databases;
    }

    /**
     * Does the server require client capabilities?
     */
    private boolean clientCapabilitiesRequired = false;

    public void setClientCapabilitiesRequired(boolean clientCapabilitiesRequired) {
        this.clientCapabilitiesRequired = clientCapabilitiesRequired;
    }

    public boolean isClientCapabilitiesRequired() {
        return this.clientCapabilitiesRequired;
    }

    /**
     * Authorized status code - used in building response message
     */
    private int authorizedStatusCode = -1;

    public void setAuthorizedStatusCode(int authorizedStatusCode) {
        this.authorizedStatusCode = authorizedStatusCode;
    }

    // ---------------------------------------------------------- Command status
    
    /**
     * The map containing the status of the commands
     */
    private Hashtable commandStatus = new Hashtable();
    
    /**
     * Sets the status code for the given command.
     *
     * @param cmd the command
     * @param statusCode the status code
     */
    public void setStatusCodeForCommand(AbstractCommand cmd, int statusCode) {
        setStatusCodeForCommand(cmd.getCommandIdentifier().getValue(), statusCode);
    }
    
    /**
     * Sets the status code for the command identified by the given id.
     *
     * @param cmdId the command id
     * @param statusCode the status code
     */
    public void setStatusCodeForCommand(String cmdId, int statusCode) {
        commandStatus.put(cmdId, new Integer(statusCode));
    }
    
    /**
     * Returns the status code for the given command. The status code must be
     * previously set with <i>setStatusCodeForCommand()</i>. If no status code
     * is associated to the given command, the default status code is returned.
     *
     * @param cmd the command
     * @param defaultCode the default status code
     *
     * @return the status code for the command if previously set, the default
     *         status code otherwise
     */
    public int getStatusCodeForCommand(AbstractCommand cmd, int defaultCode) {
        String cmdId = cmd.getCommandIdentifier().getValue();
        
        return getStatusCodeForCommand(cmdId, defaultCode);

    }
    
    /**
     * The same as <i>getStatusCodeForCommand(AbstractCommand, int) but passing
     * in the command id instead of the command.
     *
     * @param cmdId the command id
     * @param defaultCode
     *
     * @return the status code for the command if previously set, the default
     *         status code otherwise
     */
    public int getStatusCodeForCommand(String cmdId, int defaultCode) {
        Integer statusCode = (Integer)commandStatus.get(cmdId);
        
        return (statusCode == null) ? defaultCode : statusCode.intValue();
    }
 

    // ------------------------------------------------------------ Constructors

    /**
      *
      *  @param syncHeader the header of the syncronization packet
      *  @param syncBody   the body of the syncronization packet
      *
      *  @throws ProtocolException
      */
    public SyncInitialization(final SyncHeader syncHeader, 
                              final SyncBody   syncBody  )
    throws ProtocolException
    {
        super(syncHeader, syncBody);
        checkRequirements();
    }

    // ------------------------------------------------------ Public methods

    /**
     * Alerts specifying the database to be synchronized could be more
     * then one. Each can contains more than one item, which specifies
     * a single database. This method selects the items containing the
     * databases regardless in what alert command they where included.
     *
     * @return an array of Database objects 
     */
    public Database[] getDatabasesToBeSynchronized() {
        ArrayList dbList = new ArrayList();
        
        Credential c = null;

        Database    db           = null;
        Item[]      items        = null;
        MetaContent metaContent  = null;
        for (int i=0; ((clientAlerts != null) && (i < clientAlerts.length)); ++i) {
            //
            // Only database synchronization alerts are selected
            //
            if (!AlertCode.isInitializationCode(clientAlerts[i].getAlertCode())) {
                continue;
            }

            items = clientAlerts[i].getItems();
            for (int j=0; ((items != null) && (j<items.length)); ++j) {
                metaContent = items[j].getMeta().getValue();

                //
                // If the anchor does not exists, the alert does not represent
                // a database to be synchronized.
                //
                if (!(metaContent instanceof AnchorMetaContent)) {
                    continue;
                }
                
                c = syncHeader.getCredential();
                if (c == null) {
                    c = Credential.getGuestCredential();
                }
                Principal p = Sync4jPrincipal.fromCredential (
                                  c.getData(), 
                                  c.getType(),
                                  syncHeader.getSource().getURI());

                //
                // NOTE: the target becomes the database source and vice-versa
                //
                db = new Database( 
                         items[j].getTarget().getURI()                   ,
                         null /* type */                                 ,
                         ProtocolUtil.source2Target(items[j].getSource()),
                         ProtocolUtil.target2Source(items[j].getTarget()),
                         (SyncAnchor)metaContent.getContentObject()      ,
                         p
                     );
                db.setMethod(clientAlerts[i].getAlertCode());
                db.setAlertCommand(clientAlerts[i]);
                
                dbList.add(db);
            }  // next j
        }  // next i

        int dbSize = dbList.size();
        Database[] dbArray = new Database[dbSize];
        for (int i=0; i<dbSize; i++) {
            dbArray[i] = (Database)dbList.get(i);
        }
        return dbArray;
    }

    // -------------------------------------------------------------------------
    
    /**
     * Checks that all requirements regarding the header of the initialization 
     * packet are respected.
     *
     * @throws ProtocolException
     */
    public void checkHeaderRequirements() 
    throws ProtocolException {
        InitializationRequirements.checkDTDVersion     (syncHeader.getDTDVersion()           );
        InitializationRequirements.checkProtocolVersion(syncHeader.getProtocolVersion()      );
        InitializationRequirements.checkSessionId      (syncHeader.getSyncSessionIdentifier());
        InitializationRequirements.checkMessageId      (syncHeader.getMessageID()            );
        InitializationRequirements.checkTarget         (syncHeader.getTarget()               );
        InitializationRequirements.checkSource         (syncHeader.getSource()               );
    }

    /**
     * Checks that all requirements regarding the body of the initialization 
     * packet are respected.
     *
     * @throws ProtocolException
     */
    public void checkBodyRequirements() 
    throws ProtocolException {
        ArrayList listAlerts = new ArrayList();
        ArrayList mergedClientCommands = new ArrayList();
        
        AbstractCommand[] allClientCommands = syncBody.getCommands();
        
        //
        // Extracts and checks alert commands
        //
        ArrayList alertList = ProtocolUtil.filterCommands(allClientCommands    , 
                                                          AlertCommand.class);
        int size = alertList.size();
        AlertCommand[] alerts = new AlertCommand[size];
        for (int i=0; i < size; i++) {
            alerts[i] = (AlertCommand)alertList.get(i); 

            listAlerts.add(alerts[i]);
         }

        for (int i=0; (alerts != null) && (i < alerts.length); ++i) {
             InitializationRequirements.checkAlertCommand(alerts[i]);
        }  // next i
        
        //
        // All alerts are OK => they can be cached
        //
        clientAlerts = (AlertCommand[])listAlerts.toArray(new AlertCommand[0]);
        mergedClientCommands.addAll(listAlerts);
        
        ArrayList clientCapabilitiesList = 
            ProtocolUtil.filterCommands(allClientCommands, PutCommand.class);

        if ((clientCapabilities != null) && (clientCapabilitiesList.size()>0)) {
            InitializationRequirements.checkCapabilities((PutCommand)clientCapabilitiesList.get(0)     , 
                                                         InitializationRequirements.CLIENT_CAPABILITIES);
            clientCapabilities = (PutCommand)clientCapabilitiesList.get(0);
        }
        mergedClientCommands.addAll(clientCapabilitiesList);

        ArrayList capabilitiesRequest = 
            ProtocolUtil.filterCommands(allClientCommands, GetCommand.class);

        if ((capabilitiesRequest != null) && (capabilitiesRequest.size()>0)) {
            InitializationRequirements.checkCapabilitiesRequest((GetCommand)capabilitiesRequest.get(0));
            serverCapabilitiesRequest = (GetCommand)capabilitiesRequest.get(0);        
        }
        mergedClientCommands.addAll(capabilitiesRequest);
        
        clientCommands = 
            (AbstractCommand[])mergedClientCommands.toArray(new AbstractCommand[0]);
        
    }


    /**
     * Constructs a proper response message.<br>
     * NOTES
     * <ul>
     *  <li> If server capabilities are not required, they are not sent (in
     *       the SyncML protocol the server MAY send not required capabilities)
     * </ul>
     *
     * @return the response message
     *
     * @throws ProtocolException in case of error or inconsistency
     */
    public Message getResponse() throws ProtocolException {
        ArrayList commandList = new ArrayList();

        //
        // Constructs all required response commands.
        //
        // NOTE: if NoResp is specified in the header element, than no 
        //       response commands must be returned regardless NoResp is 
        //       specified or not in subsequent commands
        //
        
        if (syncHeader.getNoResponse() == false) {
            //
            // Session authorization
            //
            TargetRef[] targetRefs = new TargetRef[] { new TargetRef(syncHeader.getTarget().getURI()) };
            SourceRef[] sourceRefs = new SourceRef[] { new SourceRef(syncHeader.getSource().getURI()) };

            //
            // If the session is not authenticated, a Chal element must be returned
            //
            AuthenticationChallenge chal = null;
            if (authorizedStatusCode != StatusCode.AUTHENTICATION_ACCEPTED) {
                chal = AuthenticationChallenge.getBasicAuthenticationChallenge();
            }
            StatusCommand statusCommand = new StatusCommand(
                                              idGenerator.next()                            ,
                                              syncHeader.getMessageID()                     ,
                                              "0" /* command ref */                         ,
                                              "SyncHdr" /* see SyncML specs */              ,
                                              targetRefs                                    ,
                                              sourceRefs                                    ,
                                              null /* credential */                         ,
                                              chal                                          ,
                                              new Data(String.valueOf(authorizedStatusCode)),
                                              new Item[0]
                                          );

            commandList.add(statusCommand);

            //
            // Status for each command that requested it (it is supposed each 
            // command has bean already checked).
            //
            for (int i=0; ((clientCommands != null) && (i < clientCommands.length)); ++i) {
                if (clientCommands[i].getNoResponse()) continue;

                targetRefs = new TargetRef[0];
                sourceRefs = new SourceRef[0];
                if (clientCommands[i] instanceof ItemizedCommand) {
                    Item[] items = ((ItemizedCommand)clientCommands[i]).getItems();
                    
                    targetRefs = new TargetRef[items.length];
                    sourceRefs = new SourceRef[items.length];
                    ProtocolUtil.extractRefs(items     ,
                                             targetRefs,
                                             sourceRefs);
                }
                
                String commandReference = clientCommands[i].getCommandIdentifier().getValue();
                int status = getStatusCodeForCommand(clientCommands[i], StatusCode.OK);

                Item[] items = new Item[0];
                
				///
				//	Within Response of Alert, Item must contain the NEXT Anchor.
                // NOTE: a database represents the server point of view so that
                //       the target is the client database and the source the
                //       server database.
				//
                if (clientCommands[i] instanceof AlertCommand) {
                    for(int j=0; (databases != null) && (j<databases.length) ; ++j)
                    {
                        if((databases[j].getSource().getURI()).equals(targetRefs[0].getValue())){
                            items = new Item[1];
                            
                            AnchorMetaContent alertAnchor = new AnchorMetaContent(
                                null,                  // last
                                databases[j].getNext() // next
                            );
                            items[0] = new Item(
                               null,  // target
                               null,  // source
                               null , // meta
                               new Data(alertAnchor.toXML())
                            );
                           
                            break;
                        }
                    }
                }

                statusCommand = new StatusCommand(
                                    idGenerator.next()          ,
                                    syncHeader.getMessageID()   ,
                                    commandReference            ,
                                    clientCommands[i].getName() ,
                                    targetRefs                  ,
                                    sourceRefs                  ,
                                    null /* credential */       ,
                                    null /* challenge */        ,
                                    new Data(status)            ,
                                    items
                                );

                commandList.add(statusCommand);
            }  // next i

            //
            // Status for client capabilities
            //
            if (clientCapabilities != null) {
                String commandReference = clientCapabilities.getCommandIdentifier().getValue();
                targetRefs = new TargetRef[] {};
                sourceRefs = new SourceRef[] { 
                    new SourceRef(InitializationRequirements.CAPABILITIES_SOURCE) 
                };

                Data data = new Data(String.valueOf(clientCapabilitiesStatusCode));
                statusCommand = new StatusCommand(
                                    idGenerator.next()          ,
                                    syncHeader.getMessageID()   ,
                                    commandReference            ,
                                    clientCapabilities.getName(), 
                                    targetRefs                  ,
                                    sourceRefs                  ,
                                    null /* credential */       ,
                                    null /* challenge */        ,
                                    data                        ,
                                    null /* items */
                                );

                commandList.add(statusCommand);
            }
        }  // end if syncHeader.getNoResponse() == false

        //
        // Server capabilities
        //
        if ((serverCapabilitiesRequest != null) && (authorizedStatusCode == StatusCode.AUTHENTICATION_ACCEPTED)) {
            if (serverCapabilities == null) {
                throw new ProtocolException("Error in creating a response: server capabilities not set (use setServerCapabilities())");
            }

            String commandReference = 
                serverCapabilitiesRequest.getCommandIdentifier().getValue();
            
            Meta meta = serverCapabilitiesRequest.getMeta();
            if (meta == null) {
                meta = new Meta(
                            new StringMetaContent(
                                Constants.MIMETYPE_SYNCML_DEVICEINFO_XML, 
                                "Type"
                            )
                        );
            }
            
            Data   data     = new Data(serverCapabilities.toXML());
            Source source   = ProtocolUtil.target2Source(
                                  serverCapabilitiesRequest.getItems()[0].getTarget()
                              );
            Item[] capabilities = new Item[] { new Item(null, source, null, data) };                                      

            ResultsCommand resultsCommand = new ResultsCommand(
                                                idGenerator.next()       ,
                                                syncHeader.getMessageID(),
                                                commandReference         ,
                                                meta /* meta */          ,
                                                null /* target ref */    ,
                                                null /* source ref */    ,
                                                capabilities                      
                                            );
            commandList.add(resultsCommand);
        }

        //
        // Alerts for each database to be synchronized
        //
        for (int i=0; (databases != null) && 
                      (i<databases.length); ++i ) {
                          
            if (databases[i].getStatusCode() == StatusCode.OK) {
            
                AlertCommand alertCommand = 
                ProtocolUtil.createAlertCommand(idGenerator.next(),
                                                false             ,
                                                null              ,
                                                databases[i]      );
                commandList.add(alertCommand);
            }
        }

        //
        // If client capabilities are required but not provided, a get command
        // must be added.
        //
        if (clientCapabilitiesRequired && (clientCapabilities == null)) {
            Meta meta = new Meta(
                            new StringMetaContent(
                                ProtocolUtil.getTypeElement(
                                    Constants.NAMESPACE_METINF              ,
                                    Constants.MIMETYPE_SYNCML_DEVICEINFO_XML
                                )
                                , 
                                "Type"
                            )
                        );
            Target target = new Target( InitializationRequirements.CAPABILITIES_TARGET, 
                                        InitializationRequirements.CAPABILITIES_TARGET);
            Item[] items = new Item[1]; 

            items[0] = new Item(
                           target,
                           null  , /* source */ 
                           null  , /* meta   */
                           null    /* data   */
                       );
            GetCommand getCommand = new GetCommand(
                                        idGenerator.next()      ,
                                        false /* no response */ ,
                                        null  /* language    */ ,
                                        null  /* credentials */ ,
                                        meta                    ,
                                        items
                                    );
            commandList.add(getCommand);
        }

        //
        // Constructs return message
        //
        Target target = new Target(syncHeader.getSource().getURI(),
                                   syncHeader.getSource().getLocationName());
        Source source = new Source(syncHeader.getTarget().getURI(),
                                   syncHeader.getTarget().getLocationName());
        SyncHeader responseHeader = new SyncHeader (
                                        getDTDVersion()                      ,
                                        getProtocolVersion()                 ,
                                        syncHeader.getSyncSessionIdentifier()                ,
                                        syncHeader.getMessageID()                            ,
                                        target                                               ,
                                        source                                               ,
                                        null  /* response URI */                             ,
                                        false                                                ,
                                        null /* credentials */                               ,
                                        syncHeader.getMeta()                                 
                                    );

        int size = commandList.size();
        AbstractCommand [] aCommands = new AbstractCommand[size];
        for (int i=0; i < size; i++) {
            aCommands[i] = (AbstractCommand)commandList.get(i);
        }
        SyncBody responseBody = new SyncBody(
                                    aCommands,
                                    isFlag(Flags.FLAG_FINAL_MESSAGE)
                                );

        try {
            return new Message(responseHeader, responseBody);
        } catch (RepresentationException e) {
            //
            // It should never happen !!!!
            //
            throw new ProtocolException("Unexpected error", e);
        }
    }
    
    // --------------------------------------------------------- Private methods
    
    
}
