/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*/
package sync4j.framework.protocol;

import sync4j.framework.core.*;
import sync4j.framework.protocol.Flags;

/**
 * This is a base abstract class for classes that implement SyncML protocol 
 * packages. A SyncML package is composed of a SyncHeader and a SyncBody which
 * follows the requirements that the protocol mandates.<br>
 * SyncPackage and its subclasses are designed to be used in two steps. First a 
 * SyncPackage is created and checked for validity and compliancy with the 
 * protocol. Than <i>getResponse()</i> can be used to get a response message for 
 * the given request. During the request validation process some information about the
 * request message are cached into instance variables and used in <i>getResponse()</i>.<br>
 * <p>
 * Subclasses are required to override the following methods:
 * <ul>
 *  <li><i>checkHeaderRequirements()</i></li>
 *  <li><i>checkBodyRequirements()</i></li>
 *  <li><i>getResponse()</i></li>
 * </ul>
 * <p>
 * In addition, subclasses are required to call super(header, body) in their
 * constructors, so that the base package is constructed and validated.
 *
 * @author  Stefano Fornari @ Funambol
 * @version $Id: SyncPackage.java,v 1.6 2003/12/15 16:57:21 luigiafassina Exp $
 */
public abstract class SyncPackage 
implements Flags {
    
    protected SyncHeader syncHeader = null;
    protected SyncBody   syncBody   = null;
    
    /**
     *  Creates a SyncPackage package.
     *
     *  @param syncHeader the header of the syncronization packet
     *  @param syncBody   the body of the syncronization packet
     */
    public SyncPackage(final SyncHeader syncHeader, 
                       final SyncBody   syncBody  ) {
            this.syncHeader = syncHeader;
            this.syncBody   = syncBody  ;
    }
    
    // ------------------------------------------------------------------- Flags
    
    /**
     * Flags
     */
    private boolean[] flags = new boolean[HOW_MANY_FLAGS];
    
    /**
     * Sets the given flag to the given value. If flag is FLAG_ALL, all flags are
     * set; if flag is FLAG_ALL_RESPONSES_REQUIRED all response flags are set.
     *
     * @param flag the flag to be set
     * @param value the boolean value of the flag
     */
    public void setFlagValue(int flag, boolean value) {
        switch (flag) {
            case FLAG_ALL:
                for (int i=0; i<HOW_MANY_FLAGS; i++) {
                    flags[i] = value;
                }
                break;
            case FLAG_ALL_RESPONSES_REQUIRED:
                flags[FLAG_SYNC_RESPONSE_REQUIRED         ] = 
                flags[FLAG_MODIFICATIONS_RESPONSE_REQUIRED] = 
                flags[FLAG_SYNC_STATUS_REQUIRED           ] = value;
                break;
            default:
                flags[flag] = value;
        }
    }
    
    /**
     * Sets the given flag to true.
     *
     * @param flag the flag to be set
     */
    public void setFlag(int flag) {
        setFlagValue(flag, true);
    }
    
    /**
     * Sets the given flag to false.
     *
     * @param flag the flag to be set
     */
    public void unsetFlag(int flag) {
        setFlagValue(flag, false);
    }
    
    /**
     * Returns the value of the given flag
     *
     * @param flag the flag to be returned
     *
     * @return the flag value
     */
    public boolean isFlag(int flag) {
        return flags[flag];
    }
    
    // -------------------------------------------------------------- Properties
    
    /**
     * The command id generator to be used to get new command ids
     */
    protected CommandIdGenerator idGenerator = null;
    
    public CommandIdGenerator getIdGenerator() {
        return this.idGenerator;
    }
    
    public void setIdGenerator(CommandIdGenerator idGenerator) {
        if (idGenerator == null) {
            throw new NullPointerException("idGenerator cannot be null");
        }
        this.idGenerator = idGenerator;
    }
    
    /**
     * The session id of the session into which the message is processed
     */
    public String getSessionId() {
        return syncHeader.getSyncSessionIdentifier().getValue();
    }
    
    /**
     * The used DTD version 
     */
    public DTDVersion getDTDVersion() {
        return syncHeader.getDTDVersion();
    }
    
    /**
     * The used protocol version
     */
    public ProtocolVersion getProtocolVersion() {
        return syncHeader.getProtocolVersion();
    }
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Checks that all requirements for the package are respected. It calls the
     * abstract methods <i>checkHeaderRequirements()</i> and <i>checkBodyRequirements()</i>.
     *
     * @throws ProtocolException
     */
    public void checkRequirements() 
    throws ProtocolException
    {
        checkHeaderRequirements();
        checkBodyRequirements();
    }
    
    /**
     * Returns the header of the original message
     *
     * @return the header of the original message
     */
    public SyncHeader getSyncHeader() {
        return this.syncHeader;
    }
    
    /**
     * Returns the body of the original message
     *
     * @return the body of the original message
     */
    public SyncBody getSyncBody() {
        return this.syncBody;
    }
    
    // -------------------------------------------------------- Abstract methods
    
    /**
     * Checks that all requirements regarding the header of the initialization 
     * packet are respected.
     *
     * @throws ProtocolException
     */
    abstract public void checkHeaderRequirements() throws ProtocolException;
    
    /**
     * Checks that all requirements regarding the body of the initialization 
     * packet are respected.
     *
     * @throws ProtocolException
     */
    abstract public void checkBodyRequirements() throws ProtocolException;
    
    /**
     * Constructs a proper response message.<br>
     *
     * @return the response message
     *
     * @throws ProtocolException in case of error or inconsistency
     */
    abstract public Message getResponse() throws ProtocolException;
}
