/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*/

package sync4j.framework.protocol.v11;

import sync4j.framework.core.*;

import sync4j.framework.protocol.ProtocolException;

/**
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: BasicRequirements.java,v 1.5 2003/12/15 16:57:05 luigiafassina Exp $
 */
public class BasicRequirements
implements Errors
{
    // ---------------------------------------------------- Contraint values

    public static final DTDVersion[]      SUPPORTED_DTD_VERSIONS     
        = new DTDVersion[] { Constants.DTD_1_0, Constants.DTD_1_1 };
    public static final ProtocolVersion[] SUPPORTED_PROTOCOL_VERSIONS
        = new ProtocolVersion[] { Constants.PROT_1_0, Constants.PROT_1_1, Constants.PROT_1_1_1 };
        
    public static final String          CAPABILITIES_SOURCE        = "./devinf11"                     ;
    public static final String          CAPABILITIES_TARGET        = "./devinf11"                     ;
    public static final String SERVER_CAPABILITIES = "server";
    public static final String CLIENT_CAPABILITIES = "client";

    //----------------------------------------------------------- Public methods

    static public void checkDTDVersion(DTDVersion version)
    throws ProtocolException
    {
        for (int i=0; i<SUPPORTED_DTD_VERSIONS.length; ++i) {
            if (SUPPORTED_DTD_VERSIONS[i].equals(version)) {
                return;  // OK!
            }
        }

        // KO!

            String[] args = new String[] { version.toString() };

            throw new ProtocolException(ERRMSG_DTD_VER_NOT_SUPPORTED, args);
        }

    static public void checkProtocolVersion(ProtocolVersion version)
    throws ProtocolException
    {
        for (int i=0; i<SUPPORTED_PROTOCOL_VERSIONS.length; ++i) {
            if (SUPPORTED_PROTOCOL_VERSIONS[i].equals(version)) {
                return;  // OK!
            }
        }
        
            String[] args = new String[] { version.toString() };

            throw new ProtocolException(ERRMSG_PROTOCOL_VER_NOT_SUPPORTED, args);
    }

    static public void checkSessionId(SyncSessionIdentifier id)
    throws ProtocolException
    {
        if ((id == null) || (id.toString().trim().length() == 0)) {
            throw new ProtocolException(ERRMSG_NO_SESSION_ID);
        }
    }

    static public void checkMessageId(String id)
    throws ProtocolException
    {
        if ((id == null) || (id.trim().length() == 0)) {
            throw new ProtocolException(ERRMSG_NO_MESSAGE_ID);
        }
    }

    /**
     * A target is valid if it is not null and one of <i>location name</i> or
     * <i>URI</i> is specified.
     *
     * @param target the target to be checked - NULL
     *
     * @throws ProtocolException
     */
    static public void checkTarget(Target target)
    throws ProtocolException
    {
        boolean valid    = target != null;
        boolean location = false;
        boolean uri      = false;

        if (valid) {
            location = (  (target.getLocationName() != null)
                       && (target.getLocationName().trim().length() != 0) );
            uri = (target.getURI() != null);

            valid = location || uri;
        }

        if (!valid) {
            String[] args = { ((target == null) ? "null" : target.toXML()) };
            throw new ProtocolException(ERRMSG_INVALID_TARGET, args);
        }
    }

    /**
     * A source is valid if it is not null and one of <i>location name</i> or
     * <i>URI</i> is specified.
     *
     * @param source the source to be checked - NULL
     *
     * @throws ProtocolException
     */
    static public void checkSource(Source source)
    throws ProtocolException
    {
        boolean valid    = source != null;
        boolean location = false;
        boolean uri      = false;

        if (valid) {
            location = (  (source.getLocationName() != null)
                       && (source.getLocationName().trim().length() != 0) );
            uri = (source.getURI() != null);

            valid = location || uri;
        }

        if (!valid) {
            String[] args = { ((source == null) ? "null" : source.toXML()) };
            throw new ProtocolException(ERRMSG_INVALID_SOURCE, args);
        }
    }

    static public void checkCommandId(CommandIdentifier commandId)
    throws ProtocolException
    {
        String cmdValue = null;

        if (  ( commandId == null                       )
           || ((cmdValue = commandId.getValue()) == null)
           || ( cmdValue.trim().length() == 0)          ) {
            throw new ProtocolException(ERRMSG_NO_MESSAGE_ID);
        }
    }

    static public void checkAlertCommand(AlertCommand alert)
    throws ProtocolException
    {
        //
        // Check the command id
        //
        String[] args = new String[] {"alert is null!"};

        if (alert == null) {
            throw new ProtocolException(ERRMSG_INVALID_ALERT, args);
        }

        try {
            checkCommandId(alert.getCommandIdentifier());
        } catch (ProtocolException e) {
            args = new String[] { e.getMessage() };
            throw new ProtocolException(ERRMSG_INVALID_ALERT, args);
        }
    }

    /**
     * Checks if the given command contains valid device information.
     *
     * @param cmd the command containing data
     * @param device specifies if the command should contain client or server
     *        capabilities
     *
     * @throws ProtocolException
     */
    static public void checkCapabilities(ItemizedCommand cmd, String device)
    throws ProtocolException
    {
        //
        // Checks command id
        //
        try {
            checkCommandId(cmd.getCommandIdentifier());
        } catch (ProtocolException e) {
            String[] args = new String[] { device, e.getMessage() };
            throw new ProtocolException(ERRMSG_INVALID_CAPABILITIES, args);
        }

        //
        // Checks the <Type> tag in metadata
        //
        Meta        meta        = null;
        MetaContent metaContent = null;

        meta = cmd.getMeta();
        if (meta == null) {
            Object[] args = new Object[] { device, "missing metadata" };
            throw new ProtocolException(ERRMSG_INVALID_CAPABILITIES, args);
        }
        metaContent = meta.getValue();

        if (!("Type".equals(metaContent.getContentTag()))) {
            Object[] args = new Object[] { device, "invalid metadata type" };
            throw new ProtocolException(ERRMSG_INVALID_CAPABILITIES, args);
        }

        //
        // NOTE: the validation of the metadata should be done at XML level
        //       (see http://www.syncml.org/docs/syncml_protocol_v101_20010615.pdf)
        //

        //
        // Checks source
        //
        Item[] items = cmd.getItems();

        if ((items == null) || (items.length ==0)) {
            String[] args = new String[] { device, ERRMSG_MISSING_ITEM};
            throw new ProtocolException(ERRMSG_INVALID_CAPABILITIES, args);
        }

        try {
            checkSource(items[0].getSource());
        } catch (ProtocolException e) {
            String[] args = new String[] { device, "missing source" };
            throw new ProtocolException(ERRMSG_INVALID_CAPABILITIES, args);
        }

        if (!CAPABILITIES_SOURCE.equals(items[0].getSource().getURI().toString())) {
            String[] args = new String[] { device, "URI not " + CAPABILITIES_SOURCE};
            throw new ProtocolException(ERRMSG_INVALID_CAPABILITIES, args);
        }

    }
}
