/*
 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*/
package sync4j.framework.protocol.v11;

import sync4j.framework.core.*;

import sync4j.framework.protocol.ProtocolException;
import sync4j.framework.protocol.v11.BasicRequirements;
import sync4j.framework.protocol.v11.Errors;

/**
 * This class groups utility methods used for checking that a client modifications
 * massage follows the requirements that the protocol mandates.
 *
 * @author Stefano Fornari @ Funambol
 */
public class ClientModificationsRequirements 
extends BasicRequirements 
implements Errors {
    
    // --------------------------------------------------------------- Constants
    
    // ---------------------------------------------------------- Public methods
 
    /**
     * Checks if the given command contains a valid request for server capablities.
     *
     * @param cmd the command containing the request
     *
     * @throws ProtocolException
     */
    static public void checkSync(SyncCommand cmd)
    throws ProtocolException
    {
        //
        // Checks command id
        //
        try {
            checkCommandId(cmd.getCommandIdentifier());
        } catch (ProtocolException e) {
            String[] args = new String[] { e.getMessage() };
            throw new ProtocolException(ERRMSG_INVALID_SYNC_COMMAND, args);
        }
        
        //
        // Checks modifications
        //
        AbstractCommand[] modifications = cmd.getCommands();
        
        for (int i=0; ((modifications != null) && (i<modifications.length)); ++i) {
            checkModification((ItemizedCommand)modifications[i]);
        }  // next i
    }
    
    /**
     * Checks the requirements for a modification command.
     *
     * @param cmd the modification command
     *
     * @throws ProtocolException
     */
    static public void checkModification(ItemizedCommand cmd)
    throws ProtocolException {
        //
        // Checks command id
        //
        try {
            checkCommandId(cmd.getCommandIdentifier());
        
        
            Item[] items = cmd.getItems();

            //
            // The type of each single item can be specified at command level
            // or at item level. If the type is specified at command level, items
            // without type will take the command type as default. If the type
            // is not specified at command level, each single item MUST specify
            // its type.
            //
            Meta meta = cmd.getMeta();
            boolean checkType =  meta.isTypeContent() 
                              && !Util.isEmpty(((StringMetaContent)meta.getValue()).getContent())
                              ;
            for(int i=0; ((items != null) && (i<items.length)); ++i) {
                // 
                // NOTE: all commands but delete use <Data> to carry data about 
                // the modification
                //
                checkModificationItem(items[i], checkType, !cmd.getName().equals("Delete"));
            }
            
        } catch (ProtocolException e) {
            String[] args = new String[] { e.getMessage() };
            throw new ProtocolException(ERRMSG_INVALID_MODIFICATION_COMMAND, args);
        }
    }
    
    /**
     * Checks an item included into a modification command
     *
     * @param item the item to be checked
     * @param checkType indicats if the type specified in the <Meta> tag is
     *                  mandatory and must be checked
     * @param checkData indicates when check for the existance of the <Data> tag
     *
     * @throws ProtocolException
     */
    static public void checkModificationItem(final Item    item     , 
                                             final boolean checkType,
                                             final boolean checkData) 
    throws ProtocolException {
        checkSource(item.getSource());
        
        if (checkType) {
            Meta meta = item.getMeta();
            
            if (  (meta.isTypeContent() == false)
               || Util.isEmpty(((StringMetaContent)meta.getValue()).getContent())) {
               String[] args = new String[] { item.toString() };
               throw new ProtocolException(ERRMSG_MISSING_TYPE, args);
            }
        }
        
        if (checkData) {
            if (item.getData() == null) {
                String[] args = new String[] { item.toString() };
                throw new ProtocolException(ERRMSG_MISSING_DATA, args);
            }
        }
    }
        
}
