/*

 $Id: JAASOfficer.java,v 1.8 2003/10/13 09:32:47 luigiafassina Exp $

 Copyright (c) 2001 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.security;


import java.util.logging.Logger;
import java.util.logging.Level;

import java.security.Principal;
import javax.security.auth.*;
import javax.security.auth.login.*;
import javax.security.*;

import sync4j.framework.security.Officer;
import sync4j.framework.security.jaas.CredentialHandler;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.core.Credential;


/**
 * This implementation of <i>Officier</i> uses JAAS for authentication and
 * authorization.
 * <p>
 * In order to use this implementation, remember to set the system property
 * java.security.auth.login.config accordingly as specified in JAAS documentation
 * or in the documentation of the application server in use.<br>
 * For example, in the case of the J2EE RI, in order to use the SimpleLoginModule
 * delivered with the framework, the following lines have to be added to the
 * file <i>{J2EESDK_HOME}/lib/security/login.config</i>:
 * <pre>
 * sync4j {
 *   sync4j.framework.security.jaas.SimpleLoginModule required required debug=true;
 * }
 * </pre>
 * The security policy must be changed as well; add the following lines to the
 * file <i>{J2EESDK_HOME}/lib/security/server.policy</i>:
 * <pre>
 * permission javax.security.auth.AuthPermission "createLoginContext.sync4j";
   permission javax.security.auth.AuthPermission "modifyPrincipals";
 * </pre>
 *
 * TO DO: authorization
 *
 * @author  Stefano Fornari @ Funambol.com
 */
public class JAASOfficer implements Officer, java.io.Serializable {
   
    // ---------------------------------------------------------- Protected data

    protected transient Logger log = Sync4jLogger.getLogger();

    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for incorrect login/password?
     */
    private boolean loginFailed = false;

    public boolean isLoginFailed() {
        return loginFailed;
    }

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    public boolean isLoginExpired() {
        return loginExpired;
    }
    
    /**
     * Can use a guest credential if there is not credential in the message?
     */
    private boolean guestEnabled = true;
        
    public boolean isGuestEnabled() {       
        return this.guestEnabled;
    }
    
    public void setGuestEnabled(boolean guestEnabled) {
        this.guestEnabled = guestEnabled;
    }

    
    // ---------------------------------------------------------- Public methods

    /**
     * Authenticates a credential based on the configured JAAS subsystem.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Credential credential) {
        CredentialHandler handler = null;

        if (log.isLoggable(Level.INFO)) {
            log.info("Authenticating credential: " + credential);
        }

        try {
            handler = new CredentialHandler(credential);
            LoginContext lc = new LoginContext("sync4j", handler);
            lc.login();

            if (log.isLoggable(Level.INFO)) {
                log.info(lc.getSubject() + " authenticated!");
            }
        } catch (AccountExpiredException e) {
             if (log.isLoggable(Level.FINEST)) {
                log.throwing(getClass().getName(), "authenticate", e);
            }
            log.info( "Login failed for "
                    + handler.getLogin()
                    );

            loginFailed  = false;
            loginExpired = true;

            return false;
        } catch (FailedLoginException e) {
             if (log.isLoggable(Level.FINEST)) {
                log.throwing(getClass().getName(), "authenticate", e);
            }
            log.info( "Login failed for "
                    + handler.getLogin()
                    );

            loginFailed  = true;
            loginExpired = false;

            return false;
        } catch (LoginException e) {
            if (log.isLoggable(Level.FINEST)) {
                log.throwing(getClass().getName(), "authenticate", e);
            }
            log.info( "Login failed for "
                    + handler.getLogin()
                    );

            loginFailed  = false;
            loginExpired = false;

            return false;
        } catch (IllegalArgumentException e) {
            log.throwing(getClass().getName(), "authenticate", e);
            return false;
        } catch (Throwable t) {
            //
            // Maybe a configuration error
            //
            t.printStackTrace();
        }

        log.info( handler.getLogin() + " logged in");
        return true;
    }

    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource the name (or the identifier) of the resource to be authorized
     *
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /** Un-authenticates a credential.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Credential credential) {
        //
        // Do nothing. In the current implementation, the authentication is
        // discarde as soon as the LoginContext is garbage collected.
        //
    }


    private void readObject(java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject();
        log = Sync4jLogger.getLogger();
    }

}
