/*
 Copyright (c) 2001 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.security;

import java.security.Principal;

import sync4j.framework.core.Constants;
import sync4j.framework.tools.Base64;

/**
 * This class implements the <i>Pricipal</i> interface and represents a Sync4j
 * principal. In Sync4j a principal is represented by a couple
 * <i>(username, device id)</i> and is associated to a unique id.
 *
 * @author  Stefano Fornari @ Funambol.com
 *
 * @version $Id: Sync4jPrincipal.java,v 1.8 2003/12/19 13:36:42 luigiafassina Exp $
 */
public class Sync4jPrincipal implements Principal, java.io.Serializable {
    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data

    private String id       = null; // principalid
    private String username = null;
    private String deviceId = null;
    private String email    = null;
    private String firstName= null;
    private String lastName = null;

    // ------------------------------------------------------------ Constructors
    /**
     * Disable the default constructor.
     */
    protected Sync4jPrincipal() {};

    /**
     * Creates a new instance of UsernamePrincipal
     *
     * @param id principal id
     * @param username the username element of the principal
     * @param deviceId the deviceId element of the principal
     */
    public Sync4jPrincipal(String id, String username, String deviceId) {
        this.id       = id      ;
        this.username = username;
        this.deviceId = deviceId;
    }

    /**
     * Creates a new instance of UsernamePrincipal
     *
     * @param username the username element of the principal
     * @param deviceId the deviceId element of the principal
     *
     */
    public Sync4jPrincipal(String username, String deviceId) {
        this(null, username, deviceId);
    }


    /**
     * Creates a new instance of UsernamePrincipal
     *
     * @param id principal id
     */
    public Sync4jPrincipal(String id) {
        this(id, null, null);
    }

    /**
     * Creates a new instance of CredentialPrinciple from the given Credential
     *
     * @param userpwd user and password in the form <user>:<password>
     * @param type the username coding as expressed in the a Credential object
     * @param deviceId the deviceId element of the principal. Can be NULL
     *
     * @see sync4j.framework.core.Credential
     *
     * @throws IllegalArgumentException
     */
    public static Sync4jPrincipal fromCredential( String userpwd  ,
                                                  String type     ,
                                                  String deviceId ) throws IllegalArgumentException {

        Sync4jPrincipal ret = new Sync4jPrincipal();

        if (Constants.AUTH_SUPPORTED_TYPES.indexOf(type) < 0) {
            throw new IllegalArgumentException( "Authorization type '"
                                              + type
                                              + "' not supported"
                                              );
        }

        if (Constants.AUTH_TYPE_BASIC.equals(type)) {
            String s = new String(Base64.decode(userpwd.getBytes()));

            int p = s.indexOf(':');

            if (p == -1) {
                ret.setUsername(s);
            } else {
                ret.setUsername((p>0) ? s.substring(0, p) : "");
            }
        } else if (Constants.AUTH_TYPE_CLEAR.equals(type)) {
            String s = userpwd;

            int p = s.indexOf(':');

            if (p == -1) {
                ret.setUsername(s);
            } else {
                ret.setUsername((p>0) ? s.substring(0, p) : "");
            }
        }

        ret.setId(null);
        ret.setDeviceId(deviceId);

        return ret;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * The name of this principal is in the form:
     *   {username}.{deviceId}
     *
     * @return the principal's name
     */
    public String getName() {

        return (deviceId == null)
             ? username
             : (username + '.' + deviceId);
    }

    /**
     * The username part of the principal name.
     *
     * @return the username part of the principal name.
     */
    public String getUsername() {
        return username;
    }

    /**
     * Set the username part of the principal name.
     *
     * @param username the username principal's username
     */
    public void setUsername(String username) {
        this.username = username;
    }

    /**
     * The device id part of the principal name.
     *
     * @return the device id part of the principal name.
     */
    public String getDeviceId() {
        return deviceId;
    }

    /**
     * Set the device id part of the principal name.
     *
     * @param deviceId the device id principal's username
     */
    public void setDeviceId(String deviceId) {
        this.deviceId = deviceId;
    }

    /**
     * The principal unique id.
     *
     * @return the unique principal id.
     */
    public String getId() {
        return id;
    }

    /**
     * Set the principal unique id.
     *
     * @param id the principal unique id.
     */
    public void setId(String id) {
        this.id = id;
    }

    /**
     * The user's email of the principal.
     *
     * @return the user's email of the principal.
     */
    public String getEmail() {
        return email;
    }

    /**
     * Set the user's email of the principal.
     *
     * @param email the user's email of the principal
     */
    public void setEmail(String email) {
        this.email = email;
    }

    /**
     * The user's first name of the principal.
     *
     * @return the user's first name of the principal.
     */
    public String getFirstName() {
        return firstName;
    }

    /**
     * Set the user's first name of the principal.
     *
     * @param firstName the user's first name of the principal
     */
    public void setFirstName(String firstName) {
        this.firstName = firstName;
    }
    
    /**
     * The user's last name of the principal.
     *
     * @return the user's last name of the principal.
     */
    public String getLastName() {
        return lastName;
    }

    /**
     * Set the user's last name of the principal.
     *
     * @param lastName the user's last name of the principal
     */
    public void setLastName(String lastName) {
        this.lastName = lastName;
    }
    
    /**
     * Two Sync4jPrincipals are equals if and only if:
     * <ul>
     *   <li>o is null or
     *   <li>o is not an instance of Sync4jPrincipal or
     *   <li>the two getName()s do not match
     * </ul>
     *
     * @param o the reference object with which to compare.
     *
     * @return true if this object is the same as the obj argument; false otherwise.
     */
    public boolean equals(Object o) {
        if ((o == null) || !(o instanceof Sync4jPrincipal)) {
            return false;
        }

        return (getName().equals(((Sync4jPrincipal)o).getName()));
    }

    public int hashCode() {
        return getName().hashCode();
    }

    public String toString() {
        return getName();
    }
}
