/*

 $Id: CredentialHandler.java,v 1.5 2003/12/15 16:57:39 luigiafassina Exp $

 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.security.jaas;

import java.io.*;
import java.util.logging.Logger;
import java.util.logging.Level;

import javax.security.auth.callback.*;

import sync4j.framework.core.Credential;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.tools.Base64;
import org.apache.commons.lang.StringUtils;


/**
 * This handler implements the JAAS <i>CallbackHandler</i> interface. It stores 
 * a <i>Credential</i> object for later use as principal and credentials provider. 
 * This simple implementation supports basic authentication stored bas64 encoded
 * in the form login:password.
 *
 * TO DO: supports MD5 authentication
 *
 * @author  Stefano Fornari @ Funambol.com
 */
public class CredentialHandler implements CallbackHandler {
    
    // --------------------------------------------------------------- Constants
    
    public static final String TYPE_BASIC = "syncml:auth-basic";
    public static final String TYPE_CLEAR = "syncml:auth-clear";
    public static final String SUPPORTED_TYPES = TYPE_BASIC + ',' + TYPE_CLEAR;
    
    // ------------------------------------------------------------ Private data
    
    private String     login      = null;
    private char[]     password   = null;
    
    // ---------------------------------------------------------- Protected data
    
    protected final Logger log = Sync4jLogger.getLogger();
    
    // ------------------------------------------------------------ Constructors
    
    /**
     * Creates a new instance of CredentialHandler 
     *
     * @param credential
     * 
     * @throws IllegalArgumentException
     */
    public CredentialHandler(Credential credential)
    throws IllegalArgumentException {
        
        String type = credential.getType();
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("credential: " + credential);
        }
        
        if (SUPPORTED_TYPES.indexOf(type) < 0) {
            throw new IllegalArgumentException( "Authorization type '"
                                              + type
                                              + "' not supported"
                                              );
        }
        
        if (TYPE_BASIC.equals(type)) {
            String s = new String(Base64.decode(credential.getData()));
            
            int p = s.indexOf(':');
            
            if (p == -1) {
                login = s;
                password = null;
            } else {
                login = (p>0) ? s.substring(0, p) : "";
                password = toChars((p == (s.length()-1)) ? "" : s.substring(p+1));
            }
        } else if (TYPE_CLEAR.equals(type)) {
            String s = credential.getData();
            
            int p = s.indexOf(':');
            
            if (p == -1) {
                login = s;
                password = null;
            } else {
                login = (p>0) ? s.substring(0, p) : "";
                password = toChars((p == (s.length()-1)) ? "" : s.substring(p+1));
            }
        }
    }
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Returns the login of this credential.
     * 
     * @return the login value
     */
    public String getLogin() {
        return login;
    }
    
    // ------------------------------------------------------------------ handle
    
    public void handle(Callback[] callbacks) 
    throws IOException, UnsupportedCallbackException {
        for (int i = 0; i < callbacks.length; i++) {
            log.finest("Handling " + callbacks[i]);
            
            if (callbacks[i] instanceof TextOutputCallback) {

                // display the message according to the specified type
                TextOutputCallback toc = (TextOutputCallback)callbacks[i];
                switch (toc.getMessageType()) {
                   case TextOutputCallback.INFORMATION:
                      log.info(toc.getMessage());
                      break;
                   case TextOutputCallback.ERROR:
                      log.severe(toc.getMessage());
                      break;
                   case TextOutputCallback.WARNING:
                      log.warning(toc.getMessage());
                     break;
                   default:
                      throw new IOException( "Unsupported message type: " 
                                           + toc.getMessageType()
                                           );
                }

            } else if (callbacks[i] instanceof NameCallback) {
                NameCallback nc = (NameCallback)callbacks[i];
                nc.setName(login);
            } else if (callbacks[i] instanceof PasswordCallback) {
                PasswordCallback nc = (PasswordCallback)callbacks[i];
                nc.setPassword(password);
            } else {
                throw new UnsupportedCallbackException
                 (callbacks[i], "Unrecognized Callback");
            }
          }
    }
    
    // --------------------------------------------------------- Private methods
    
    private char[] toChars(String str) {
        if (StringUtils.isEmpty(str)) {
            return new char[0];
        }
        
        int l = str.length();
        char[] ret = new char[l];
        
        str.getChars(0, l, ret, 0);
        
        return ret;
    }
    
}
