/*

 $Id: SimpleLoginModule.java,v 1.3 2003/01/23 22:33:39 stefano_fornari Exp $

 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.framework.security.jaas;

import java.util.logging.*;
import java.util.Map;
import java.security.Principal;

import javax.security.auth.*;
import javax.security.auth.spi.*;
import javax.security.auth.login.*;
import javax.security.auth.callback.*;

import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.logging.Sync4jLogger;

/**
 * This is a simple implementation of a JAAS login module. It always authenticates
 * and uthorizes.
 *
 * @author  Stefano Fornari @ Funambol.com
 */
public class SimpleLoginModule implements LoginModule {
    
    // ------------------------------------------------------------ Private data
    
    private Subject         subject        = null ;
    private CallbackHandler handler        = null ;
    private boolean         loginSucceded  = false;
    private boolean         loginCommitted = false;
    private Principal       principal      = null ;
    private String          username       = null ;
    private String          password       = null ;
    
    // ---------------------------------------------------------- Protected data
    
    protected final Logger log = Sync4jLogger.getLogger();
    
    // ---------------------------------------------------------- Public methods
    
    public void initialize(Subject         subject            ,
                           CallbackHandler handler            ,
                           Map             sharedState        ,
                           Map             options            ) {
                               
        this.subject = subject;
        this.handler = handler;
        this.loginSucceded = this.loginCommitted = false;
        this.username = null;
        this.password = null;
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("sharedState: " + sharedState);
            log.finest("options: " + options);
        }
    }
    
    public boolean login() throws LoginException {
        if (handler == null) {
            throw new LoginException("No CallbackHandler defined");
        }
        
        //
        // Creates two callbacks: one for user, one for password. They are for
        // demonstration purpose only, because they will never be used by this
        // implementation of <i>LoginModule</i>
        //
        Callback[] callbacks = new Callback[2];
        
        NameCallback nameCallback = new NameCallback("Username");
        PasswordCallback passwordCallback = new PasswordCallback("Password", false);
        
        callbacks[0] = nameCallback;
        callbacks[1] = passwordCallback;
        
        try {
            handler.handle(callbacks);
            
            username = nameCallback.getName();
            password = (passwordCallback.getPassword() == null) ? null 
                                                                : new String(passwordCallback.getPassword());
            
            passwordCallback.clearPassword();
        } catch (Exception e) {
            throw new LoginException(e.toString());
        }
        
        loginSucceded = true;
        return true;
    }
    
    public boolean commit() throws LoginException {
        if (loginSucceded == false) {
            return false;
        }
        
        //
        // Login succeded, so create a Principal and add it to the subject
        //
        principal = new Sync4jPrincipal(null, username, null);
        
        if (!(subject.getPrincipals().contains(principal))) {
            subject.getPrincipals().add(principal);
        }
        
        username = null;
        password = null;
        loginCommitted = true;
        
        return true;
    }
    
    public boolean abort() throws LoginException {
        if (loginSucceded == false) {
            return false;
        } else if (loginCommitted == false) {
            loginSucceded = false;
            username  = password  = null;
            principal = null;
        } else {
            logout();
        }
        
        return true;
    }
    
    public boolean logout() throws LoginException {
        subject.getPrincipals().remove(principal);
        
        loginSucceded = loginCommitted = false;
        username = password = null;
        principal = null;
        
        return true;
    }
    
}
