/*

 $Id: AbstractSessionManager.java,v 1.1 2002/10/07 20:51:29 stefano_fornari Exp $

 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.server.session;

import java.io.Serializable;

import java.util.Collection;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.server.session.SessionHandler;
import sync4j.framework.server.session.SessionExpiredException;

/**
 * This is a basic implementation of a session manager. Subclasses have to 
 * extend it in order to get and create sessions.
 *
 * @author  Stefano Fornari @ Funambol.com
 */
public abstract class AbstractSessionManager implements Serializable {
    
    // -------------------------------------------------------------- Properties
    
    /**
     * The lifetime of a session in minutes
     */
    private int lifeTime = 5;
    
    /** Getter for property lifeTime.
     * @return Value of property lifeTime.
     */
    public int getLifeTime() {
        return lifeTime;
    }
    
    /** Setter for property lifeTime.
     * @param lifeTime New value of property lifeTime.
     */
    public void setLifeTime(int lifeTime) {
        this.lifeTime = lifeTime;
    }

    
    // ------------------------------------------------------------ Private data
    
    /**
     * The logger
     */
    private static final Logger log = Sync4jLogger.getLogger();
        
    // ---------------------------------------------------------- Public methods
    
    public SessionHandler getSessionHandler(String sessionId) 
    throws SessionExpiredException {
        log.finest("Looking for session " + sessionId);
        
        SessionHandler handler = getSessionFromBag(sessionId);
        
        if (handler == null) {
            handler  = createNewSession(sessionId);
            putSessionInBag(sessionId, handler);
        } else {
            if (handler.isNew()) {
                handler.setNew(false);
                putSessionInBag(sessionId, handler);
            }
        }
        
        return handler;
    }
    
    public void removeSession(String sessionId) {
        log.fine("Removing session " + sessionId);
        
        try {
            SessionHandler oldSession = getSessionFromBag(sessionId);
            oldSession.expire();
            removeSessionFromBag(sessionId);
        } catch (SessionExpiredException e) {
            log.finest("Session already expired");
        }
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Remaining sessions: " + getSessionBag());
        }
    }
    
    public void removeSession(SessionHandler handler) {
        log.fine("Removing session " + ((handler!=null) ? handler.getSessionId() : "null"));
        
        handler.expire();
        removeSessionFromBag(handler.getSessionId());
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Remaining sessions: " + getSessionBag());
        }
    }
    
    public void storeSessionHandler(SessionHandler handler) {
        log.fine("Storing session " + handler);
        
        putSessionInBag(handler.getSessionId(), handler);
    }
    
    
    // -------------------------------------------------------- Abstract methods
    
    abstract protected SessionHandler createNewSession(String sessionId);
    abstract protected SessionHandler getSessionFromBag(String sessionId) throws SessionExpiredException;
    abstract protected void putSessionInBag(String sessionId, SessionHandler handler);
    abstract protected void removeSessionFromBag(String sessionId);
    abstract protected Collection getSessionBag();
}
