/*
 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE. 
 */

package sync4j.framework.server.store;

import java.util.Map;
import java.util.logging.Logger;

import java.sql.*;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import sync4j.framework.server.store.PersistentStoreException;
import sync4j.framework.logging.Sync4jLogger;

import org.apache.commons.lang.StringUtils;

/**
 * This is a base class for <i>PersistenStore</i> objects. It does not persist
 * any object, but it provides services common to concrete implementations.
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: BasePersistentStore.java,v 1.4 2003/12/15 16:57:58 luigiafassina Exp $
 *
 */
public abstract class BasePersistentStore {
    
    // --------------------------------------------------------------- Constants
    
    public static final String
    CONFIG_JNDI_DATA_SOURCE_NAME = "jndi-data-source-name";
    
    /**
     * Logger
     */
    protected transient final Logger log = Sync4jLogger.getLogger();
    
    // -------------------------------------------------------------- Properties
    
    /**
     * The JNDI name of the datasource to be used
     */
    protected String jndiDataSourceName = null;
    
    public String getJndiDataSourceName() {
        return this.jndiDataSourceName;
    }
    
    public void setJndiDataSourceName(String jndiDataSourceName) throws PersistentStoreException {
        this.jndiDataSourceName = jndiDataSourceName;
        
        if (jndiDataSourceName == null) {
            dataSource = null;
        }
        
        try {
            InitialContext ctx = new InitialContext();
            dataSource = (DataSource) ctx.lookup(jndiDataSourceName);
        } catch (NamingException e) {
            throw new PersistentStoreException("Data source "
            + jndiDataSourceName
            + " not found"
            , e
            );
        }
    }
    
    // ---------------------------------------------------------- Protected data
    
    protected transient DataSource dataSource = null;
    
    // ------------------------------------------------------------ Constructors
    
    // ---------------------------------------------------------- Public methods

    /** Configures the persistent store
     *
     * @param config an <i>Map</i> containing configuration parameters.
     *
     * @throws ConfigPersistentStoreException
     *
     */
    public void configure(Map config) throws ConfigPersistentStoreException {
        
        checkConfigParams(config);
        
        try {
            setJndiDataSourceName((String) config.get(CONFIG_JNDI_DATA_SOURCE_NAME));
        } catch (PersistentStoreException e) {
            throw new ConfigPersistentStoreException( "Error creating the datasource: "
                                                    + e.getMessage()
                                                    , e
                                                    );
        }
    }
        
    // ------------------------------------------------------- Protected methods
        
    // ------------------------------------------------------- Protected methods
        
    // --------------------------------------------------------- Private methods
    
    /**
     * Checks if the given configuration parameters contain all required
     * parameters. If not a <i>ConfigPersistentStoreException</i> is thrown.
     *
     * @param config the <i>Map</i> containing the configuration parameters
     *
     * @throws ConfigPersistentStoreException in case of missing parameters
     */
    private void checkConfigParams(Map config)
    throws ConfigPersistentStoreException {
        StringBuffer sb = new StringBuffer();
        
        if (StringUtils.isEmpty((String) config.get(CONFIG_JNDI_DATA_SOURCE_NAME))) {
            sb.append(CONFIG_JNDI_DATA_SOURCE_NAME);
        }
    }
}
