/*

 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.server.store;

import java.util.Map;

import sync4j.framework.server.store.Clause;
import sync4j.framework.server.store.ConfigPersistentStoreException;
import sync4j.framework.server.store.PersistentStoreException;

/**
 * A <i>PersistentStore</i> is a class that stores objects in a persistent media
 * such as a database. The work of saving and reading data to and from the store 
 * is delegated to the store() and read() methods that can take the appropriate
 * actions based on the type of the object that has to be written or read.
 * <p>
 * To store an object just call <i>store(obj)</i>.<br>
 * To read an object call read(obj).
 * <p>
 * Note that those two methods return true if they know how to deal with the 
 * given object; that return value is not intended to be a success indicator. It 
 * just tells the caller that the <i>PersistentStore</i> knew how to process the
 * given object.
 * <p>
 * A <i>PersistentStore</i> can be configured calling <i>configure()</i> with a 
 * <i>java.util.Map</i> parameter containing configuration information. The 
 * content of the map is implementation specific.
 * 
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: PersistentStore.java,v 1.6 2004/01/05 13:08:27 stefano_fornari Exp $
 */
public interface PersistentStore {
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Configure the persistent store
     * 
     * @param config an <i>Map</i> containing configuration parameters.
     *
     * @throws ConfigPersistentStoreException
     */
    public void configure(Map config) throws ConfigPersistentStoreException;
    
    /**
     * Store the given object to the persistent media.
     *
     * @param o object to be stored
     *
     * @return true if this <i>PersistentStore</i> processed the given object,
     *         false otherwise
     *
     * @throws PersistentStoreException
     */
    public boolean store(Object o) throws PersistentStoreException;
    
    /**
     * Read from the persistent media the given object.
     *
     * @param o object to be read
     *
     * @return true if this <i>PersistentStore</i> processed the given object,
     *         false otherwise
     *
     * @throws PersistentStoreException
     */
    public boolean read(Object o) throws PersistentStoreException;
    
    /**
     * Read all objects stored the persistent media.
     *
     * @param objClass the object class handled by the persistent store
     *
     * @return an array containing the objects read. If no objects are found an
     *         empty array is returned. If the persistent store has not 
     *         processed the quest, null is returned.
     *
     * @throws PersistentStoreException
     */
    public Object[] read(Class objClass) throws PersistentStoreException;
    
    /**
     * Delete the given object to the persistent media
     *
     * @param o object to be deleted
     *
     * @return true if this <i>PersistentStore</i> processed the given object,
     *         false otherwise
     *
     * @throws PersistentStoreException
     */
    public boolean delete(Object o) throws PersistentStoreException;
    
    /**
     * Read all objects stored the persistent media.
     *
     * @param objClass the object class handled by the persistent store
     * @param clause the array of select conditions
     *
     * @return an array containing the objects read. If no objects are found an
     *         empty array is returned. If the persistent store has not 
     *         processed the quest, null is returned.
     *
     * @throws PersistentStoreException
     */
    public Object[] read(Object o, Clause clause) throws PersistentStoreException;
    
    /**
     * Store the given object to the persistent media.
     *
     * @param id the id of the object
     * @param o object to be stored
     * @param operation the operation to execute on the object (insert or update)
     *
     * @return true if this <i>PersistentStore</i> processed the given object,
     *         false otherwise
     *
     * @throws PersistentStoreException
     */
    public boolean store(String id, Object o, String operation) throws PersistentStoreException;    
    
}
