/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
package sync4j.framework.tools;

import java.io.*;
import java.util.Vector;

import org.kxml.parser.ParseEvent;
import org.kxml.parser.XmlParser;
import org.kxml.parser.Tag;
import org.kxml.wap.SyncML;
import org.kxml.wap.SyncMLWriter;
import org.kxml.wap.SyncMLParser;
import org.kxml.Xml;
import org.kxml.Attribute;

import sync4j.framework.core.Message;
import sync4j.framework.core.Sync4jException;


/**
 *  Utility class for WBXML stuff
 *
 * @author  Stefano Fornari
 * @version $Id: WBXMLTools.java,v 1.5 2003/12/15 16:58:28 luigiafassina Exp $
 */
public class WBXMLTools {
    
    // --------------------------------------------------------------- Constants
    
    public static final String WELL_KNOWN_NS = ",DevInf,";
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Converts a string to a WBXML message.
     *
     * @param s the String to convert - NOT NULL
     *
     * @return the WBXML message
     *
     * @throws Sync4jException in case of parser errors
     */
    public static byte[] toWBXML(final String s) 
    throws Sync4jException {
        SyncMLWriter writer = null;
        try {
            writer = new SyncMLWriter();
            XmlParser xml = new XmlParser(new StringReader(s));
            traverseXML(xml, writer);
        } catch (IOException e) {
            throw new Sync4jException(e.getMessage(), e);
        } finally {
            if (writer != null) try {writer.close();} catch (Exception e) {}
        }

        return writer.getBytes();
    }
    
    /**
     * Encodes a <i>Message</i> to WBXML
     * <p>
     * The message is fixed before encoding in order to get a converted message 
     * that makes sense. For instance, the Meta type of a Results element (if
     * there is any), must be changed from <code>application/vnd.syncml-devinf+xml</code>
     * to <code>application/vnd.syncml-devinf+wbxml</code>
     *
     * @param msg the message to encode
     *
     * @return the encoded stream of bytes (as a byte[] buffer).
     *
     * @throws Sync4jException in case of errors
     */
    public static byte[] toWBXML(Message msg) 
    throws Sync4jException {
        return toWBXML(msg.toXML());
    }
        
    /**
     * Converts a WBXML message into the corresponding XML message.
     *
     * @param msg the message to convert - NOT NULL
     *
     * @return the XML message or NULL if an error occurred
     *
     * @throws Sync4jException in case of parser errors
     */
    public static String wbxmlToXml(final byte[] msg)
    throws Sync4jException {

        ByteArrayInputStream in = null;
        try {
            in = new ByteArrayInputStream(msg);

            SyncMLParser parser = new SyncMLParser(in);

            return parseWBXML(parser);
        } catch (Throwable t) {
            throw new Sync4jException(t.getMessage(), t);
        }
    }
    
    public static boolean isWellKnownNamespace(String ns) {
        return (WELL_KNOWN_NS.indexOf(',' + ns + ',') >= 0);
    }
    
    // --------------------------------------------------------- Private methods    
    private static void traverseXML(XmlParser parser, SyncMLWriter writer) throws IOException {
        //
        // NOTE: when the namespace changes in one of the namespaces listed 
        //       in WELL_KNOWN_NS, a well known document must be inserted; 
        //       therefore a new inner writer is created when the tag is opened
        //       and its content flushed in the original writer when the tag is 
        //       closed
        //
        
        boolean leave = false;

        do {
            ParseEvent event = parser.read();

            switch (event.getType()) {
                case Xml.START_TAG:
                    SyncMLWriter tagWriter = null;
                    
                    String name = event.getName();
                    if (isWellKnownNamespace(name)) {
                        tagWriter = new SyncMLWriter(((Tag)event).getNamespace(), SyncML.VER_10);
                    } else {
                        tagWriter = writer;
                    }
                    
                    // see API doc of StartTag for more access methods
                    tagWriter.startTag(name);
                    traverseXML(parser, tagWriter); // recursion
                    
                    if (tagWriter != writer) {
                        tagWriter.close();
                        writer.writeOpaque(new String(tagWriter.getBytes()));
                        tagWriter = null;
                    }
                    break;

                case Xml.END_TAG:
                    writer.endTag();
                    leave = true;
                    break;

                case Xml.END_DOCUMENT:
                    leave = true;
                    break;

                case Xml.TEXT:
                    writer.write(event.getText());
                    break;

                case Xml.WHITESPACE:
                    break;

                default:
            }
        } while (!leave);
    }
    
    private static String parseWBXML(SyncMLParser parser) throws IOException{

        StringBuffer buf=new StringBuffer();
        boolean leave = false;

        do {
            ParseEvent event = parser.read();

            switch (event.getType()) {
                case Xml.START_TAG:

                    buf.append("<");
                    buf.append(event.getName());
                    Vector attrs=event.getAttributes();
                    if(attrs!=null){
                        for(int i=0;i<attrs.size();i++){
                            Attribute attr=(Attribute)attrs.elementAt(i);
                            buf.append(" ");
                            buf.append(attr.getName());
                            buf.append("='");
                            buf.append(attr.getValue());
                            buf.append("'");
                        }
                    }
                    buf.append(">");
                    buf.append(parseWBXML(parser));
                    break;

                case Xml.END_TAG:
                    buf.append("</");
                    buf.append(event.getName());
                    buf.append(">");
                    leave = true;
                    break;

                case Xml.END_DOCUMENT:
                    leave = true;
                    break;

                case Xml.TEXT:
                    buf.append(event.getText());
                    break;
                    
                case Xml.WAP_EXTENSION:
                    buf.append(event.getText());
                    break;

                case Xml.WHITESPACE:
                    break;

                default:
            }
        } while (!leave);
        
        return buf.toString();
    }
}
