/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.framework.tools.beans;

import java.beans.XMLEncoder;
import java.beans.XMLDecoder;

import java.io.*;

import sync4j.framework.tools.beans.LazyInitBean;
import sync4j.framework.tools.beans.BeanException;
import sync4j.framework.tools.beans.BeanInstantiationException;
import sync4j.framework.tools.beans.BeanInitializationException;



/**
 * This is a JavaBeans factory. At the moment it has a very simple 
 * implementation: it loads the class from a serialized file at runtime.
 * 
 * @author Stefano Fornari @ Funambol
 * @version $Id: BeanFactory.java,v 1.6 2003/12/15 16:58:28 luigiafassina Exp $
 */
public class BeanFactory {

    protected BeanFactory() { }

    /**
     * Creates an instance of the given bean. The bean is first search as a
     * class. If not found as class, it is searched as a resource.
     *
     * @param       classLoader the class loader to use for loading the bean
     * @param       beanName    name of the bean (for now is also the file name,
     *                          without .ser, into which the bean has been serialized)
     *                          NOT NULL
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     */
    public static Object getBeanInstance(ClassLoader classLoader, String beanName)
           throws BeanInstantiationException, BeanInitializationException {
              
        if (classLoader == null) {
            classLoader = ClassLoader.getSystemClassLoader();
        }
        
        Object bean = null;

        //
        // Search beanName as a class
        //
        try {
            bean = Class.forName(beanName, true, classLoader).newInstance();
            if (bean instanceof LazyInitBean) {
                ((LazyInitBean)bean).init();
            }
            return bean;
        }  catch (ClassNotFoundException e) {
            //
            // beanName does not represent a class!
            //
        } catch (BeanInitializationException e) {
            throw e;
        } catch (Exception e) {
            //
            // Other exceptions are not good!
            //
            throw new BeanInstantiationException( "Error instantiating "
                                                + beanName
                                                , e
                                                );
        }
        InputStream is = classLoader.getResourceAsStream(beanName);
        
        if (is == null) {
            throw new BeanInstantiationException( "Resource " 
                                                 + beanName 
                                                 + " not found"
                                                 + " with "
                                                 + classLoader
                                                 );
        }
        XMLDecoder d = new XMLDecoder(is);
        bean = d.readObject();
        d.close();
        
        if (bean instanceof LazyInitBean) {
            ((LazyInitBean)bean).init();
        }

        return bean;          
    }
    
    /**
     * Creates an instance of the given bean using the system class loader.
     *
     * @param       beanName    name of the bean (for now is also the file name,
     *                          without .ser, into which the bean has been serialized)
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     */
    public static Object getBeanInstance(String beanName)
           throws BeanInstantiationException, BeanInitializationException {
           return getBeanInstance(null, beanName);
    }
    

    /**
     * Creates an instance of the given bean. This version deserializes the bean
     * from the given file.
     *
     * @param       beanFile    the filename of the serialized file
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     */
    public static Object getBeanInstance(File beanFile)
           throws BeanInstantiationException, BeanInitializationException {
        try {
            Object bean = null;

            XMLDecoder e = new XMLDecoder(new FileInputStream(beanFile));
            bean = e.readObject();
            e.close();
            
            if (bean instanceof LazyInitBean) {
                ((LazyInitBean)bean).init();
            }
            
            /*
            ObjectInputStream osi = new ObjectInputStream(new FileInputStream(beanFile));
            bean = osi.readObject();
            osi.close();
            */
            return bean;
        } catch (IOException e) {
            String msg = "Bean creation ("
                       +          beanFile
                       +       ") failed: "
                       +     e.getMessage()
                       ;
            throw new BeanInstantiationException(msg, e);
//        } catch (BeanInstantiationException e) {
//            throw e;
        } catch (Exception e) {
            String msg = "Bean creation (" + beanFile + ") failed: " + e.getMessage();
            throw new BeanInstantiationException(msg, e);
        }
    }
    
    /**
     * Serialize the given object in the give file using XMLEncoder.
     *
     * @param obj the object to be serialized
     * @param file the file into wich serialize the object
     *
     * @throws BeanException in case of errors
     */
    public static void saveBeanInstance(Object obj, File file) 
    throws BeanException {
        XMLEncoder encoder = null;
        
        try {
            encoder = new XMLEncoder(new FileOutputStream(file));
            encoder.writeObject(obj);
            
            /**
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(file));
            oos.writeObject(obj);
            oos.close();
            */
        } catch (IOException e) {
            String msg = "Bean saving (" + file + ") failed: " + e.getMessage();
            throw new BeanException(msg, e);
        } finally {
            if (encoder != null) {
                encoder.close();
            }
        } 
    }
    
    /**
     * Creates a bean instance given the class name and serialize it in the
     * specified file. The object is serialized in XML.
     * <p>
     * Syntax:<br>
     * sync4j.framework.tools.beans.BeanFactory <i>class name</i> </i>file name</i>
     **/
    public static void main(String[] args) throws Exception {
        if (args.length != 2) {
            System.out.println("Syntax: sync4j.framework.tools.beans.BeanFactory <class name> <file name>");
            return;
        }
        
        saveBeanInstance(getBeanInstance(args[0]), new File(args[1]));
    }
}
