/*
 Copyright (c) 2001 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows 
    these conditions in the documentation and/or other materials 
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.  
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the 
 end-user documentation provided with the redistribution and/or in the 
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */
package sync4j.server.engine;

import java.util.List;
import java.util.Arrays;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Collection;

import java.util.logging.*;
import java.security.Principal;
import java.sql.Timestamp;

import sync4j.framework.core.ModificationCommand;
import sync4j.framework.core.StatusCode;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.*;

import sync4j.server.engine.SyncItemHelper;
import sync4j.server.engine.Sync4jOperationStatusOK;
import sync4j.server.engine.Sync4jOperationStatusError;
import sync4j.server.engine.Sync4jOperationStatusConflict;

import org.apache.commons.collections.ListUtils;

/**
 * This class represents a synchronization process.
 *
 * The base concrete implementation of a synchronization strategy. It implements
 * the <i>ConcreteStrategy</i> partecipant of the Strategy Pattern.
 * <p>
 * The synchronization process is implemented in this class as follows:
 * <p>
 * Given a set of sources A, B, C, D, etc, the synchronization process takes place 
 * between two sources at a time: A is first synchronzed with B, then AB with
 * C, then ABC with D and so on. <br>
 * The synchronization process is divided in three phases: preparation, synchronization,
 * finalization. These phases correspond to the methdos prepareSync, sync and
 * endSync; they are call-back methods called sequentially by a driver object,
 * usually a SyncMethod object. In this way the SyncMethod object has a chance 
 * to communicate with the external world before carry on the process. 
 * For example a SyncMethod object can show to the user the items that are going
 * to be synchronized, so that the user can let the process carry on or stop it. <br>
 * prepareSync returns an array of SyncOperation, in which each element represents
 * a particular synchronization action, ie. create an item in the source A,
 * delete the item X from the source B, etc. Sometime it is not possible decide 
 * what to do, thus a SyncConflict operation is used. A conflict must be solved
 * by something external the synchronization process, for instance by a user
 * action. Below is a table of all possible situations.
 * <pre>
 *
 * | -------- | --- | --- | --- | --- | --- |
 * | Source A |     |     |     |     |     |
 * |    /     |  N  |  D  |  U  |  S  |  X  |     N : item new
 * | Source B |     |     |     |     |     |     D : item deleted
 * | -------- | --- | --- | --- | --- | --- |     U : item updated
 * |       N  |  O  |  O  |  O  |  O  |  B  |     S : item synchronized/unchanged
 * | -------- | --- | --- | --- | --- | --- |     X : item not existing
 * |       D  |  O  |  X  |  O  |  X  |  X  |     O : conflict
 * | -------- | --- | --- | --- | --- | --- |     A : item A replaces item B
 * |       U  |  O  |  O  |  O  |  B  |  B  |     B : item B replaces item A
 * | -------- | --- | --- | --- | --- | --- |
 * |       S  |  O  |  X  |  A  |  =  |  B  |
 * | -------- | --- | --- | --- | --- | --- |
 * |       X  |  A  |  X  |  A  |  A  |  X  |
 * | -------- | --- | --- | --- | --- | --- |
 *
 * </pre>
 *
 * @author Stefano Fornari @ Funambol
 * 
 * @version  $Id: Sync4jStrategy.java,v 1.21 2003/12/15 17:06:05 luigiafassina Exp $
 */
public class Sync4jStrategy implements SyncStrategy, java.io.Serializable {
    
    // --------------------------------------------------------------- Constants
    
    public static String LOG_NAME = "sync4j.framework.engine";
    
    private transient static Logger log = Logger.getLogger(LOG_NAME);
    
    // -------------------------------------------------------------- Properties
    
    /**
     * The synchronization source 
     */
    private transient SyncSource[] sources = null; 
    public SyncSource[] getSources(){ return sources; }

    public void setSources(SyncSource[] sources){ this.sources = sources; }

    /**
     * The process' name
     */
    public String getName(){ return name; }

    private String name;
    public void setName(String name){ this.name = name; }
    
    // ------------------------------------------------------------ Constructors
    
    public Sync4jStrategy() {
    }
    
    public Sync4jStrategy(SyncSource[] syncSources) {
        sources = syncSources;
    }
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Preparation for the synchronization. If <i>sources</i> is not null,
     * the preparation works on the given sources. Otherwise it works on the 
     * sources as they were set in the constructor.
     *
     * @param sources the sources to be synchronized
     * @param principal the entity for which the synchronization is required
     *
     * @return an array of SyncOperation, one for each SyncItem that must be
     *         created/updated/deleted or in conflict.
     *
     * @see sync4j.framework.engine.SyncStrategy
     */
    public SyncOperation[] prepareSlowSync(SyncSource[] sources  , 
                                           Principal    principal) 
    throws SyncException {
        if (sources != null) {
            this.sources = sources;
        }
        
        List syncOperations = null           ,
             Am             = new ArrayList(),
             Bm             = new ArrayList(),
             AmBm           = null           ;

        Logger log = Logger.getLogger(LOG_NAME);
        
        if (log.isLoggable(Level.INFO)) {
            log.info("Preparing slow synchronization for " + principal + " ...");
        }
            
        Am.addAll(Arrays.asList(sources[1].getAllSyncItems(principal)));
        Bm.addAll(Arrays.asList(sources[0].getAllSyncItems(principal)));
        
        //
        // Because is a slow sync, items state must be reset to S
        //
        EngineHelper.resetState(Am);
        EngineHelper.resetState(Bm);
        
        AmBm = EngineHelper.intersect(Am, Bm);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Am: "   + Am  );
            log.finest("Bm: "   + Bm  );
            log.finest("AmBm: " + AmBm);
        }
        
        syncOperations = 
            checkSyncOperations(principal, Am, Bm, AmBm, new ArrayList(), new ArrayList());
        
        //
        // In the case of slow sync, delete operations are not needed
        //
        SyncOperation o = null;
        
        ArrayList ret = new ArrayList();
        
        Iterator i = syncOperations.iterator();
        while (i.hasNext()) {
            o = (SyncOperation)i.next();
            
            if (o.getOperation() != SyncOperation.DELETE) {
                ret.add(o);
            }
        }
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("operations: " + ret);
        }
        
        log.info("Preparation completed.");
        
        return (SyncOperation[])ret.toArray(new SyncOperationImpl[] {});
    }
    
    /**
     * Preparation for faset synchronization. If <i>sources</i> is not null,
     * the preparation operates on the given sources. Otherwise it works on the 
     * sources as they were set in the constructor.
     * <p>
     * Refer to the <a href="http://sync4j.sourceforge.net/web/project/architecture/sync4j-architecture.html">
     * architecture document</a> for details about the algoritm applied.
     *
     * @param sources the sources to be synchronized
     * @param principal the entity for which the synchronization is required
     *
     * @return an array of SyncOperation, one for each SyncItem that must be
     *         created/updated/deleted or in conflict.
     *
     * @see sync4j.framework.engine.SyncStrategy
     */
    public SyncOperation[] prepareFastSync(SyncSource[] sources  , 
                                           Principal    principal,
                                           Timestamp    since    ) 
    throws SyncException {
        if (sources != null) {
            this.sources = sources;
        }
        
        // ---------------------------------------------------------------------
        
        List Am    ,  // items modified in A
             Bm    ,  // items modified in B
             AmBm  ,  // Am intersect Bm
             AAmBm ,  // items unmodified in A, but modified in B
             AmBBm ;  // items unmodified in B, but modified in A
        
        ArrayList syncOperations = null;

        SyncItem[] newA       = null, newB       = null,
                   updatedA   = null, updatedB   = null,
                   deletedA   = null, deletedB   = null;
               
        Iterator i = null;

        if (log.isLoggable(Level.INFO)) {
            log.info( "Preparing fast synchronization for "
                    + principal
                    + " since "
                    + since
                    );
        }
        
        // ---------------------------------------------------------------------

        //
        // NOTE: simplified version - only two sources, the first one of which 
        // is the client
        //
        newA       = sources[1].getNewSyncItems      (principal, since);
        updatedA   = sources[1].getUpdatedSyncItems  (principal, since);
        deletedA   = sources[1].getDeletedSyncItems  (principal, since);

        newB       = sources[0].getNewSyncItems      (principal, since);
        updatedB   = sources[0].getUpdatedSyncItems  (principal, since);
        deletedB   = sources[0].getDeletedSyncItems  (principal, since);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("newA: " + Util.arrayToString(newA));
            log.finest("newB: " + Util.arrayToString(newB));

            log.finest("updatedA: " + Util.arrayToString(updatedA));
            log.finest("updatedB: " + Util.arrayToString(updatedB));

            log.finest("deletedA: " + Util.arrayToString(deletedA));
            log.finest("deletedB: " + Util.arrayToString(deletedB));
        }
        
        //
        // If any item from the client has not a corresponding mapping, the 
        // server source must be queried for the item, since the client item 
        // could be the same of an existing server item. In this case,  the old
        // unmapped item is replaced in Ma by the newly mapped item.
        //
        ArrayList newlyMappedItems = new ArrayList();
        
        fixMappedItems(newlyMappedItems, newA,     sources[0], principal);
        fixMappedItems(newlyMappedItems, updatedA, sources[0], principal);
        fixMappedItems(newlyMappedItems, deletedA, sources[0], principal);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Newly mapped items: " + newlyMappedItems);
        }
        
        Am = new ArrayList(); Bm = new ArrayList();
        Am.addAll(Arrays.asList(newA      )); 
        Am.addAll(Arrays.asList(updatedA  ));
        Am.addAll(Arrays.asList(deletedA  ));
        Bm.addAll(Arrays.asList(newB      )); 
        Bm.addAll(Arrays.asList(updatedB  ));
        Bm.addAll(Arrays.asList(deletedB  ));
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Am: " + Am);
            log.finest("Bm: " + Bm);
            log.finest("Am-Bm: " + ListUtils.subtract(Am, Bm));
            log.finest("Bm-Am: " + ListUtils.subtract(Bm, Am));
        }
        
        //
        // Now calculate subsets: AmBm, AmBBm, AAmBm
        //
        AmBm  = EngineHelper.intersect(Am, Bm);
        AmBBm = EngineHelper.buildAmBBm(ListUtils.subtract(Am, Bm), sources[0], principal);
        AAmBm = EngineHelper.buildAAmBm(ListUtils.subtract(Bm, Am), sources[1], principal);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("AmBm:  " + AmBm );
            log.finest("AmBBm: " + AmBBm);
            log.finest("AAmBm: " + AAmBm);
        }
        
        //
        // Ready for conflict detection!
        //
        syncOperations = 
            checkSyncOperations(principal, Am, Bm, AmBm, AmBBm, AAmBm);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("operations: " + syncOperations);
        }
        
        log.info("Preparation completed.");
        
        return (SyncOperation[])syncOperations.toArray(new SyncOperationImpl[] {});
    }

    /**
     * Implements Synchronizable.sync
     *
     * 
     */
    public SyncOperationStatus[] sync(SyncOperation[] syncOperations) {
        Logger log = Logger.getLogger(LOG_NAME);
        
        log.info("Synchronizing...");
        
        if ((syncOperations == null) || (syncOperations.length == 0)) {
            return new SyncOperationStatus[0];
        }
        
        ArrayList status = new ArrayList();
        
        SyncOperationStatus[] operationStatus = null;
        for(int i=0; i<syncOperations.length; ++i) {
            if (log.isLoggable(Level.FINE)) {
                log.fine("Executing " + syncOperations[i]);
            }
            
            //
            // execSyncOperation can return more than one status for one 
            // operation when more than one source are involved
            //
            operationStatus = execSyncOperation((SyncOperationImpl)syncOperations[i]);
            
            for (int j=0; j<operationStatus.length; ++j) {
                status.add(operationStatus[j]);
            }  // next j
        } // next i
        
        log.fine("Synchronization completed.");
        if (log.isLoggable(Level.FINE)) {
            log.fine("status: " + status);
        }

        return (SyncOperationStatus[])status.toArray(new SyncOperationStatus[0]);
    }
        
    /**
     * Implements Synchronizable.endSync
     *
     * @see beanblue.sync.event.Synchronizable
     */
    public void endSync() throws SyncException {
        Logger log = Logger.getLogger(LOG_NAME);
        
        log.info("Ending synchronization...");
        log.info("Synchronization ended.");
    }
    
    // -------------------------------------------------------- Protected methds

    /**
     * Checks the given SyncItem lists and creates the needed SyncOperations
     * following the rules described in the class descriprio and in the .
     *
     * @param principal who has requested the synchronization
     * @param AmBm
     * @param AmBBm
     * @param AAmBm
     *
     * @return an ArrayList containing all the collected sync operations
     */
    protected ArrayList checkSyncOperations(Principal principal,
                                            List      Am       ,
                                            List      Bm       ,
                                            List      AmBm     ,
                                            List      AmBBm    , 
                                            List      AAmBm    ) {
        SyncItemMapping mapping = null;
        SyncItem syncItemA = null, syncItemB = null;
        
        ArrayList all        = new ArrayList();
        ArrayList operations = new ArrayList();
        
        SyncItem notExistingSyncItem = null;
        
        // ---------------------------------------------------------------------
        
        all.addAll(AmBm );
        all.addAll(AmBBm);
        all.addAll(AAmBm);
        
        //
        // 1st check: items in both sources
        //
        Iterator i = all.iterator();
        while (i.hasNext()) {
            mapping = (SyncItemMapping)i.next();
            
            syncItemA = mapping.getSyncItemA();
            syncItemB = mapping.getSyncItemB();
            
            operations.add(
                checkSyncOperation(
                    principal, 
                    syncItemA,
                    syncItemB
                )
            );
            Am.remove(syncItemA);
            Bm.remove(syncItemB);
        }
        
        //
        // 2nd check: items in source A and not in source B
        //
        i = Am.iterator();
        while (i.hasNext()) {
           syncItemA = (SyncItem)i.next();
            
           operations.add(checkSyncOperation(principal, syncItemA, null));
        }  // next i 
        
        //
        // 3rd check: items in source B and not in source A
        //
        i = Bm.iterator();
        while (i.hasNext()) {
           syncItemB = (SyncItem)i.next();
            
           operations.add(checkSyncOperation(principal, null, syncItemB));
        }  // next i 
        
        return operations;
    }
    
    /**
     * Create a SyncOperation based on the state of the given SyncItem couple.
     *
     * @param principal     the entity that wnats to do the operation
     * @param syncItemA     the SyncItem of the source A - NULL means <i>not existing</i/
     * @param syncItemB     the SyncItem of the source B - NULL means <i>not existing</i/
     *
     * @return the SyncOperation object
     */
    protected SyncOperation checkSyncOperation(Principal principal,
                                               SyncItem  syncItemA, 
                                               SyncItem  syncItemB) {
        if (log.isLoggable(Level.FINE)) {
            log.fine( "check: syncItemA: "
                    + syncItemA
                    + " syncItemB: "
                    + syncItemB
                    );
        }
        
        if (syncItemA == null) {
            syncItemA = SyncItemImpl.getNotExistingSyncItem(null);
        }
        if (syncItemB == null) {
            syncItemB = SyncItemImpl.getNotExistingSyncItem(null);
        }
        
        switch (syncItemA.getState()) {
            //
            // NEW
            //
            case SyncItemState.NEW:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW) +
                                                String.valueOf(SyncItemState.NEW) );
                    case SyncItemState.UPDATED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW)     +
                                                String.valueOf(SyncItemState.UPDATED) );
                   case SyncItemState.DELETED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW)     +
                                                String.valueOf(SyncItemState.DELETED) );
                   case SyncItemState.SYNCHRONIZED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW)       +
                                                String.valueOf(SyncItemState.SYNCHRONIZED) );
                   case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, false, true);
                }  // end inner switch
            
            //
            // DELETED
            //
            case SyncItemState.DELETED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.DELETED) +
                                                String.valueOf(SyncItemState.NEW)     );
                    case SyncItemState.UPDATED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.DELETED) +
                                                String.valueOf(SyncItemState.UPDATED) );
                   
                   case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.DELETE, false, true);
                        
                   case SyncItemState.DELETED:
                   case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemB, syncItemA, SyncOperation.NOP, false, false);
                }  // end inner switch
            
            //
            // UPDATED
            //
            case SyncItemState.UPDATED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.NEW)     );
                    case SyncItemState.UPDATED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.UPDATED) );
                    case SyncItemState.DELETED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.DELETED) );
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.UPDATE, false, true);
                    case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, false, true);
                }  // end inner switch
            
           //
           // SYNCHRONIZED
           //
           case SyncItemState.SYNCHRONIZED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.SYNCHRONIZED) +
                                                String.valueOf(SyncItemState.NEW)       );
                    case SyncItemState.UPDATED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.UPDATE, true, false);
                   case SyncItemState.DELETED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.DELETE, true, false);
                   case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
                   case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, false, true);
                }  // end inner switch
                
           //
           // NOTEXISITNG
           //
           case SyncItemState.NOT_EXISTING:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                    case SyncItemState.UPDATED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, true, false);
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, true, false);
                    case SyncItemState.NOT_EXISTING:
                    case SyncItemState.DELETED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
                }  // end inner switch
        }  // end switch
        
        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
    }
    
    /**
     * Executes the given SyncOperation. Note that conflicts are ignored!
     * <p>
     * Note also that the number of status returned is equal to the number of
     * sources affected by the operation (0 for NOP, 1 for NEW and UPDATE and 
     * 1 or 2 for DELETE).
     *
     * @param operation     the SyncOperation to execute
     *
     * @return an array of <i>SyncOperationStatus</i> objects representing the 
     *         status of the executed operation. For instance, in case of error,
     *         <i>SyncOperation.error</i> will be set to the catched exception.
     */
    protected SyncOperationStatus[] execSyncOperation(SyncOperationImpl operation) {
                SyncItem syncItemA = operation.getSyncItemA(), 
                 syncItemB = operation.getSyncItemB();
        
        Principal owner = operation.getOwner();
        SyncOperationStatus[] status = null;
        ModificationCommand cmd = null;
        
        int size = 0, s = 0;
        
        switch (operation.getOperation()) {
            case SyncOperation.NEW:
            case SyncOperation.UPDATE:
                status = new SyncOperationStatus[1];
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    try {
                        syncItemA = sources[1].setSyncItem(owner, syncItemB);
                        operation.setSyncItemA(syncItemA);
                        status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                    }
                } else if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    syncItemB.setProperty(syncItemA.getProperty(SyncItem.PROPERTY_TIMESTAMP)); // this contains the
                                                                                               // current sync
                    try {
                        syncItemB = sources[0].setSyncItem(owner, syncItemA);
                        operation.setSyncItemB(syncItemB);
                        status[0] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                    }
                }
                break;

            case SyncOperation.DELETE:
                //
                // How many status we need? One if we have to delete the item
                // from only one source, two if we have to delete it from both
                // sources
                //
                size = 1;
                if (operation.isAOperation() && operation.isBOperation()) {
                    size = 2;
                }
                status = new SyncOperationStatus[size];
                                
                s = 0;
                if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    syncItemB.setProperty(syncItemA.getProperty(SyncItem.PROPERTY_TIMESTAMP)); // this contains the
                                                                                               // timestamp of the 
                                                                                               // current sync
                    try {
                        sources[0].removeSyncItem(owner, syncItemB);
                        status[s++] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[s++] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                    }
                } 
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                                       
                    try {
                        sources[1].removeSyncItem(owner, syncItemA);
                        status[s++] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[s++] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                    }
                }
                break;

            case SyncOperation.NOP:
                //
                // A NOP operation can be due by one of the following conditions
                // (see checkOperation(...)):
                //
                // 1. both items are flagged as "Deleted"
                // 2. both items are flagged "Synchronized"
                // 3. itemA is "Not existing" and itemB is "Not existing" or "Deleted"
                //
                // In case 1. we want a status is returned, thought no real 
                // action is performed.
                //
                s = 0;
                if (operation.isAOperation()) ++s;
                if (operation.isBOperation()) ++s;
                
                status = new SyncOperationStatus[size];
                
                if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    status[s++] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                }
                
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    status[s++] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                }
                break;

            case SyncOperation.CONFLICT:
                //
                // The conflic is solved with data from server
                //
                try {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    syncItemA = sources[1].setSyncItem(owner, syncItemB);
                    operation.setSyncItemA(syncItemA);
                    operation.setAOperation(true);
                    
                    status = new SyncOperationStatus[1];
                    status[0] = new Sync4jOperationStatusConflict(
                                    operation, 
                                    sources[1], 
                                    cmd, 
                                    StatusCode.CONFLICT_RESOLVED_WITH_SERVER_DATA
                                );
                } catch (SyncException e) {
                    log.severe("Error executing sync operation: " + e.getMessage());
                    log.throwing(getClass().getName(), "execSyncOperation", e);
                    status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                }
                
                break;
        }  // end switch
        
        return status;
    }
    
    // --------------------------------------------------------- Private methods
    
    /**
     * If any item from the client has not a corresponding mapping, the 
     * server source must be queried for the item, since the client item 
     * could be the same of an existing server item. In this case,  the old
     * unmapped item is replaced in Ma by the newly mapped item.
     *
     * @param newlyMappedItems the collection that contains all the checked items
     * @param source the source that has to be queried for twins
     * @param syncItems the items to be searched if not mapped
     * @param principal the principal items are belonging to
     *
     */
    private void fixMappedItems(Collection newlyMappedItems,
                                SyncItem[] syncItems       , 
                                SyncSource source          ,
                                Principal  principal       ) {
                               
        SyncItem itemB = null;
        
        for (int i = 0; ((syncItems != null) && (i<syncItems.length)); ++i) {
            itemB = null;
            if (!syncItems[i].isMapped()) {
                try {
                    itemB = source.getSyncItemFromTwin(principal, syncItems[i]);
                } catch (SyncSourceException e) {
                      String msg = "Error retrieving the twin item of "
                                 + syncItems[i].getKey()
                                 + " from source "
                                 + source
                                 + ": "
                                 + e.getMessage();
                      log.severe(msg);
                      log.throwing(getClass().getName(), "fixSyncMapping", e);
                }
                if (itemB != null) {
                    SyncItemMapping mapping = new SyncItemMapping(itemB.getKey());
                    mapping.setMapping(syncItems[i], itemB);
                    newlyMappedItems.add(mapping);
                    syncItems[i] = SyncItemHelper.newMappedSyncItem(itemB.getKey(), syncItems[i]);
                }
            }
        }  // next i

    }
}
