/*
 Copyright (c) 2001-2002-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE. 
 */

package sync4j.server.store;

import java.util.ArrayList;

import java.sql.*;

import sync4j.server.engine.Sync4jSource;

import sync4j.framework.tools.DBTools;
import sync4j.framework.server.store.*;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.security.Sync4jPrincipal;

/**
 * This is the store for information regarding the Sync4j engine such as users,
 * principals, ecc. Currently it persistes the following classes:
 *
 * <ul>
 * <li>sync4j.framework.security.Sync4jPrincipal</li>
 * </ul>
 *
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: EnginePersistentStore.java,v 1.14 2004/01/07 10:39:39 stefano_fornari Exp $
 *
 */
public class EnginePersistentStore 
extends BasePersistentStore
implements PersistentStore, java.io.Serializable {
    
    // --------------------------------------------------------------- Constants

    private static final String OPT_INSERT     = "INSERT";
    private static final String OPT_UPDATE     = "UPDATE";
    
    public static final int SQL_INSERT_PRINCIPAL      = 0;
    public static final int SQL_GET_PRINCIPAL         = 1;
    public static final int SQL_SELECT_PRINCIPAL      = 2;
    public static final int SQL_UPDATE_PRINCIPAL      = 3;
    public static final int SQL_SELECT_ALL_PRINCIPALS = 4;
    public static final int SQL_INSERT_SOURCE         = 5;
    public static final int SQL_GET_SOURCE            = 6;
    public static final int SQL_UPDATE_SOURCE         = 7;
    public static final int SQL_SELECT_ALL_SOURCES    = 8;
    
    public static final int SQL_SELECT_ALL_DEVICES    = 9;
    public static final int SQL_GET_DEVICE            = 10;
    public static final int SQL_INSERT_DEVICE         = 11;
    public static final int SQL_UPDATE_DEVICE         = 12;
    public static final int SQL_DELETE_DEVICE         = 13;
    public static final int SQL_DELETE_DEVICE_PRINCIPAL = 14;
    public static final int SQL_DELETE_PRINCIPAL      = 15;
    public static final int SQL_DELETE_CLIENT_MAPPING = 16;
    public static final int SQL_DELETE_LAST_SYNC      = 17;    
    public static final int SQL_GET_COUNTER           = 18;
    public static final int SQL_UPDATE_COUNTER        = 19;
    public static final int SQL_DELETE_SOURCE         = 20;
    public static final int SQL_INSERT_SYNCSOURCE     = 21;
    public static final int SQL_UPDATE_SYNCSOURCE     = 22;
    
    public static final String SQL_COUNT_DEVICES = 
        "select count(*) as devices from sync4j_device ";
    public static final String SQL_COUNT_PRINCIPALS =
        "select count(*) as principals from sync4j_principal ";
    
    // -------------------------------------------------------------- Properties    
    
    private String[] sql = null;
    
    public void setSql(String[] sql) {
        this.sql = sql;
    }
    
    public String[] getSql() {
        return this.sql;
    }
    
    // ------------------------------------------------------------ Private data
    private static final String SEARCH_COUNT_DEVICES    = "SCD";
    private static final String SEARCH_COUNT_PRINCIPALS = "SCP";
    
    // ------------------------------------------------------------ Constructors
    // ---------------------------------------------------------- Public methods
    
    public boolean delete(Object o)
    throws PersistentStoreException {
        if (o instanceof Sync4jDevice) {
            deleteDevice((Sync4jDevice) o);
            return true;
        } else if (o instanceof Sync4jPrincipal) {
            deletePrincipal((Sync4jPrincipal) o);
            return true;
        } else if (o instanceof Sync4jSource) {
            deleteSource((Sync4jSource) o);
            return true;
        }
        
        return false;
    }
        
    public boolean store(String id, Object o, String operation) throws PersistentStoreException {
        if (o instanceof Sync4jSource) {
            if (operation.equals(OPT_INSERT)) {
                insertSyncSource(id, (Sync4jSource)o);
            } else {
                updateSyncSource(id, (Sync4jSource)o);
            }
            return true;
        }
        return false;
    }
    
    public boolean store(Object o)
    throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            storePrincipal((Sync4jPrincipal) o);
            return true;
        } else if (o instanceof Sync4jSource) {
            storeSource((Sync4jSource) o);
            return true;
        } else if (o instanceof Sync4jDevice) {
            storeDevice((Sync4jDevice) o);
            return true;
        }
        
        return false;
    }
    
    public boolean read(Object o)
    throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            readPrincipal((Sync4jPrincipal) o);
            return true;
        } else if (o instanceof Sync4jSource) {
            readSource((Sync4jSource) o);
            return true;
        } else if (o instanceof Sync4jDevice) {
            readDevice((Sync4jDevice) o);
            return true;
        }
        
        return false;
    }
    
    public Object[] read(Class objClass) throws PersistentStoreException {
        if (objClass.equals(Sync4jSource.class)) {
            return readAllSources();
        }
        
        return null;
    }

    /**
     * Read all informations
     * @param object whichever object Sync
     * @param clause condition where for select
     * 
     * @return Object[] array of object
     */
    public Object[] read(Object o, Clause clause) throws PersistentStoreException {        
        if (o instanceof Sync4jPrincipal) {
            return readAllPrincipals(clause);
        } else if (o instanceof Sync4jDevice) {
            return readAllDevices(clause);
        } else if (o instanceof String) {
            if (o.equals(SEARCH_COUNT_DEVICES)) {
                return readCountDevices(clause);
            } else if (o.equals(SEARCH_COUNT_PRINCIPALS)) {
                return readCountPrincipals(clause);
            }
        }
        
        return null;
    }
    
    // --------------------------------------------------------- Private methods
    
    private void storePrincipal(Sync4jPrincipal p)
    throws PersistentStoreException {        
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        int n = 0;
        
        try {
            conn = dataSource.getConnection();
            
            if (p.getId() != null && !p.getId().equals("")) {
                        
                stmt = conn.prepareStatement(sql[SQL_UPDATE_PRINCIPAL]);

                stmt.setString(1, p.getUsername());
                stmt.setString(2, p.getDeviceId());
                stmt.setString(3, p.getId()      );

                n = stmt.executeUpdate();
                
                stmt.close();
            } else {
                //check if the principal already exist: verify username-deviceid

                stmt = conn.prepareStatement(sql[SQL_SELECT_PRINCIPAL]);   
                
                stmt.setString(1, p.getUsername());
                stmt.setString(2, p.getDeviceId());
                
                rs = stmt.executeQuery();
                
                if (rs.next() == false) {
                    
                    n = 0;
                    
                } else {
            
                    n = 1;
                    
                    p.setId       (rs.getString(1));
                    p.setUsername (rs.getString(2));
                    p.setDeviceId (rs.getString(3));
                    p.setEmail    (rs.getString(4));
                    p.setFirstName(rs.getString(5));
                    p.setLastName (rs.getString(6));
                }

                stmt.close();
                rs.close();
            }
            
            if (n == 0) {
                //
                // The first time!!!
                //
                
                int principalId = this.readCounter("principal");
                p.setId("" + principalId);
                
                stmt = conn.prepareStatement(sql[SQL_INSERT_PRINCIPAL]);
                
                stmt.setString(1, p.getId()      );
                stmt.setString(2, p.getUsername());
                stmt.setString(3, p.getDeviceId());
                
                stmt.executeUpdate();
            }
        } catch (SQLException e) {
            throw new PersistentStoreException("Error storing principal " + p, e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }

    /**
     * Read the principal from the data store. If <i>getId()</i> returns null, 
     * it tries to read the principal from username/device, otherwise throws id
     * is used for the lookup.
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void readPrincipal(Sync4jPrincipal p)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
                
        try {
            conn = dataSource.getConnection();
            
            if (p.getId() == null) {
                
                stmt = conn.prepareStatement(sql[SQL_SELECT_PRINCIPAL]);   
                
                stmt.setString(1, p.getUsername());
                stmt.setString(2, p.getDeviceId());
            } else {
                
                stmt = conn.prepareStatement(sql[SQL_GET_PRINCIPAL]);
                
                stmt.setString(1, p.getId());
            }

            rs = stmt.executeQuery();
            
            if (rs.next() == false) {
                throw new NotFoundException("Principal not found for "
                + p.toString()
                );
            }
            
            p.setId       (rs.getString(1));
            p.setUsername (rs.getString(2));
            p.setDeviceId (rs.getString(3));
            p.setEmail    (rs.getString(4));
            p.setFirstName(rs.getString(5));
            p.setLastName (rs.getString(6));
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading principal " + p, e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
    
    /**
     * Read all principals
     *
     * @return an array with the principals (empty if no objects are found)
     *
     * @throws PersistentException in case of error reading the data store
     */
    private Object[] readAllPrincipals(Clause clause)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
          
        ArrayList ret = new ArrayList();
        
        try {
            conn = dataSource.getConnection();
            
            PreparedWhere where = clause.getPreparedWhere();
            
            String query = sql[SQL_SELECT_ALL_PRINCIPALS];
            if (where.sql.length() > 0){
                query += " where " + where.sql;
            }
            
            stmt = conn.prepareStatement(query);
            
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }

            rs = stmt.executeQuery();
            
            while (rs.next()) {
                ret.add(
                    new Sync4jPrincipal(
                        rs.getString(1),
                        rs.getString(2),
                        rs.getString(3)
                    )
                );
            }
            
            return ret.toArray(new Sync4jPrincipal[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading principals", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
    
    
    /**
     * Delete the principal from the data store. 
     *
     * @param p the principal
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void deletePrincipal(Sync4jPrincipal p)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
                        
            stmt = conn.prepareStatement(sql[SQL_DELETE_PRINCIPAL]);
            
            stmt.setString(1, p.getId());

            stmt.executeUpdate();
            
            //delete client_mapping
            stmt.close();
            stmt = conn.prepareStatement(sql[SQL_DELETE_CLIENT_MAPPING]);            
            stmt.setString(1, p.getId());

            stmt.executeUpdate();

            stmt.close();
            stmt = conn.prepareStatement(sql[SQL_DELETE_LAST_SYNC]);
            stmt.setString(1, p.getId());

            stmt.executeUpdate();

        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the principal " + p, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    /**
     * Select the number of principals that satisfy the conditions specified in input
     * 
     * @return int number of principals
     */
    public Object[] readCountPrincipals(Clause clause) throws PersistentStoreException {       
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        int n = 0;
        
        try {
            conn = dataSource.getConnection();
            
            PreparedWhere where = clause.getPreparedWhere();
                                    
            String query = SQL_COUNT_PRINCIPALS;
            
            if (where.sql.length()>0) {
                query += " where " + where.sql;
            }
            
            stmt = conn.prepareStatement(query);
            
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }

            rs = stmt.executeQuery();
            while (rs.next()) {
                n = rs.getInt(1);
            }
            
            String[] s = new String[1];
            s[0] = "" + n;
            return s;
            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading count principals ", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
    
    private void storeSource(Sync4jSource s)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
                        
            stmt = conn.prepareStatement(sql[SQL_UPDATE_SOURCE]);
            
            stmt.setString(1, s.getConfig());
            stmt.setString(2, s.getUri()   );
                        
            int n = stmt.executeUpdate();
            
            if (n == 0) {
                //
                // The first time!!!
                //
                stmt.close();
                
                stmt = conn.prepareStatement(sql[SQL_INSERT_SOURCE]);
                
                stmt.setString(1, s.getUri()   );
                stmt.setString(2, s.getConfig());
                
                stmt.executeUpdate();
            }
        } catch (SQLException e) {
            throw new PersistentStoreException("Error storing the source " + s, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }

    /**
     * Read the source from the data store. 
     *
     * @param s the source
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void readSource(Sync4jSource s)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
                
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_GET_SOURCE]);
            stmt.setString(1, s.getUri());
            
            rs = stmt.executeQuery();
            
            if (rs.next() == false) {
                throw new NotFoundException("Source not found for "
                + s.getUri()
                );
            }
            
            s.setName  (rs.getString(1));
            s.setConfig(rs.getString(2));
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading the source " + s, e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
    
    /**
     * Read all principals
     *
     * @return an array with the principals (empty if no objects are found)
     *
     * @throws PersistentException in case of error reading the data store
     */
    private Object[] readAllSources()
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
          
        ArrayList ret = new ArrayList();
        
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_SELECT_ALL_SOURCES]);

            rs = stmt.executeQuery();
            
            while (rs.next()) {
                ret.add(
                    new Sync4jSource(
                        rs.getString(1),
                        rs.getString(2)
                    )
                );
            }
            
            return ret.toArray(new Sync4jSource[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading sources", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }

    
    
    /**
     * Read the device from the data store. 
     *
     * @param d the device
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void readDevice(Sync4jDevice d)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
                
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_GET_DEVICE]);
            stmt.setString(1, d.getDeviceId());

            rs = stmt.executeQuery();
            
            if (rs.next() == false) {
                throw new NotFoundException("Device not found for "
                + d.getDeviceId()
                );
            }
            
            d.setDescription(rs.getString(1));
            d.setType       (rs.getString(2));
            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading the device " + d, e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
    
    /**
     * Insert or Update the device from the data store. 
     *
     * @param d the device
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void storeDevice(Sync4jDevice d)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        int n = 0;
        
        try {
            conn = dataSource.getConnection();

            if (d.getDeviceId() != null && !d.getDeviceId().equals("")) {
                stmt = conn.prepareStatement(sql[SQL_UPDATE_DEVICE]);

                stmt.setString(1, d.getDescription());
                stmt.setString(2, d.getType());
                stmt.setString(3, d.getDeviceId());

                n = stmt.executeUpdate();
                stmt.close();
            }
            
            if (n == 0) {
                //
                // The first time!!!
                //
                if (d.getDeviceId() == null || d.getDeviceId().equals("")) {
                    int deviceId = this.readCounter("device");
                    d.setDeviceId("" + deviceId);
                }
                
                stmt = conn.prepareStatement(sql[SQL_INSERT_DEVICE]);
                
                stmt.setString(1, d.getDeviceId());
                stmt.setString(2, d.getDescription());
                stmt.setString(3, d.getType());
                
                stmt.executeUpdate();
            }
        } catch (SQLException e) {
            throw new PersistentStoreException("Error storing the device " + d, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    /**
     * Delete the device from the data store. 
     *
     * @param d the device
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void deleteDevice(Sync4jDevice d)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
                        
            stmt = conn.prepareStatement(sql[SQL_DELETE_DEVICE]);
            
            stmt.setString(1, d.getDeviceId());

            stmt.executeUpdate();
            
            stmt.close();
            
            //select and delete all principal for rhis device
            WhereClause c = new WhereClause("device", new String[] {d.getDeviceId()}, WhereClause.OPT_EQ, true);
            Sync4jPrincipal[] sp = (Sync4jPrincipal[])readAllPrincipals(c);

            for (int i=0; (sp != null) && i<sp.length; i++) {
                Sync4jPrincipal syncPrincipal = sp[i];                
                this.deletePrincipal(syncPrincipal);                
            }
            
            stmt = conn.prepareStatement(sql[SQL_DELETE_DEVICE_PRINCIPAL]);
            
            stmt.setString(1, d.getDeviceId());
            
            stmt.executeUpdate();

            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the device " + d, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    /**
     * Select the number of devices that satisfy the conditions specified in input
     * 
     * @return int number of devices
     */
    public Object[] readCountDevices(Clause clause) throws PersistentStoreException {       
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        int n = 0;
        
        try {
            conn = dataSource.getConnection();
            
            PreparedWhere where = clause.getPreparedWhere();
                                    
            String query = SQL_COUNT_DEVICES;
            if (where.sql.length()>0) {
                query += " where " + where.sql;
            }
            
            stmt = conn.prepareStatement(query);
            
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }

            rs = stmt.executeQuery();
            while (rs.next()) {
                n = rs.getInt(1);
            }
            
            String[] s = new String[1];
            s[0] = "" + n;
            return s;
            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading count devices ", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
    
    /**
     * Read the counter of the specific id in order to use it like primary key
     * in case of inserting a new record into data store.
     *
     * @param idSpace the idSpace
     *
     * @throws PersistentException in case of error reading the data store
     */
    private int readCounter(String idSpace)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        int counter = 0;
                
        try {
            conn = dataSource.getConnection();
            
            java.sql.DatabaseMetaData dmd = conn.getMetaData();
           
            if ( dmd.supportsSelectForUpdate() ) {
                stmt = conn.prepareStatement(sql[SQL_GET_COUNTER]);
            } else {
                stmt = conn.prepareStatement("select counter from sync4j_id where idspace=? ");
            }
            
            stmt.setString(1, idSpace);

            rs = stmt.executeQuery();
            
            if (rs.next() == false) {
                throw new NotFoundException("Counter not found for "
                + idSpace
                );
            }
            
            counter = rs.getInt(1) + 1;
            
            stmt.close();
            
            stmt = conn.prepareStatement(sql[SQL_UPDATE_COUNTER]);
            stmt.setInt(1, counter);
            stmt.setString(2, idSpace);

            stmt.executeUpdate();
            
        } catch (SQLException e) {
            e.printStackTrace();
            throw new PersistentStoreException("Error reading the counter " + counter, e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
        return counter;
    }

    private Object[] readAllDevices(Clause clause)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
          
        ArrayList ret = new ArrayList();
        
        try {
            conn = dataSource.getConnection();
            
            PreparedWhere where = clause.getPreparedWhere();
            
            String query = sql[SQL_SELECT_ALL_DEVICES];
            if (where.sql.length() > 0) {
                query += " where " + where.sql;
            }
            
            stmt = conn.prepareStatement(query);
            
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }

            rs = stmt.executeQuery();
            
            while (rs.next()) {
                ret.add(
                    new Sync4jDevice(
                        rs.getString(1),
                        rs.getString(2),
                        rs.getString(3)
                    )
                );
            }
            
            return ret.toArray(new Sync4jDevice[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading devices", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }
   
    private void deleteSource(Sync4jSource s)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();

            stmt = conn.prepareStatement(sql[SQL_DELETE_SOURCE]);
            
            stmt.setString(1, s.getUri());

            stmt.executeUpdate();
            
            stmt.close();
            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the source " + s, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    /**
     * Insert a new SyncSource
     */
    private void insertSyncSource(String sourcetypeid, Sync4jSource ss)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
           
            stmt = conn.prepareStatement(sql[SQL_INSERT_SYNCSOURCE]);

            stmt.setString(1, ss.getUri());
            stmt.setString(2, ss.getConfig());
            stmt.setString(3, ss.getSourceName());
            stmt.setString(4, sourcetypeid);

            stmt.executeUpdate();
            stmt.close();
            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error storing the syncsource " + ss, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }

    /**
     * Update an existing SyncSource
     */
    private void updateSyncSource(String sourcetypeid, Sync4jSource ss)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_UPDATE_SYNCSOURCE]);
            
            stmt.setString(1, ss.getConfig());
            stmt.setString(2, ss.getSourceName());
            stmt.setString(3, ss.getSourceTypeId());
            stmt.setString(4, ss.getUri()   );
                        
            stmt.executeUpdate();
            stmt.close();
            
        } catch (SQLException e) {
            throw new PersistentStoreException("Error updating the syncsource " + ss, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }

}
