/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.server.syncbean;

import javax.ejb.*;
import javax.naming.InitialContext;

import java.io.*;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.net.URL;
import java.net.MalformedURLException;

import sync4j.framework.core.*;
import sync4j.framework.server.SyncResponse;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncEngineFactory;
import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigClassLoader;
import sync4j.framework.config.ConfigurationException;

import sync4j.server.syncbean.session.SimpleSessionHandler;
import sync4j.framework.protocol.ProtocolException;

import sync4j.framework.core.Sync4jException;
import sync4j.framework.core.RepresentationException;
import sync4j.framework.server.session.SessionExpiredException;
import sync4j.framework.server.error.*;
import sync4j.framework.server.store.PersistentStoreException;
import sync4j.framework.tools.WBXMLTools;

import sync4j.framework.tools.beans.BeanFactory;
import sync4j.framework.tools.beans.BeanException;
import sync4j.framework.engine.pipeline.PipelineManager;
import sync4j.framework.engine.pipeline.MessageProcessingContext;

/**
 *  This is the session enterprise java bean that handles a synchronization
 *  request. It is designed to be a stateful session bean so that the state
 *  and the session management are on charge of the application server.
 *  <p>
 *  This server accepts synchronization requests addressed to the hostname
 *  indicated by the configuration property pointed by {CONFIG_SERVER_URI} (see
 *  Sync4j.properties).
 *  <p>
 *  SyncBean uses two environment properties:
 *  <ul>
 *    <li><i>{ENV_ENGINE_FACTORY_NAME}</i> points to the bean to be used as a
 *        factory for the synchronization engine</li>
 *    <li><i>{ENV_SERVER_CONFIG_URI}</i> contains the URI of the server
 *        configuration. It can be a file or any other kind of URI that returns
 *        a stream readable from the <i>java.util.Properties</i> class.
 *  </ul>
 *
 *  LOG NAME: sync4j.server
 *
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: SyncBean.java,v 1.43 2003/12/29 11:57:47 luigiafassina Exp $
 *
 */
public class SyncBean
        implements javax.ejb.SessionBean
{
    // ------------------------------------------------------- Private constants

    private static final String ENV_ENGINE_FACTORY_NAME
        = "java:comp/env/syncengine/factory/bean";
    private static final String ENV_SERVER_CONFIG_URI
        = "java:comp/env/server/config_uri";
    private static final String ENV_SERVER_CONFIG_PATH
        = "java:comp/env/server/config_path";

    private static final String CONFIG_SERVER_URI      = "server.uri"     ;
    private static final String CONFIG_ENGINE_PIPELINE = "engine.pipeline";

    // ------------------------------------------------------------ Private data

    private transient Logger log = null;

    private SessionContext sessionContext       = null;

    private SyncEngineFactory syncEngineFactory = null;

    private SimpleSessionHandler sessionHandler = null;

    private Configuration config                = null;

    private String sessionId                    = null;

    private PipelineManager pipelineManager     = null;
    
    private MessageProcessingContext mpc        = null;

    // ------------------------------------------------------------- EJB methods

    public void ejbCreate(String sessionId)
            throws CreateException {

        log = Sync4jLogger.getLogger("server");

        //
        // Create and initialize the SyncEngineFactory
        //
        try {
            loadConfiguration();
            getSyncEngineFactory();
        } catch (Sync4jException e) {
            log.throwing("SyncBean", "ejbCreate", e);
            throw new CreateException( "Error "
                                     + e.getClass().getName()
                                     + " creating the SyncBean: "
                                     + e.getMessage()
                                     );
        }

        this.sessionId = sessionId;

        sessionHandler = new SimpleSessionHandler();
        sessionHandler.setSyncEngineFactory(syncEngineFactory);

        pipelineManager = (PipelineManager)config.getBeanInstance(CONFIG_ENGINE_PIPELINE);
        mpc = new MessageProcessingContext();

    }

    /**
     * When the session terminates this object is removed. This can happen
     * because the synchronization process has been completed or because the
     * session has expired. If the session completes its job, the synchro-
     * nization can be committed.
     *
     */
    public void ejbRemove() {
        log = null;
    }

    public void ejbActivate()
    {
        log = Sync4jLogger.getLogger();
    }

    public void ejbPassivate()
    {
        log = null;
    }

    public void setSessionContext (SessionContext sessionContext)
    throws EJBException {
        this.sessionContext = sessionContext;
    }

    /**
     *
     * process the incoming sync message
     *
     * @param msg must be non-null
     * @param mimeType may be null.
     *
     */
    public SyncResponse processMessage( final byte[] msg, final String mimeType)
    throws ServerException {
        SyncBean.Sync4jResponse response = null;

        if (msg == null) {
            String err = "msg is null!";
            log.severe(err);
            throw new ServerException(err);
        }

        if (mimeType == null) {
            String err = "mimeType is null!";
            log.severe(err);
            throw new ServerException(err);
        }

        log.finest("mimeType: " + mimeType);

        Message msgIn         = null;
        Message returnMessage = null;
        if (Constants.MIMETYPE_SYNCML_WBXML.equals(mimeType)) {
            // convert WBXML to XML, and then pass to processXMLMessage
            // todo: change this so that it mirrors processXMLMessage
            //  instead of converting to XML then calling processXMLMessage
            try {
                String xmlMsg = WBXMLTools.wbxmlToXml(msg);

                if (log.isLoggable(Level.FINE)) {
                    log.fine("Calling input pipeline");
                }
                msgIn = new Message(xmlMsg);
                pipelineManager.preProcessMessage(mpc, msgIn);

                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Processing XML message: " + msgIn.toString());
                }

                returnMessage = processXMLMessage(msgIn.toString());

                if (log.isLoggable(Level.FINEST) && returnMessage!=null) {
                    log.finest("Returning: " + returnMessage.toXML());
                }

            } catch (Sync4jException e) {
                String err = "Error processing WBXML message: " + e.getMessage();
                log.severe(err);
                throw new ServerException(err, e);
            }
        } else if (Constants.MIMETYPE_SYNCML_XML.equals(mimeType)) {
            try {
                String xmlMsg = new String(msg);

                if (log.isLoggable(Level.FINE)) {
                    log.fine("Calling input pipeline");
                }
                msgIn = new Message(xmlMsg);
                pipelineManager.preProcessMessage(mpc, msgIn);

                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Processing XML message: " + msgIn.toString());
                }

                returnMessage = processXMLMessage(msgIn.toString());

                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Returning: " + returnMessage.toXML());
                }
            } catch (Sync4jException e) {
                String err = "Error processing XML message: " + e.getMessage();
                log.severe(err);
                throw new ServerException(err, e);
            }
        } else {
            //
            // its an unsupported MIME type
            //
            String err = "Unknown mime type:  " + mimeType;

            log.severe(err);

            throw new NotImplementedException(err);
        }

        // todo : don't hardcode this:
        final String resultMimeType = mimeType;

        if (log.isLoggable(Level.FINE)) {
            log.fine("Calling output pipeline");
        }

        pipelineManager.postProcessMessage(mpc, returnMessage);

        if (log.isLoggable(Level.FINE)) {
            log.fine( "The synchronization is "
                    + ((returnMessage.isLastMessage()) ? "" : "not ")
                    + "complete"
                    );
        }

        response =
            new SyncBean.Sync4jResponse( returnMessage ,
                                         resultMimeType);

        return (SyncResponse) response;
    }

    /**
     * Used to process a status information as needed by the client object (i.e.
     * errors or success).
     *
     * @param statusCode the status code
     * @param statusMessage additional descriptive message
     *
     * @see sync4j.framework.core.StatusCode for valid status codes.
     */
    public SyncResponse processStatusCode(int statusCode, String info){
        if (statusCode != StatusCode.OK) {
            sessionHandler.abort(statusCode);
        }
        return null;
    }

    // --------------------------------------------------------- private methods

    /**
     * Processes the given SyncML XML message. See the class description for
     * more information.
     *
     * @param message the message to be processed
     *
     * @return the response message
     *
     * @throws ProtocolException
     */
    private Message processXMLMessage(String msg) throws ServerException {
        Message message = null;

        try {
            message = new Message(msg);
        } catch (RepresentationException e) {
            throw new ServerException("The message is not a valid SyncML message", e);
        }

        try {
            try {
                checkMessage(message);
                return sessionHandler.processMessage(message);
            } catch (InvalidCredentialsException e) {
                return sessionHandler.processError(message, e);
            } catch (ServerException e) {
                return sessionHandler.processError(message, e);
            } catch (ProtocolException e) {
                return sessionHandler.processError(message, new BadRequestException(e.getMessage()));
            }
        } catch (Sync4jException e1) {
            //
            // This can be due only to processError
            //
            throw new ServerException(e1);
        }
    }

    /**
     * Returns the <i>Sync4jEngineFactory</i> object that a <i>SessionHandler</i>
     * should use to create a <i>SyncEngine</i>. The method dynamically loads the
     * <i>SyncEngineFactory</i> specified by the environment key
     * <i>syncengine/factory/bean</i>.
     *
     * @return the <i>Sync4jEngineFactory</i> object
     *
     * @throws Sync4jException
     */
    private void getSyncEngineFactory() throws Sync4jException {
        if (syncEngineFactory != null) {
            return;
        }

        String factoryName = null;
        try {
            InitialContext ctx = new InitialContext();

            factoryName = (String)ctx.lookup(ENV_ENGINE_FACTORY_NAME);

            if (log.isLoggable(Level.FINE)) {
                log.fine("factory name: " + factoryName);
                log.fine("Getting a new syncEngineFactory...");
            }

            if ((factoryName == null) || (factoryName.length() == 0)) {
                throw new Sync4jException( ENV_ENGINE_FACTORY_NAME
                                         + " must be specified"    );
            }

            syncEngineFactory =
                (SyncEngineFactory)Class.forName(factoryName).newInstance();
        } catch (Exception e) {
            log.throwing(getClass().getName(), "getSyncEngineFactory", e);
            throw new Sync4jException( "Error in loading "
                                     + factoryName
                                     , e
                                     );
        }

        //
        // The engine factory is configured with the server configuration so
        // that the engine  object has the opportunity to access to the server
        // configuration
        //
        syncEngineFactory.setConfiguration(config);
    }

    /**
     * Loads the configuration for the server. The URI from which the configuration
     * is must be set as the envirnoment property named as <i>ENV_SERVER_CONFIG_URI</i>.
     *
     * @throws ConfigurationException in case of errors.
     */
    private void loadConfiguration()
    throws ConfigurationException {
        URL configURI = null, configPath = null;

        config = new Configuration();

        try {
            InitialContext ctx = new InitialContext();

            //
            // The following two entries should be two URIs, but in the case
            // they are not (they do not start with http://, https:// or file://,
            // they are considered simple files
            //
            configURI  = fixURI((String)ctx.lookup(ENV_SERVER_CONFIG_URI ));
            configPath = fixURI((String)ctx.lookup(ENV_SERVER_CONFIG_PATH));

            config.setClassLoader(
                new ConfigClassLoader(
                    new URL[] { configPath },
                    getClass().getClassLoader()
                )
            );

            config.load(configURI.toString());

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Configuration: " + config);
            }
        } catch (ConfigurationException e) {
            throw e;
        } catch (Exception e) {
            throw new ConfigurationException("Error loading configuration from " + configURI, e);
        }
    }

    /**
     * Checks if the given string is an accepted URL (starting with http://,
     * https:// or file://). If yes, a new URL object representing the given
     * url is returned; otherwise, the given string is considered a file name
     * and a new URL is obtained calling File.toURL().
     *
     * @param s the string to check
     *
     * @return the corresponding URL if the string represents a URL or the
     *         fixed URL if the string is a pathname/filename
     *
     * @throws MalformedURLException
     */
    private URL fixURI(final String s)
    throws MalformedURLException {
        String minS = s.toLowerCase();

        try {
            return new URL(s);
        } catch (MalformedURLException e) {
            //
            // This is not a URL, let's consider it just a file
        }

        try {
            return new File(new File(s).getCanonicalPath()).toURL();

        } catch (IOException e) {
            throw new MalformedURLException("Unable to convert" + s + " to a URL");
        }
    }

    /**
     * Checks if the given message can be processed by this server implementation.<br>
     * In detail, it checks:
     * <ul>
     *  <li> if server_uri (as specified by the configuration parameter
     *       CONFIG_SERVER_URI) is a URL, the taget uri hostname and the server
     *       uri hostname must match.
     *  <li> if the server uri is not a valid url, server uri and target uri
     *       must match as strings.
     * </ul>
     *
     * @param msg the message to be checked
     *
     * @throws ServerException in case the message is wrong
     */
    private void checkMessage(Message msg)
    throws ServerException {
        String serverURI = null;
        String targetURI = null;

        serverURI = config.getStringValue(CONFIG_SERVER_URI);
        targetURI = msg.getHeader().getTarget().getURI()    ;

        try {
            URL serverURL = new URL(serverURI);
            URL targetURL = new URL(targetURI);

            serverURI = serverURL.getHost();
            targetURI = targetURL.getHost();
        } catch (MalformedURLException e) {
            //
            // The server uri is not a valid URL. Simple string comparison
            // will be performed
            //
        }

        if (log.isLoggable(Level.FINE)) {
            log.fine("serverURI: " + serverURI);
            log.fine("targetURI: " + targetURI);
        }

        if (!serverURI.equals(targetURI)) {
            String err = "The message is addressed to "
                       + targetURI
                       + " not to "
                       + serverURI
                       ;
            if (log.isLoggable(Level.INFO)) {
                log.info(err);
            }
            throw new BadRequestException(err);
        }
    }

    private static class Sync4jResponse
            implements SyncResponse
    {
        private Message msg;
        private String  resultMimeType;

        private Sync4jResponse(
                final Message msg           ,
                final String  resultMimeType) {
            this.msg = msg;
            this.resultMimeType = resultMimeType;
        }

        public Message getMessage() {
            return this.msg;
        }

        public String getMessageString() {
            return msg.toXML();
        }

        public String getMimeType() {
            return resultMimeType;
        }

        /**
         * Is this message the last message allowed for the current session?
         *
         * @return true if yes, false otherwise
         *
         */
        public boolean isCompleted() {
            return msg.isLastMessage();
        }

    }



}

