/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.server.syncbean.session;

import java.io.File;

import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.framework.tools.beans.BeanFactory;
import sync4j.framework.tools.beans.BeanException;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.server.session.AbstractSessionManager;
import sync4j.framework.server.session.SessionHandler;
import sync4j.framework.server.session.SessionExpiredException;
import sync4j.server.syncbean.session.SimpleSessionHandler;

/**
 * This implementation of a <i>AbstractSessionManager</i> stores and retrieves
 * sessions to/from the file system. <br>
 * For each session is serialized under the directory pointed by <i>sessionsPath</i>
 * in a file called {session_id}.session . This is a very simple implementation 
 * that allows many servers to access the same session repository. In fact, in
 * a real deployment, we can have SyncML servers replicated for reliability and
 * scalability purposes. In that case they must share the directory specified
 * by <i>sessionsPath</i>.
 * <p>
 * <b>This class is not thread safe. It is supposed that an instance is not
 * shared by multiple threads</b>.
 *
 * @author  Stefano Fornari @ Funambol.com
 * @version $Id: SessionManager.java,v 1.4 2003/12/15 17:07:06 luigiafassina Exp $
 */
public class SessionManager extends AbstractSessionManager implements java.io.Serializable {
    
    // ------------------------------------------------------------ Private data
    
    /**
     * The logger
     */
    private static final Logger log = Sync4jLogger.getLogger();
    
    // -------------------------------------------------------------- Properties
   
    private String sessionsPath = null;
    
    /** Getter for property sessionsPath.
     * @return Value of property sessionsPath.
     */
    public String getSessionsPath() {
        return sessionsPath;
    }
    
    /** Setter for property sessionsPath.
     * @param sessionsPath New value of property sessionsPath.
     */
    public void setSessionsPath(String sessionsPath) {
        this.sessionsPath = sessionsPath;
    }
    
    // ---------------------------------------------------------- Public methods
    
    
    // --------------------------------------------------------- Private methods
    protected Collection getSessionBag() {
        ArrayList sessions = new ArrayList();
        
        //
        // Get a list of the content of the sessionsPath directory. Than for 
        // each file deserialized it and add it to sessions.
        //
        File sessionsPathFile = new File(sessionsPath);
        
        if (!sessionsPathFile.exists()) {
            return sessions; // returns an empty collection
        }
        
        File[] files = sessionsPathFile.listFiles();
        
        SessionHandler handler = null;
        for (int i=0; (files != null) && (i<files.length); ++i) {
            try {
                handler = (SessionHandler)BeanFactory.getBeanInstance(files[i]);
                sessions.add(handler);
            } catch (BeanException e) {
                log.throwing(getClass().getName(), "getSessionBag", e.getCause());
            }
        }
        
        return sessions;
    }
    
    protected SessionHandler getSessionFromBag(String sessionId) throws SessionExpiredException {
        File sessionFile = new File(sessionsPath, sessionId + ".session");
        
        if (!sessionFile.exists()) {
            return null;
        }
        
        SessionHandler handler = null;
        try {
            handler = (SessionHandler)BeanFactory.getBeanInstance(sessionFile);
        } catch (BeanException e) {
            log.throwing(getClass().getName(), "getSessionFromBag", e.getCause());
        }
        
        return handler;
    }
    
    protected void putSessionInBag(String sessionId, SessionHandler handler) {
        File sessionFile = new File(sessionsPath, sessionId + ".session");
        
        try {
            BeanFactory.saveBeanInstance(handler, sessionFile);
        } catch (BeanException e) {
            log.severe("Error in serializing " + sessionFile);
            log.throwing(getClass().getName(), "putSessionInBag", e.getCause());
        }
    }
    
    protected void removeSessionFromBag(String sessionId) {
        File sessionFile = new File(sessionsPath, sessionId + ".session");
        
        if (sessionFile.exists()) {
            sessionFile.delete();
        }
    }

    protected SessionHandler createNewSession(String sessionId) {
        log.finest("Creating a new session with id " + sessionId);
        
        SimpleSessionHandler sessionHandler = new SimpleSessionHandler();
        
        putSessionInBag(sessionId, sessionHandler);
        
        return sessionHandler;
    }    
}