/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */

package sync4j.server.syncbean.session;

import java.util.logging.Logger;
import java.util.logging.Level;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import sync4j.framework.core.*;
import sync4j.framework.protocol.*;
import sync4j.framework.database.Database;
import sync4j.server.engine.Sync4jEngine;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncOperation;
import sync4j.framework.engine.SyncEngineFactory;
import sync4j.framework.engine.source.MemorySyncSource;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.security.SecurityConstants;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.session.SessionHandler;
import sync4j.framework.server.error.ServerException;
import sync4j.framework.server.SyncTimestamp;
import sync4j.framework.server.LastTimestamp;
import sync4j.framework.server.ClientMapping;

import sync4j.framework.engine.SyncEngine;
import sync4j.framework.engine.SyncItemState;

import sync4j.framework.security.JAASOfficer;

import sync4j.framework.server.error.InvalidCredentialsException;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.server.store.PersistentStoreException;

import sync4j.framework.server.session.SyncState;

/**
 * This class represents the handler for a SyncML session. It coordinates and
 * handles the packages and messages as dictated by the protocol.
 * <p>
 * The entry point is <i>processMessage()</i>, which determines which message is
 * expected and what processing has to be done (depending on the value of
 * <i>currentState</i>). If an error accours, the session goes to the state
 * <i>STATE_ERROR</i>; in this state no other messages but initialization can be
 * performed.
 * <p>
 * In the current implementation separate initialization is required.
 * <p>
 * <i>SessionHandler</i> makes use of a <i>SyncEngine</i> for all
 * tasks not related to the handling of the protocol.
 * See <i>sync4j.framework.engine.SyncEngine</i> for more information.
 *
 * LOG NAME: sync4j.handler
 *
 * @see sync4j.framework.engine.SyncEngine
 *
 * @author Stefano Fornari @ Funambol
 * @author Richard Wafer
 * @author Chintan Shah
 *
 * @version $Id: SimpleSessionHandler.java,v 1.56 2003/12/29 12:08:30 luigiafassina Exp $
 *
 */
public class SimpleSessionHandler implements SessionHandler, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    public static final int STATE_START                      = 0x0000;
    public static final int STATE_END                        = 0x0001;
    public static final int STATE_ERROR                      = 0xFFFF;
    public static final int STATE_INITIALIZATION_PROCESSING  = 0x0010;
    public static final int STATE_INITIALIZATION_PROCESSED   = 0x0011;
    public static final int STATE_SYNCHRONIZATION_PROCESSING = 0x0012;
    public static final int STATE_SYNCHRONIZATION_PROCESSED  = 0x0013;
    public static final int STATE_SYNCHRONIZATION_COMPLETION = 0x0014;


    // ------------------------------------------------------------ Private data

    private int currentState = STATE_START;
    private int nextState = STATE_ERROR;

    /**
     * Gets the current state
     *
     * @return the current state
     */
    public int getCurrentState() {
        return currentState;
    }

    private long creationTimestamp = -1;

    /**
     * Gets the creation timestamp of the session
     *
     * @return the creation timestamp
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }

    private transient Logger log = Sync4jLogger.getLogger("handler");

    /**
     * SyncTimestamp for the current synchronization
     */
    private SyncTimestamp nextTimestamp = null;

    private transient SyncInitialization syncInit = null;
    private transient ClientModifications modifications = null;

    /**
     * When action command received from client set to true
     */
    private boolean syncCommandFromClient = false;

    /**
     * This Map contains all the mapping for the databases involved in the
     * synchronization for the current logged principal.<br>
     * The database names are used as keys and the corresponding mapping
     * contains the luig-guid mapping. The map is created and initialized
     * in <i>getClientMappings()</i>
     */
    private Map clientMappings = null;

    /**
     * Contain the client device Id for the current session
     */
    private String clientDeviceId = null;

    /**
     * The databases that have to be synchronized. It is set in the initialization
     * process.
     */
    Database[] dbs = null;

    /**
     * The flag for established if credentials are valid
     */
    private boolean validCredentials = false;

    // -------------------------------------------------------------- Properties

    /**
     * The session id - read only
     */
    private String sessionId = null;

    public String getSessionId() {
        return this.sessionId;
    }

    /**
     * The command id generator (it defaults ro a <i>CommandIdGenerator</i> instance)
     */
    private CommandIdGenerator cmdIdGenerator = new CommandIdGenerator();

    public void setCommandIdGenerator(CommandIdGenerator cmdIdGenerator) {
        this.cmdIdGenerator = cmdIdGenerator;
    }

    public CommandIdGenerator getCommandIdGenerator() {
        return this.cmdIdGenerator;
    }

    /**
     * The cmdIdGenerator must be reset each time the process
     * of a message is starting
     */
    private void resetIdGenerator() {
        this.cmdIdGenerator.reset();
        syncEngine.setCommandIdGenerator(this.cmdIdGenerator);
    }

    /**
     * The message id generator (it defaults ro a <i>SimpleIdGenerator</i> instance)
     */
    private SimpleIdGenerator msgIdGenerator = new SimpleIdGenerator();

    /**
     * The Last message Id from the client
     */
    private String lastMsgIdFromClient = new String();

    /**
     * The factory for the <i>SyncEngine</i> object
     */
    private SyncEngineFactory syncEngineFactory = null;

    /**
     * The <i>SyncEngine</i>
     */
    private Sync4jEngine syncEngine = null;

    /**
     * Authenticated credential. If null no credential has authenticated.
     */
    private Credential      loggedCredential = null;
    private Sync4jPrincipal loggedPrincipal  = null;

    private boolean guestEnabled = false;

    /**
     * The SyncState for the syncronization process
     */
    private SyncState syncState = null;

    // ---------------------------------------------------------- Public methods

    /**
     * Sets the property <i> syncEngineFactory</i> and creates the instance of
     * the <i>SyncEngine</i> objects.
     *
     * @param syncEngineFactory the factory
     */
    public void setSyncEngineFactory(SyncEngineFactory syncEngineFactory) {
        this.syncEngineFactory = syncEngineFactory;
        try {
            this.syncEngine = (Sync4jEngine) syncEngineFactory.getSyncEngine();
        } catch (Sync4jException e) {
            log.throwing(getClass().getName(), "setSyncEngineFacytory", e);
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest(syncEngineFactory + " sync engine factory set");
            log.finest(syncEngine + " sync engine created");
        }
    }

    public SyncEngineFactory getSyncEngineFactory() {
        return this.syncEngineFactory;
    }

    public SyncEngine getSyncEngine() {
        return this.syncEngine;
    }

    /**
     * Indicates if the session is a new session
     */
    private boolean newSession = true;

    public void setNew(boolean newSession) {
        this.newSession = newSession;
    }

    public boolean isNew() {
        return this.newSession;
    }

    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new instance of SimpleSessionHandler
     */
    public SimpleSessionHandler() {
        this.creationTimestamp = System.currentTimeMillis();
    }

    /**
     * Creates a new instance of SimpleSessionHandler with a given session id
     */
    public SimpleSessionHandler(String sessionId) {
        this();
        this.sessionId = sessionId;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Returns true if the sessione has been authenticated.
     *
     * @return true if the sessione has been authenticated, false otherwise
     */
    public boolean isAuthenticated() {
        return loggedCredential != null;
    }

    /**
     * Returns true if the session has not been authenticated because of an
     * expired account.
     *
     * @return true if the account is expired
     */
    public boolean isAccountExpired() {
        JAASOfficer officer = (JAASOfficer) syncEngine.getOfficer();

        return officer.isLoginExpired();
    }

    /**
     * Processes the given message. See the class description for more
     * information.
     *
     * @param message the message to be processed
     *
     * @return the response message
     *
     * @throws ProtocolException
     */
    public Message processMessage(Message message) throws ProtocolException, InvalidCredentialsException {
        boolean syncWithInit = false;

        Message response     = null;
        Message syncResponse = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("About processing message: " + message.toXML());
        }

        //
        // Reset the cmdIdGenerator has specified in the spec
        //
        resetIdGenerator();

        //
        // Each time a message is received for a particular session adjust the message ID
        //
        msgIdGenerator.next();

        //
        //  We maintain the message Id from client
        //
        lastMsgIdFromClient = message.getHeader().getMessageID();

        //
        // Initialize the device ID from the client request
        //
        clientDeviceId = message.getHeader().getSource().getURI();

        if (log.isLoggable(Level.FINE)) {
            log.fine("current state: " + getStateName(currentState));
        }

        try {
            switch (currentState) {
                case STATE_ERROR: // in case of error you can start a new initialization
                case STATE_START:
                    nextTimestamp = new SyncTimestamp();
                    nextTimestamp.start = System.currentTimeMillis();
                    nextTimestamp.tag   = String.valueOf(nextTimestamp.start);

                    syncState = new SyncState();
                    login(message.getHeader().getCredential(),
                          clientDeviceId                     );

                    if (isAuthenticated()) {
                        readPrincipal(loggedPrincipal);
                    }

                    moveTo(STATE_INITIALIZATION_PROCESSING);

                case STATE_INITIALIZATION_PROCESSING:
                    response = processInitMessage(message);
                    if (!isAuthenticated()) {
                        moveTo(STATE_START);
                        break;
                    }

                    //
                    //	Checking for message with Sync with Initialization
                    //	If yes, set CurrentState to STATE_SYNCHRONIZATION_PROCESSING
                    //	and proceed ahead for synchornization...
                    //
                    syncWithInit = checkSyncInit(message);
                    if (syncWithInit) {
                        moveTo(STATE_INITIALIZATION_PROCESSED);
                        if (log.isLoggable(Level.FINE)) {
                            log.fine("Sync message without separate initialization");
                        }
                    } else {
                        if (log.isLoggable(Level.FINE)) {
                            log.fine("Sync with separate initalization");
                        }

                        if (message.getBody().isFinal()) {
                            moveTo(STATE_INITIALIZATION_PROCESSED);
                        }
                        break;
                    }


                case STATE_INITIALIZATION_PROCESSED:
                    moveTo(STATE_SYNCHRONIZATION_PROCESSING);

                case STATE_SYNCHRONIZATION_PROCESSING:
                    syncResponse = processSyncMessage(message);

                    if (syncWithInit) {
                        response = mergeResponse(response, syncResponse);
                    } else {
                        response = syncResponse;
                    }

                    storeClientMappings();
                    resetClientMappings();

                    if (message.getBody().isFinal()) {
                        moveTo(STATE_SYNCHRONIZATION_PROCESSED);
                        nextTimestamp.end = System.currentTimeMillis();
                        commit();
                        moveTo(STATE_SYNCHRONIZATION_COMPLETION);
                    }
                    break;

                case STATE_SYNCHRONIZATION_COMPLETION:
                    response = processCompletionMessage(message);

                    storeClientMappings();
                    resetClientMappings();
                    if (message.getBody().isFinal()) {
                        response.setLastMessage(true);
                        moveTo(STATE_END);
                    }
                    break;
            }
        } catch (ProtocolException e) {
            log.throwing(getClass().getName(), "processMessage", e);
            moveTo(STATE_ERROR);
            throw e;
        } catch (NotFoundException e) {
            log.throwing(getClass().getName(), "processMessage", e);
            moveTo(STATE_ERROR);
            throw new InvalidCredentialsException("Invalid credential error", e);
        } catch (PersistentStoreException e) {
            log.throwing(getClass().getName(), "processMessage", e);
            moveTo(STATE_ERROR);
            throw new ProtocolException("Persistent store error", e);
        } catch (Throwable t) {
            t.printStackTrace();
            log.throwing(getClass().getName(), "processMessage", t);
            moveTo(STATE_ERROR);
        }

        if (log.isLoggable(Level.FINER)) {
            log.finer("About returning message: " + response);
        }

        return response;
    }

	/**
     * Merges Initalization and Synchronization responses into a single message.
     *
	 * @param init Initialization Response Message
	 * @param sync Synchronization Response Message
     *
     * @throws sync4.framework.core.RepresentationException in case of a
     *         representation error
     *
	 *	@return the combined message
	 */
	private Message mergeResponse(Message init, Message sync)
    throws RepresentationException {

        SyncHeader header = init.getHeader();
        ArrayList mergedCmd = new ArrayList();

        AbstractCommand[] initCmd = init.getBody().getCommands();
		for(int i=0; i<initCmd.length; ++i) {
			mergedCmd.add(initCmd[i]);
        }

        if (sync != null) {
            AbstractCommand[] syncCmd = sync.getBody().getCommands();
            for(int i=0; i<syncCmd.length; ++i){
                if (syncCmd[i] instanceof StatusCommand) {
                    if ("0".equals(((StatusCommand)syncCmd[i]).getCommandRef()) &&
                        "SyncHdr".equals(((StatusCommand)syncCmd[i]).getCommandName())) {
                        continue;
                    }
                }
                mergedCmd.add(syncCmd[i]);
            }
            header = sync.getHeader();
        }

        AbstractCommand[] cmds =
            (AbstractCommand[])mergedCmd.toArray(new AbstractCommand[0]);

		SyncBody body = new SyncBody(cmds, init.getBody().isFinal());

        return (new Message(header, body));
	}

    /**
     * Processes an error condition. This method is called when the error is
     * is not fatal and is manageable at a protocol/session level. This results
     * in a well formed SyncML message with an appropriete error code.
     * <p>
     * Note that the offending message <i>msg</i> cannot be null, meaning that
     * at least the incoming message was a SyncML message. In this context,
     * <i>RepresentationException</i>s are excluded.
     *
     * @param the offending message - NOT NULL
     * @param the exception representing the error condition - NOT NULL
     *
     * @throws sync4j.framework.core.Sync4jException in case of unexpected errors
     *
     * @return the response message
     */
    public Message processError(Message msg, Throwable error)
    throws Sync4jException {
        SyncHeader msgHeader = msg.getHeader();

        Item[] items = new Item[0];
        int status = StatusCode.SERVER_FAILURE;

        if (syncEngine.isDebug()) {
            items = new Item[1];

            items[0] = new Item(
                           null, // target
                           null, // source
                           null, // meta
                           new Data(error.getMessage())
                       );
        }

        if (error instanceof ServerException) {
            status = ((ServerException)error).getStatusCode();
        }

        StatusCommand statusCommand =
            new StatusCommand(
                cmdIdGenerator.next()                ,
                msgHeader.getMessageID()            ,
                "0" /* command ref */                ,
                "SyncHdr" /* see SyncML specs */     ,
                new TargetRef(msgHeader.getTarget()),
                new SourceRef(msgHeader.getSource()),
                null /* credential */                ,
                null /* challenge */                 ,
                new Data(status)         ,
                new Item[0]
            );

        String serverURI =
            syncEngine.getConfiguration().getStringValue(syncEngine.CFG_SERVER_URI);
        SyncHeader syncHeader = new SyncHeader (
                                    new DTDVersion     ("1.1"       )         ,
                                    new ProtocolVersion("SyncML/1.1")         ,
                                    msgHeader.getSyncSessionIdentifier()      ,
                                    msgHeader.getMessageID()                  ,
                                    new Target(msgHeader.getSource().getURI()),
                                    new Source(serverURI)                     ,
                                    null  /* response URI */                  ,
                                    false                                     ,
                                    null /* credentials */                    ,
                                    null /* metadata */
                               );

        SyncBody syncBody = new SyncBody(
                                new AbstractCommand[] { statusCommand },
                                true /* final */
                            );

        moveTo(STATE_ERROR);

        return new Message(syncHeader, syncBody);
    }

    /**
     * Called by the <i>SessionManager</i> when the session is expired.
     * It logs out the credential and release aquired resources.
     */

    public void expire() {
        logout();
    }

    /**
     * Called to interrupt the processing in case of errors depending on
     * extenal causes (i.e. the transport). The current implementation just move
     * the session state to the error state.
     * <p>
     * NOTE that the current implementation simply moves the state of the session
     * to <i>STATE_ERROR</i>.
     *
     * @param statusCode the error code
     *
     * @see sync4j.framework.core.StatusCode for valid status codes
     *
     */
    public void abort(int statusCode) {
        moveTo(STATE_ERROR);
    }

    /**
     * Called to permanently commit the synchronization. It does the following:
     * <ul>
     *  <li>persists the <i>last</i> timestamp to the database for the sources
     *      successfully synchronized
     * </ul>
     */
    public void commit() {
        assert (loggedPrincipal != null);

        LastTimestamp last = null;

        PersistentStore ps = syncEngine.getStore();

        Database[] dbsGen = syncEngine.getDbs();

        for (int i = 0;  (dbsGen != null)
                      && (i < dbsGen.length)
                      && (dbsGen[i].getStatusCode() == StatusCode.OK); ++i) {
            last = new LastTimestamp(
                    loggedPrincipal.getId(),
                    dbsGen[i].getName(),
                    dbsGen[i].getAnchor().getNext(),
                    nextTimestamp.start,
                    nextTimestamp.end
            );

            if (log.isLoggable(Level.FINE)) {
                log.fine("Commiting databse "
                        + dbsGen[i].getName()
                        + " ( "
                        + last
                        + " )"
                );
            }

            try {
                boolean stored = ps.store(last);
                log.fine("LastTimeStamp stored: " + stored);
            } catch (Sync4jException e) {
                log.severe("Error in saving persistent data");
                log.throwing(getClass().getName(), "commit", e);
            }
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * If property "isGuestEnabled" is true and credential is null use the
     * credential of user guest
     *
     * @return true if the property "isGuestEnabled" is true else false
     */
    private boolean isGuestEnabled() {
        return this.guestEnabled;
    }

    /**
     * If flag "isValidCredential" is true then the login and password are correct
     * else are invalid credential
     */
    private boolean isValidCredential() {
        return this.validCredentials;
    }

    /**
     * Processes the given initialization message.
     *
     * @param message the message to be processed
     *
     * @return the response message
     *
     * @throws ProtocolException
     */
    private Message processInitMessage(Message message)
    throws ProtocolException {

        syncInit = new SyncInitialization(message.getHeader(),
                                          message.getBody()  );

        //Store the informations that will be used in the Syncronization process
        syncState.addClientCommands(syncInit.getClientCommands());
        syncState.addClientAlerts(syncInit.getClientAlerts());

        syncInit.setIdGenerator(cmdIdGenerator);
        syncInit.setFlag(Flags.FLAG_FINAL_MESSAGE);

        if (isAuthenticated()) {
            syncInit.setAuthorizedStatusCode(StatusCode.AUTHENTICATION_ACCEPTED);


            syncInit.setClientCapabilitiesRequired(false);

            //
            // Gets the databases requested for synchronization and for each of
            // them checks if the database exists on the server and if the
            // credential is allowed to synchronize it
            //
            dbs = syncInit.getDatabasesToBeSynchronized();

            //
            // This will change the status code of the elements of clientDBs to
            // reflect the availability and accessibility of the given databases
            // on the server
            //
            syncEngine.prepareDatabases(loggedPrincipal, dbs, nextTimestamp);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Requested databases: " + Arrays.asList(dbs));
            }

            boolean noDataSource = true; // there are no datasource to allowed to synchronize

            for (int i = 0; ((dbs != null) && (i < dbs.length)); ++i) {
                syncInit.setStatusCodeForCommand(
                        dbs[i].getAlertCommand(),
                        dbs[i].getStatusCode()
                );

                if (dbs[i].getStatusCode() == StatusCode.OK) {
                    noDataSource = false;
                    syncEngine.addClientSource(
                            new MemorySyncSource(
                                dbs[i].getName(),
                                null,
                                dbs[i].getSource().getURI())
                    );
                }
            }

            syncEngine.setDbs(dbs);

            //
            // Setting the databases to synchronize. This will force the relative
            // <Alert>s to be inserted in the response.
            //
            syncInit.setDatabases(dbs);

            //
            // Setting server capabilities
            //
            syncInit.setServerCapabilities(
                syncEngine.getServerCapabilities(syncInit.getDTDVersion())
            );
        } else {
            if (isAccountExpired()) {
                syncInit.setAuthorizedStatusCode(StatusCode.PAYMENT_REQUIRED);
            } else if (!isGuestEnabled() && isValidCredential()) {
                syncInit.setAuthorizedStatusCode(StatusCode.MISSING_CREDENTIALS);
            } else if (!isValidCredential()) {
                syncInit.setAuthorizedStatusCode(StatusCode.INVALID_CREDENTIALS);
            } else {
                syncInit.setAuthorizedStatusCode(StatusCode.FORBIDDEN);
            }
        }

        return syncInit.getResponse();
    }

    /**
     * Processes the given synchronization message.
     *
     * @param syncRequest the message to be processed
     *
     * @return the response message
     *
     * @throws ProtocolException
     */
    private Message processSyncMessage(Message syncRequest)
            throws ProtocolException {
        log.finest("client sources: " + syncEngine.getClientSources());

        try {
            modifications =
                    new ClientModifications(syncRequest.getHeader(),
                                            syncRequest.getBody()  ,
                                            syncEngine.getDbs());

            SyncCommand[] syncCommands = modifications.getClientSyncCommands();
            if (syncCommands == null || syncCommands.length == 0) {
                throw new ProtocolException("Sync command expected in this package but no Sync commands were found");
            }


            List responseCommands = processModifications(modifications);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("responseCommands: " + responseCommands);
            }

            modifications.setIdGenerator(cmdIdGenerator);
            modifications.setFlag(Flags.FLAG_ALL_RESPONSES_REQUIRED);

            modifications.setClientModificationsStatus(
                    (StatusCommand[]) filterCommands(
                            responseCommands,
                            new String[]{ StatusCommand.COMMAND_NAME }
                    ).toArray(new StatusCommand[0])
            );

			// .setServerModifications sets all the modification commands
			// required to be send to client from Server


			// cannot block this whole code,as this code may be used
			// by other DBS within same sync but differnt alert code.
            modifications.setServerModifications(
                    (AbstractCommand[]) filterCommands(
                            responseCommands,
                            new String[]{
                                SyncCommand.COMMAND_NAME
                            }
                    ).toArray(new AbstractCommand[0])
            );

            modifications.setFlag(Flags.FLAG_FINAL_MESSAGE);

            Message response = modifications.getResponse();

            //response.setLastMessage(
            //    syncRequest.getBody().isFinal() && noMoreResponse(response)
            //);

            return response;
        } catch (Sync4jException e) {
            throw new ProtocolException(e);
        }
    }

    /**
     * Executes the given modifications and for each of them returns a status
     * command. It forwards the execution to the synchronization engine.
     *
     * @param modifications synchronization commands containing the modifications
     *                 that the client requires
     *
     * @return an array of command objects, each containig the result of
     *         a modification or a modification itself
     * @throws Sync4jException
     */
    private List processModifications(ClientModifications modifications)
            throws Sync4jException {
        SyncHeader header           = modifications.getSyncHeader();
        String     msgId            = header.getMessageID()        ;
        boolean    headerNoResponse = header.getNoResponse()       ;

        ArrayList responseCommands = new ArrayList();

        syncEngine.setCommandIdGenerator(cmdIdGenerator);

        SyncCommand[] syncCommands = modifications.getClientSyncCommands();

        //
        // Retrieves existing LUID-GUID mapping
        //
        getClientMappings();

        //
        // First of all prepare the memory sources with the modification commands
        // receveid by the client
        //
        prepareMemorySources(syncCommands);

        syncEngine.setClientMappings(clientMappings);

        try {
			syncEngine.sync(loggedPrincipal);
        } catch (Sync4jException e) {
            log.throwing(getClass().getName(), "processModifications", e);
        }

        //
        // Status code for commands
        //
        if (headerNoResponse == false) {
            //
            // Sync commands
            //
            responseCommands.addAll(statusForSyncs(syncCommands));

            //
            // Status for server-side executed modification commands
            //
            StatusCommand[] operationStatus =
                syncEngine.getModificationsStatusCommands(msgId);

            for (int i=0; i<operationStatus.length; ++i) {
                responseCommands.add(operationStatus[i]);
            }
        }

        //
        // SyncCommands sent back to the client
        // No sync sent to client if ONE_WAY_FROM_CLIENT
        //

        AlertCommand[] alertCommands = syncState.getClientAlerts();

        Item[] alertItems = null;
        ItemizedCommand[] commands = null;
        String uri = null;
        for (int k = 0; ((alertCommands != null) && (k < alertCommands.length)); ++k) {
            //
            // If the alert represents a client only action, just ignore it
            //
            if (AlertCode.isClientOnlyCode(alertCommands[k].getAlertCode())) {
                continue;
            }

            alertItems = alertCommands[k].getItems();

            for (int i=0; ((alertItems != null) && (i < alertItems.length)); ++i) {
                uri = alertItems[i].getTarget().getURI();

                for (int s=0; ((syncCommands != null) && s<syncCommands.length); s++) {
                    if ((syncCommands[s].getTarget() != null) && !uri.equals(syncCommands[s].getTarget().getURI())) {
						continue;
                    }
                    for (int j = 0; ((syncEngine.getDbs() != null) && (j < syncEngine.getDbs().length)); ++j) {
                        if ((syncEngine.getDbs())[j].getName().equals(uri)){
                            SyncOperation[] operations = syncEngine.getSyncOperations(uri);
                            commands = syncEngine.operationsToCommands(
                                           (ClientMapping)clientMappings.get(uri),
                                           operations,
                                           uri
                                       );
                            responseCommands.add(
                                    new SyncCommand(
                                            cmdIdGenerator.next(),
                                            false,
                                            null,
                                            ProtocolUtil.source2Target(alertItems[i].getSource()),
                                            ProtocolUtil.target2Source(alertItems[i].getTarget()),
                                            null,
                                            commands
                                    )
                            );

                            //
                            // Now we can update the client-side LUID-GUID mappings
                            //
                            syncEngine.updateClientMappings(clientMappings, operations);
                        }
                    }  // next j
                }
            } // next i
        } // next k

        //
        // Resets client alerts so that they are considered processed
        //
        syncState.resetClientAlerts();

        //
        // And here we can update the server-side LUID-GUID mappings
        //
        syncEngine.updateServerMappings(clientMappings);

        return responseCommands;
    }

    /**
     * Process the completion message from the client check status for sync
     * command and update client mapping
     * @param message
     * @return the response message
     * @throws ProtocolException
     */
    private Message processCompletionMessage(Message message)
            throws ProtocolException {
        log.finest("processCompletionMessage");
        Message ret = null;

        try {
            ClientCompletion clientCompletion = new ClientCompletion(
                    message.getHeader(),
                    message.getBody()
            );

            clientCompletion.setIdGenerator(cmdIdGenerator);
            clientCompletion.setFlag(Flags.FLAG_FINAL_MESSAGE);

            //
            // IF the client completion request contains MapCommand
            //
            if (clientCompletion.isMapCommandFind()) {
                String uri = null;

                ClientMapping mapping = null;
                MapCommand[] mapCommands = clientCompletion.getMapCommands();
                MapItem[]    mapItems    = null                             ;
                for (int j=0; ((mapCommands != null) && (j<mapCommands.length)); ++j) {
                    uri = mapCommands[j].getTarget().getURI();
                    mapItems = mapCommands[j].getMapItems();

                    mapping = (ClientMapping)clientMappings.get(uri);
                    if (mapping == null) {
                        mapping = new ClientMapping(loggedPrincipal, uri);
                        clientMappings.put(uri, mapping);
                    }

                    for (int i = 0; i < mapItems.length; i++) {
                        MapItem mapItem = mapItems[i];

                        //Adding item properties to the persistent mapping
                        String GUID = mapItem.getTarget().getURI();
                        String LUID = mapItem.getSource().getURI();
                        mapping.updateMapping(LUID, GUID);
                    }
                }
            }

            ret = clientCompletion.getResponse();
        } catch (Sync4jException e) {
            log.severe("Error in process completion");
            throw new ProtocolException(e);
        }
        return ret;
    }

    /**
     * Makes a state transition. Very simple implementation at the moment: it
     * changes the value of <i>currentState</i> to the given value.
     *
     * @param state the new state
     */
    private void moveTo(int state) {
        if (log.isLoggable(Level.FINE)) {
            log.fine("moving to state " + getStateName(state));
        }
        currentState = state;
    }

    /**
     * Prepares the memory sources with the modification commands receveid
     * by the client.
     * <p>
     * Note that if the requested synchronization is a slow sync, the items are
     * inserted as "existing" items, regardless the command they belong to.
     * (maybe this will change as soon as the specification becomes clearer)
     *
     * @param syncCommands the commands used to prepare the source
     *
     */
    private void prepareMemorySources(SyncCommand[] syncCommands) {

        List sources = syncEngine.getClientSources();

        //
        // For efficiency: put the databases in a HashMap
        //
        HashMap dbMap = new HashMap();
        for (int i=0; ((syncEngine.getDbs() != null) && (i<syncEngine.getDbs().length)); ++i) {
            dbMap.put((syncEngine.getDbs())[i].getName(), (syncEngine.getDbs())[i]);
        }

        //
        // First of all prepare the memory sources with the modification commands
        // receveid by the client
        //
        int method;
        boolean slow;
        Target target = null;
        MemorySyncSource mss = null;
        AbstractCommand[] modifications = null;
        for (int i = sources.size(); i > 0; --i) {
            if (log.isLoggable(Level.FINE)) {
                log.fine("Preparing "
                        + syncEngine.getClientSources().get(i - 1)
                        + " with "
                        + Arrays.asList(syncCommands)
                );
            }

            mss = (MemorySyncSource) sources.get(i - 1);

            String  uri = mss.getSourceURI();

            modifications = new AbstractCommand[0];
            for (int j = 0; ((syncCommands != null) && (j < syncCommands.length)); ++j) {
                target = syncCommands[j].getTarget();
                if ((target != null) && (uri.equals(target.getURI()))) {
                    modifications = syncCommands[j].getCommands();

                    break;
                }
            }

            method = ((Database)dbMap.get(uri)).getMethod();
            slow = ((method == AlertCode.SLOW) || (method == AlertCode.REFRESH_FROM_SERVER));

            prepareMemorySource(mss, modifications, slow);
        }
    }

    /**
     * Prepares a source that represents the image of the client database. This
     * is done combining the existing client mapping with the modifications sent
     * by the client. Than this source can be compared with the server view of
     * the database.
     *
     * @param source the e source to prepare
     * @param commands the client modifications
     * @param slowSync true if the preparation is for a slow sync, false
     *                    otherwise
     *
     */
    private void prepareMemorySource(MemorySyncSource  source  ,
                                     AbstractCommand[] commands,
                                     boolean           slowSync) {
        ArrayList existing = new ArrayList();
        ArrayList deleted  = new ArrayList();
        ArrayList created  = new ArrayList();
        ArrayList updated  = new ArrayList();

        ClientMapping guidluid = null;

        //
        // First of all, in the case of fast sync, the items already mapped are
        // added as existing items. Note that server ids are used.
        // In the case of slow sync, the mappinngs are cleared so that they
        // won't generate conflicts.
        //
        guidluid = (ClientMapping)clientMappings.get(source.getSourceURI());
        if (slowSync) {
            if (guidluid != null) {
                guidluid.clearMappings();
            }
        } else {
            if (guidluid != null) {
                Map map = guidluid.getMapping();

                Iterator i = map.keySet().iterator();
                while (i.hasNext()) {
                    existing.add(
                        new SyncItemImpl(source, map.get(i.next()), SyncItemState.SYNCHRONIZED)
                    );
                }
            }
        }

        String name = null;
        for (int i = 0; ((commands != null) && (i < commands.length)); ++i) {
            name = commands[i].getName();
            if (slowSync && !DeleteCommand.COMMAND_NAME.equals(name)) {
                existing.addAll(
                    Arrays.asList(
                            syncEngine.itemsToSyncItems(
                                (ClientMapping)clientMappings.get(source.getSourceURI()),
                                source,
                                (ModificationCommand)commands[i],
                                SyncItemState.SYNCHRONIZED,
                                nextTimestamp.start
                            )
                    )
                );
                continue;
            }
            if (AddCommand.COMMAND_NAME.equals(commands[i].getName())) {
                created.addAll(
                        Arrays.asList(
                                syncEngine.itemsToSyncItems(
                                    (ClientMapping)clientMappings.get(source.getSourceURI()),
                                    source,
                                    (ModificationCommand)commands[i],
                                    SyncItemState.NEW,
                                    nextTimestamp.start
                                )
                        )
                );
                continue;
            }

            if (DeleteCommand.COMMAND_NAME.equals(commands[i].getName())) {
                deleted.addAll(
                        Arrays.asList(
                                syncEngine.itemsToSyncItems(
                                    (ClientMapping)clientMappings.get(source.getSourceURI()),
                                    source,
                                    (ModificationCommand)commands[i],
                                    SyncItemState.DELETED,
                                    nextTimestamp.start
                                )
                        )
                );
                continue;
            }

            if (ReplaceCommand.COMMAND_NAME.equals(commands[i].getName())) {
                updated.addAll(
                        Arrays.asList(
                                syncEngine.itemsToSyncItems(
                                    (ClientMapping)clientMappings.get(source.getSourceURI()),
                                    source,
                                    (ModificationCommand)commands[i],
                                    SyncItemState.UPDATED,
                                    nextTimestamp.start
                                )
                        )
                );
                continue;
            }
        }

        source.initialize(existing, deleted, created, updated);
    }


    /**
     * Filters a list of commands extracting the ones of the given types.
     *
     * @param commands the list of command to be filtered
     * @param types the command types to extract
     *
     * @return an array of the selected commmands
     */
    private List filterCommands(List commands, String[] types) {
        StringBuffer sb = new StringBuffer(",");

        for (int i = 0; ((types != null) && (i < types.length)); ++i) {
            sb.append(types[i]).append(',');
        }

        ArrayList selectedCommands = new ArrayList();
        AbstractCommand command = null;
        Iterator i = commands.iterator();
        while (i.hasNext()) {
            command = (AbstractCommand) i.next();

            if (sb.indexOf(',' + command.getName() + ',') >= 0) {
                selectedCommands.add(command);
            }
        }

        return selectedCommands;
    }

    /**
     * Create and return the status commands for the executed &lt;Sync&gt;s.
     *
     * @param syncCommands the Sync commands
     *
     * @return the status commands in a List collection
     */
    private List statusForSyncs(SyncCommand[] syncCommands) {
        ArrayList ret = new ArrayList();

        String uri = null;
        Target target = null;
        Source source = null;
        TargetRef targetRef = null;
        SourceRef sourceRef = null;
        int statusCode = StatusCode.OK;
        for (int i = 0; (  (syncCommands     != null )
                        && (i < syncCommands.length)); ++i) {

            target = syncCommands[i].getTarget();
            source = syncCommands[i].getSource();

            //
            // A Sync command can be empty....
            //
            uri = (target==null) ? null : target.getURI();
            if ((uri == null) || (syncEngine.getClientSource(uri) != null)) {
                statusCode = StatusCode.OK;
            } else {
                statusCode = StatusCode.NOT_FOUND;
            }

            targetRef = (target == null) ? null : new TargetRef(uri);

            sourceRef = (source == null) ? null : new SourceRef(syncCommands[i].getSource());
            ret.add(
                new StatusCommand(
                        cmdIdGenerator.next(),
                        lastMsgIdFromClient,
                        syncCommands[i].getCommandIdentifier().getValue(),
                        syncCommands[i].COMMAND_NAME,
                        targetRef,
                        sourceRef,
                        null,
                        null,
                        new Data(statusCode),
                        new Item[0]
                )
            );
        } // next i

        return ret;
    }


    /**
     * Checks that the credentials of the given message are allowed to start a
     * session.
     *
     * @param credential the message
     * @param deviceId the deviceId
     */
    private boolean login(Credential credential, String deviceId) {
        //
        // May be the credential is already logged in...
        //
        logout();

        //
        // If the credential is not specified, create a new "guest" credential
        // but only if the property isGuestEnabled is true
        //
        if (credential == null) {

            if (syncEngine.isGuestEnabled()) {
                credential = Credential.getGuestCredential();
            } else {

                validCredentials = true;
                guestEnabled = false;
                return false;
            }
        }

        Sync4jPrincipal p = Sync4jPrincipal.fromCredential(credential.getData(),
                                                           credential.getType(),
                                                           deviceId            );

        if (syncEngine.login(credential)
                && syncEngine.authorize(p, SecurityConstants.RESOURCE_SESSION)) {
            loggedCredential = credential;
            loggedPrincipal  = p         ;

            validCredentials = true;

            return true;
        }

        return false;
    }

    /**
     * Logs out the logged in credential
     */
    private void logout() {
        if (isAuthenticated()) {
            syncEngine.logout(loggedCredential);
        }
        loggedCredential = null;
        loggedPrincipal  = null;
    }

    /**
     * Called by the <i>SyncBean</i> when the container release the session.
     * It commit the change to the DB, logs out the credential and
     * release aquired resources.
     */
    public void endSession() {
        commit();
        logout();
    }


    /**
     *	Checks if the message uses a separate initialization or not.
     *  If it contains &gt;Alert&lt; and &gt;Sync&lt; tag, it implies Sync with
     *  Initialization..
     *
     *	@param message Message to be checked for Sync with Initialization.
     *	@return TRUE (if syncWithInitialization) / FALSE (if not Sync with Init)
     *	@throws ProtocolException
    */
    private boolean checkSyncInit(Message message)
    throws ProtocolException{

		// check for the message of the SyncPackage
		// if it contains <alert> & <sync> tag, it implies Sync with Initialization..
		// As the Initialization process is completed, only <sync> package is needed
		// to be checked, but <alert> tag is also checked for Cross Verification.

    	AbstractCommand[] clientCommands = message.getBody().getCommands();

        List syncs  = ProtocolUtil.filterCommands(clientCommands    ,
                                                  SyncCommand.class);
		List alerts = ProtocolUtil.filterCommands(clientCommands	,
                                                  AlertCommand.class);

        return ((syncs.size() != 0) && (alerts.size() != 0));
	}

    /**
     * Checks if a message require a response from the client.<p>
     * A message requires a response if its body contains no commands other than
     * status commands.
     *
     * @param msg the message to check - NOT NULL and properly constructed
     *
     * @return true if the message requires a response, false otherwise
     *
     */
    private boolean noMoreResponse(Message msg) {
        AbstractCommand[] commands = msg.getBody().getCommands();

        for(int i=0; ((commands != null) && (i<commands.length)); ++i) {
            if (!StatusCommand.COMMAND_NAME.equals(commands[i].getName())) {
        	return false;
        }
	}

        return true;
    }

    /**
     * Resets the clientMappings map.
     */
    private void resetClientMappings() {
        clientMappings = new HashMap();
    }

    /**
     * Retrieves the existing LUID-GUID mappings for the logged in principal.
     * The mappings for all the databases involved in the synchronization are
     * loaded.
     *
     * @return a new <i>ClientMapping</i> object containing the mapping for the
     *         given client id
     *
     * @throws Sync4jException in case of error reading the mapping from the
     *         persistent store.
     */
    private void getClientMappings()
    throws Sync4jException {

        resetClientMappings();

        ClientMapping clientMapping = null;
        String uri = null;
        try {
            PersistentStore ps = syncEngine.getStore();

            for (int i = 0; ((syncEngine.getDbs() != null) && (i<syncEngine.getDbs().length)); ++i) {

                uri = (syncEngine.getDbs())[i].getName();

                clientMapping = new ClientMapping(loggedPrincipal, uri);

                ps.read(clientMapping);

                clientMappings.put(uri, clientMapping);
            }
        } catch (PersistentStoreException e) {
            log.severe("Unable to read clientMappings from the persistent store");
            throw new Sync4jException(e);
        }
    }

    /**
     * Stores the LUIG-GUID mappings into the database
     *
     * @throws Sync4jException in case of database error
     */
    private void storeClientMappings() throws Sync4jException {
        if (clientMappings == null) {
            return;
        }

        try {
            PersistentStore ps = syncEngine.getStore();

            ClientMapping cm = null;
            Iterator i = clientMappings.keySet().iterator();
            while (i.hasNext()) {
                cm = (ClientMapping)clientMappings.get(i.next());
                if (log.isLoggable(Level.FINE)) {
                    log.fine("Saving client mapping: " + cm);
                }
                ps.store(cm);
            }
        } catch (PersistentStoreException e) {
            log.severe("Unable to save clientMappings to the persistent store");
            throw new Sync4jException(e);
        }
    }

    /**
     * Read the given principal from the store. The principal must be
     * already configured with username and device. The current implementation
     * reads the following additional informatioin:
     * <ul>
     *  <li>principal id
     * </ul>
     *
     * @param principal the principal to be read
     *
     * @throws sync4j.framework.server.store.PersistentStoreException
     */
    private void readPrincipal(Sync4jPrincipal principal)
    throws PersistentStoreException {
        assert(principal               != null);
        assert(principal.getUsername() != null);
        assert(principal.getDeviceId() != null);

        PersistentStore ps = syncEngine.getStore();

        ps.read(principal);
    }

    private void readObject(java.io.ObjectInputStream in)
            throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject();
        log = Sync4jLogger.getLogger();
    }

    private String getStateName(int state) {
        String stateName = "STATE_UNKNOWN";

        switch (state) {
            case STATE_START                      : stateName = "STATE_START"                     ; break;
            case STATE_END                        : stateName = "STATE_END"                       ; break;
            case STATE_ERROR                      : stateName = "STATE_ERROR"                     ; break;
            case STATE_INITIALIZATION_PROCESSING  : stateName = "STATE_INITIALIZATION_PROCESSING" ; break;
            case STATE_INITIALIZATION_PROCESSED   : stateName = "STATE_INITIALIZATION_PROCESSED"  ; break;
            case STATE_SYNCHRONIZATION_PROCESSING : stateName = "STATE_SYNCHRONIZATION_PROCESSING"; break;
            case STATE_SYNCHRONIZATION_PROCESSED  : stateName = "STATE_SYNCHRONIZATION_PROCESSED" ; break;
            case STATE_SYNCHRONIZATION_COMPLETION : stateName = "STATE_SYNCHRONIZATION_COMPLETION"; break;
        }

        return stateName;
    }

    public SyncState getSyncState() {
        return this.syncState;
    }

}
