/*
 Copyright (c) 2001-2002 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */
package sync4j.test.tools.ant;

import java.io.*;
import java.util.ArrayList;

import org.apache.tools.ant.Project;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.types.PatternSet;

import sync4j.framework.tools.IOTools;
import sync4j.test.TestFailedException;
import sync4j.test.tools.PostSyncML;


/**
 * This is an Ant task that executes the protocol level tests.
 * <p>
 * @see sync4j.test.tools.PostSyncML
 * @see sync4j.test.tools.CompareXml
 * <p
 * It can be used in the following manner:
 *
 * <pre>
 * &lt;task name="test-syncml" ...&gt;
 *   &lt;testsyncml url="{initial URL}" test="{test name}"&gt;
 *     &lt;patternset&gt;
          &lt;exclude name="/SyncML/SyncHdr/SessionID/self::node()[1]"/&gt;
          &lt;exclude name="/SyncML/SyncHdr/RespURI/self::node()[1]"/&gt;
        &lt;/patternset&gt;
 *   &lt;/testsyncml&gt;
 * &lt;/task&gt;
 * </pre>
 *
 * @author Stefano Fornari
 * @version $Id: TestSyncMLTask.java,v 1.4 2003/12/15 17:08:08 luigiafassina Exp $
 */
public class TestSyncMLTask extends org.apache.tools.ant.Task {
    
    // -------------------------------------------------------------- Properties
    
    protected String url = null;
    public void setUrl(String url) {
        this.url = url;
    }
    
    protected String test = null;
    public void setTest(String test) {
        this.test = test;
    }
        
    protected ArrayList xpathPatterns = new ArrayList();
    
    public void addPatternset(PatternSet set) {
        xpathPatterns.add(set);
    }
    
    // ----------------------------------------------------- Task Implementation 
    
    public void execute() throws BuildException {
        //
        // make sure we don't have an illegal set of options
        //
        validateAttributes();
        
        //
        // Deal with the specified XPath to exclude from the comparison of 
        // the XML messages. They are specified through one or more inner 
        // <patternset> elements.
        //
        ArrayList allExcludeXPaths = new ArrayList();

        PatternSet ps = null;
        String[] xpaths = null;
        for (int i = 0; i < xpathPatterns.size(); i++) {
            ps = (PatternSet) xpathPatterns.get(i);
            xpaths = ps.getExcludePatterns(project);
            for (int j = 0; j < xpaths.length; ++j) {
                allExcludeXPaths.add(xpaths[j]);
            }
        }  // next i
        
        log("Initial URL: "    + url             );
        log("Test: "           + test            );
        log("Ignored XPaths: " + allExcludeXPaths);
        
        xpaths = (String[])allExcludeXPaths.toArray(new String[0]);
        
        //
        // Execute the tests now
        //
        FilenameFilter filter = IOTools.getFileTypeFilter(".xml");
        
        String[] msgFiles = new File(test).list(filter);

        try {
            new PostSyncML(
                url                                 , 
                new File(project.getBaseDir(), test), 
                msgFiles                            , 
                xpaths
            ).syncAndTest();

            log("Test " + test + " passed!");
        } catch (IOException e) {
            log(e.getMessage(), Project.MSG_ERR);
            throw new BuildException("Error executing PostSyncMLTask", e);
        } catch (TestFailedException e) {
            log("Test " + test + " failed: " + e.getMessage(), Project.MSG_INFO);
        }
    }
    
    // ------------------------------------------------------- Protected methods
    
    protected void validateAttributes() throws BuildException {
        try {
            new java.net.URL(url); 
        } catch (Exception e) {
            new BuildException("Malformed url exception: " + url);
        }
    }

}
