/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 */


package sync4j.transport.http.client;

import sync4j.framework.core.Constants;
import sync4j.framework.core.Message;
import sync4j.framework.core.Constants;
import sync4j.framework.core.RepresentationException;
import sync4j.framework.core.Sync4jException;

import sync4j.framework.tools.WBXMLTools;

import java.net.*;
import java.io.*;

/**
 *
 * @author Sean C. Sullivan
 * @author Stefano Fornari @ Funambol
 *
 * @version $Id: HttpClientConnection.java,v 1.12 2003/06/17 13:55:46 stefano_fornari Exp $
 *
 */
public final class HttpClientConnection
        implements sync4j.framework.client.IClientConnection
{
    //
    // todo : decide if HttpURLConnection is a good implementation strategy...
    // The other way to implement this would be to use a java.net.Socket
    // The problem with java.net.Socket is that I'd have to write alot
    // of low-level HTTP protocol code.  The advantage of java.net.Socket
    // is that I could explicitly set a Socket timeout value. Hmm....
    //
    // Also, another way to implement this class would be to use Apache's
    // HttpClient code:  http://jakarta.apache.org/commons/httpclient/
    //

    private final HttpURLConnection             m_conn;
    private       OutputStream                   m_out;
    private final String            m_strServerAddress;
    
    private       String                  lastResponse;  // as string 
    private       String                   lastMessage;  // as string


    /**
     *
     * @param strServerAddress must be non-null
     *
     */
    public HttpClientConnection(final String strServerAddress)
    throws IOException
    {
        if (strServerAddress == null)
        {
            throw new NullPointerException("strAddress parameter is null");
        }

        m_strServerAddress = strServerAddress;

        if (m_strServerAddress.startsWith("http:") == false)
        {
            throw new IllegalArgumentException(
                    "server address must start with 'http:'");

        }
        
        URL u = null;
        try
        {
            u = new URL(strServerAddress);
        }
        catch (java.net.MalformedURLException ex)
        {
            throw new IllegalArgumentException(
                    "server address is not a valid URL");
        }
        
        lastMessage = lastResponse = null;
        
        Object obj = u.openConnection();
        if ((obj instanceof java.net.HttpURLConnection) == false)
        {
            // todo : throw exception here
        }
        m_conn = (HttpURLConnection) obj;
        m_conn.setDoOutput(true);
        m_conn.setDoInput(true);
        m_conn.setAllowUserInteraction(false);
        m_conn.setRequestMethod("POST");
        m_conn.setUseCaches(false);
        m_conn.setInstanceFollowRedirects(false);
        m_conn.setRequestProperty("User-Agent", this.getClass().toString());
    }

    public Message sendMessage(final Message msg) throws Exception {
        return sendMessage(msg.toXML());
    }
    
    public Message sendMessage(final String msg)
    throws IOException, RepresentationException
    {
        final byte[] yaData = msg.getBytes();
        
        m_conn.setRequestProperty("Content-Type", Constants.MIMETYPE_SYNCML_XML);
        m_conn.setRequestProperty("Content-Length", "" + yaData.length);
        
        m_out = m_conn.getOutputStream();
        m_out.write(yaData);
        m_out.flush();

        InputStream in = m_conn.getInputStream();

        final int iResponseCode = m_conn.getResponseCode();

        if (iResponseCode != HttpURLConnection.HTTP_OK)
        {
            String error = "Response status: " 
                         + iResponseCode 
                         + ", Response message: " 
                         + m_conn.getResponseMessage()
                         ;
            throw new IOException(error);
        }

        final String strResponseContentType = m_conn.getContentType();

        if (strResponseContentType == null)
        {
            throw new IOException("Content type: " + strResponseContentType);
        }

        if (strResponseContentType.equals(Constants.MIMETYPE_SYNCML_XML) == false)
        {
            throw new IOException( "Content type: " 
                                 + strResponseContentType
                                 + " (should be "
                                 + Constants.MIMETYPE_SYNCML_XML
                                 + ")"
                                 );
        }
        
        lastMessage = msg;

        final int iResponseContentLength = m_conn.getContentLength();

        if (iResponseContentLength < 1)
        {
           throw new IOException("Response content length: " + iResponseContentLength);
        }

        final byte[] yaResponse = new byte[iResponseContentLength];

        int n = 0;
        int iBytesRead = 0;
        do
        {
            n = in.read(yaResponse,
                    iBytesRead,
                    yaResponse.length - iBytesRead);
            if (n > 0)
            {
                iBytesRead += n;
            }
        } while (n != -1);

        if (iBytesRead != iResponseContentLength)
        {
            // todo : throw exception - ?
        }
        
        lastResponse = new String(yaResponse);

        return new Message(lastResponse);
    }

    public Message sendWBXMLMessage(final byte[] msg)
    throws IOException, Sync4jException, RepresentationException
    {
        final byte[] yaData = msg;

        m_conn.setRequestProperty("Content-Type", Constants.MIMETYPE_SYNCML_WBXML);
        m_conn.setRequestProperty("Content-Length", "" + yaData.length);

        m_out = m_conn.getOutputStream();
        m_out.write(yaData);
        m_out.flush();

        InputStream in = m_conn.getInputStream();

        final int iResponseCode = m_conn.getResponseCode();

        if (iResponseCode != HttpURLConnection.HTTP_OK)
        {
            String error = "Response status: "
                         + iResponseCode
                         + ", Response message: "
                         + m_conn.getResponseMessage()
                         ;
            throw new IOException(error);
        }

        final String strResponseContentType = m_conn.getContentType();

        if (strResponseContentType == null)
        {
            throw new IOException("Content type: " + strResponseContentType);
        }

        if (strResponseContentType.equals(Constants.MIMETYPE_SYNCML_WBXML) == false)
        {
            throw new IOException( "Content type: "
                                 + strResponseContentType
                                 + " (should be "
                                 + Constants.MIMETYPE_SYNCML_WBXML
                                 + ")"
                                 );
        }

        final int iResponseContentLength = m_conn.getContentLength();

        if (iResponseContentLength < 1)
        {
           throw new IOException("Response content length: " + iResponseContentLength);
        }

        final byte[] yaResponse = new byte[iResponseContentLength];

        int n = 0;
        int iBytesRead = 0;
        do
        {
            n = in.read(yaResponse,
                    iBytesRead,
                    yaResponse.length - iBytesRead);
            if (n > 0)
            {
                iBytesRead += n;
            }
        } while (n != -1);

        if (iBytesRead != iResponseContentLength)
        {
            // todo : throw exception - ?
        }

        Message msgResponse = null;

        //TRN - convert yaResponse from WBXML to XML string, then create the Message
        String xmlResponse = WBXMLTools.wbxmlToXml(yaResponse);

        msgResponse = new Message(xmlResponse);

        return msgResponse;
    }

    public void close()
    {
        if (m_conn != null)
        {
            m_conn.disconnect();
            try
            {
                m_out.close();
            }
            catch (java.io.IOException ex)
            {
                // ignore exception
            }
        }
    }

    public String toString()
    {
        return m_strServerAddress;
    }
    
    public String getLastMessage() {
        return lastMessage;
    }
    
    public String getLastResponse() {
        return lastResponse;
    }
}
