/*
 Copyright (c) 2001-2003 sync4j project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.

 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:

     "This product includes software developed by the
      sync4j project."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 */


package sync4j.transport.http.server;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.StringTokenizer;

import javax.naming.*;
import javax.ejb.CreateException;
import javax.ejb.EJBException;
import javax.servlet.*;
import javax.servlet.http.*;

import sync4j.framework.core.*;
import sync4j.framework.tools.beans.BeanFactory;
import sync4j.framework.tools.Base64;
import sync4j.framework.server.*;
import sync4j.framework.tools.WBXMLTools;
import sync4j.framework.transport.http.SyncHolder;
import sync4j.framework.protocol.ProtocolException;
import sync4j.server.syncbean.SyncHomeLocal;
import sync4j.server.syncbean.SyncLocal;

import org.apache.commons.lang.StringUtils;

/**
 *  Receives the HTTP request and does session management.
 *  
 *  <b>Session Management</b>
 *  <p>
 *  If the url contains a parameter {PARAM_SESSION_ID}={SESSION_ID}, the session
 *  id is used to lookup in the <i>handlerCache</i> if there is already a 
 *  <i>SyncHolder</i> associated with the given session id it is used to process
 *  the incoming request. Otherwise (either if the session parameter is not
 *  specified or not cached handler is found), a new session id is created and a
 *  new <i>SyncHolder</i> object is instantiated and stored in the cache.<br>
 *  The session id is created as followed:
 *  <p>
 *  <ol>
 *    <li> the first four bytes contains the IP address of the remote client
 *    <li> a dash ('-') is appended
 *    <li> the creation timestamp is appended
 *    <li> the resulting string is encoded base 64
 *  </ol>
 *  Note that no session expiration is handled by this class. That is delegated 
 *  to the <i>SyncHandler</i> object.
 *
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: Sync4jHttpServlet.java,v 1.26 2004/01/13 08:05:02 luigiafassina Exp $
 *
 */
public final class Sync4jHttpServlet
        extends javax.servlet.http.HttpServlet
{

    // --------------------------------------------------------------- Constants
    
    public static final String PARAM_SESSION_ID       = "sid"              ;
    public static final String PARAM_SYNCHOLDER_CLASS = "sync-holder-class";
    
    // ------------------------------------------------------------ Private data
    
    private static String  syncHolderClass = null;
    private static HashMap holderCache     = null;
    
    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {
        ServletConfig config = getServletConfig();
        
        syncHolderClass = config.getInitParameter(PARAM_SYNCHOLDER_CLASS);
        
        if (StringUtils.isEmpty(syncHolderClass)) {
            String msg = "The servlet configuration parameter " 
                       + PARAM_SYNCHOLDER_CLASS
                       + " cannot be empty ("
                       + syncHolderClass
                       + ")"
                       ;
            log(msg);
            throw new ServletException(msg);
        }
        
        holderCache = new HashMap(97);
    }
    
    
    public void doPost(
            final HttpServletRequest httpRequest,
            final HttpServletResponse httpResponse)
            throws ServletException, IOException
    {

        final String contentType = httpRequest.getContentType();
        final int contentLength = httpRequest.getContentLength();
        
        log("contentType: " + contentType);
        log("contentLength: " + contentLength);
        
        if (contentLength < 1) {
            throw new Error("Content length < 1 (" + contentLength + ")");
        }

        byte[] requestData = new byte[contentLength];

        InputStream in = httpRequest.getInputStream();
        int n = 0;
        int bytesRead = 0;

        try {
            do {
                n = in.read(requestData,
                            bytesRead,
                            requestData.length - bytesRead);
                if (n > 0) {
                    bytesRead += n;
                }
            } while (n != -1);
        } catch (IOException ex) {
            handleError(httpRequest, httpResponse, "Error reading the request", ex);
            return;
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (Exception ex) {
                handleError(httpRequest, httpResponse, ex.getClass().getName(), ex);
                return;
            }
            in = null;
        }

        if (bytesRead != contentLength) {
            handleError(httpRequest, httpResponse, "bytesRead != contentLength", null);
            return;
        }

        //
        // If the session id is not specified in the URL, a new remote object 
        // will be created. Otherwise the session id specifies which remote 
        // object shall handles the request.
        //
        SyncHolder holder = null;
        try {
            holder = createHolder(httpRequest);
        } catch (Exception e) {
            handleError(httpRequest, httpResponse, "Error creating SyncBean", e);
            return;
        }
        
        SyncResponse resp = null;
        try {
            resp = holder.processMessage(requestData, contentType);
        } catch (Exception e) {
            Throwable cause = e.getCause();
            
            if (  (cause != null) 
               && (  (cause instanceof ProtocolException)
                  || (cause instanceof Sync4jException  )
                  )
               ) {
                handleError(httpRequest, httpResponse, "Protocol error", cause);
                return;
            } else {
                throw new ServletException(e);
            }
        }
        
        //
        // Set the <RespURI> tag to the server uri plus the session id parameter
        //
        Message responseMessage = resp.getMessage();
        responseMessage.getHeader().setResponseURI(
            httpRequest.getScheme()                              +
            "://"                                                +
            httpRequest.getServerName()                          +
            ':'                                                  +
            httpRequest.getServerPort()                          +
            httpRequest.getContextPath()                         +
            httpRequest.getServletPath()                         +
            '?'                                                  +
            PARAM_SESSION_ID                                     +
            '='                                                  +
            holder.getSessionId()
        );
        
        OutputStream out = null;
        try {
            out = httpResponse.getOutputStream();
            httpResponse.setContentType(resp.getMimeType());

            byte[] response = null;
            if (Constants.MIMETYPE_SYNCML_XML.equals(resp.getMimeType()))
            {
                log("Sending back XML");
                response = responseMessage.toXML().getBytes();
                httpResponse.setContentLength(response.length);
                out.write(response);
            } else if (Constants.MIMETYPE_SYNCML_WBXML.equals(resp.getMimeType())){
                log("Sending back WBXML");
                try {
                    response = WBXMLTools.toWBXML(responseMessage);
                } catch (Sync4jException e) {
                    handleError(httpRequest, httpResponse, "Encoding error", e);
                }
                httpResponse.setContentLength(response.length);
                out.write(response);
            }
            
            out.flush();
        } finally {
            log("Finally");
            if (out != null) out.close();
        }
        
        //
        // If the message completed the SyncML communication, the holder
        // must be closed and discarded.
        //
        if (resp.isCompleted()) {
            releaseHolder(holder);
        }
    }

    // --------------------------------------------------------- Private methods

    
    /** 
     * Factory method for <i>SyncHolder</i> objects. If the session id is
     * passed as a CGI parameter, it is transformed in a EJB handle. If the 
     * handle is not valid, it is considered expired. If the session id is not
     * specified, a new EJB is created
     *
     * @param request the associated HTTP request object
     *
     * @return a new <i>SyncHolder</i>
     *
     */
    private SyncHolder createHolder(HttpServletRequest request) 
    throws Exception {
        String sessionId = request.getParameter(PARAM_SESSION_ID);
        
        SyncHolder holder = null;
        
        if (!StringUtils.isEmpty(sessionId)) {
            holder = (SyncHolder)holderCache.get(sessionId);
        }
        
        if (holder == null) {
            holder = (SyncHolder)BeanFactory.getBeanInstance(
                getClass().getClassLoader(), syncHolderClass
            );
            
            sessionId = createSessionId(request);
            holder.setSessionId(sessionId);
            holderCache.put(sessionId, holder);
        }
        
        return holder;
    }
    
    /**
     * Closes the given <i>SyncHolder</i> and removes it from the cache.
     *
     * @param holder the holder to releases
     */
    private void releaseHolder(SyncHolder holder) {
        try {
            holder.close();
        } catch (Exception e) {
            log(e.getMessage(), e.getCause());
        }
        holderCache.remove(holder.getSessionId());
    }
    
    /**
     * Handles errors conditions returning an appropriate content to the client.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest   request,
                             final HttpServletResponse response,
                                   String                   msg,
                             final Throwable                  t) {
                             
        //
        // TO DO: returning an appropriate HTTP status code/error message
        //
        if (msg == null) {
            msg = "";
        }
        
        if (t == null) {
           log(msg);
        } else {
            log(msg, t);
        }
        try {
            response.sendError(response.SC_BAD_REQUEST, msg);
        } catch (IOException e) {
            log(e.getMessage(), e);
        }
    }
    
    /**
     * Creates the session id (see the class description for details).
     *
     * @param request the HTTP request object
     *
     * @return a newly created session id
     */
    private String createSessionId(HttpServletRequest request) {
        String clientIP  = request.getRemoteAddr()   ;
        long   timestamp = System.currentTimeMillis();
        
        StringBuffer sb = new StringBuffer();
        
        byte[] ip = null;
        
        try {
            ip = InetAddress.getByName(clientIP).getAddress();
        } catch (UnknownHostException e) {
            ip = new byte[] {0, 0, 0, 0};
        }
        
        sb.append(ip).append('-').append(timestamp);
        
        String stringSessionId = new String(Base64.encode(sb.toString().getBytes()));
        
        //
        // strips out tailing '='
        //
        int i = stringSessionId.lastIndexOf('=');
       
        return ((i>0) ? stringSessionId.substring(0, i-1) : stringSessionId);
    }

    /**
     * The same as <i>handleError(reqest, response, msg, t, null)</i>.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest   request,
                             final HttpServletResponse response,
                             final String                   msg) {
        this.handleError(request, response, msg, null);
    }
}
