/* 
Copyright (c) 2001-2002 sync4j project
All rights reserved.
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
 
1. Redistributions of source code must retain the above copyright
notice, this list of conditions, and the following disclaimer.
 
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions, and the disclaimer that follows
these conditions in the documentation and/or other materials
provided with the distribution.
 
3. The name "sync4j" must not be used to endorse or promote products
derived from this software without prior written permission.
 
4. Products derived from this software may not be called "sync4j", nor
may "sync4j" appear in their name, without prior written permission.
 
In addition, we request (but do not require) that you include in the
end-user documentation provided with the redistribution and/or in the
software itself an acknowledgement equivalent to the following:
 
"This product includes software developed by the
sync4j project."
 
THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
 
 */

package sync4j.transport.j2me.http.client;

import sync4j.framework.client.ClientConnectionInfo;
import sync4j.framework.IConnectionFactory;
import sync4j.framework.IConnection;
import sync4j.framework.core.Message;

import javax.microedition.io.HttpConnection;
import java.io.*;

/**
 * A class for Http client connection.  It assumes http-specific behavior,
 * that means, it is not using StreamConnection but HttpConnection.
 *
 * Note:
 * 1. A client can force HTTP 1.0 behavior by including 
 *  the "Connection: close" header. In this behavior, the server closes 
 *  the socket when it is done with the request. 
 * 2. For applications that make multiple round trip requests to 
 *  the same server, you can achieve the best performance by taking 
 *  advantage of the HTTP 1.1 persistent connections. 
 *  This is particularly true if the requests are made over real wireless 
 *  networks where the bandwidth is slow and the latency is very high. 
 *
*  @author	<a href="mailto:sso@users.sourceforge.net">Simon So</a>
 * @since   SyncML 1.1
 * @see javax.microedition.io.HttpConnection
 *
 * @version $Id: ClientConnectionImpl.java,v 1.2 2003/01/15 15:41:16 stefano_fornari Exp $
 */
public final class ClientConnectionImpl implements IConnection {
    private final HttpConnection conn;
    InputStream input;
    OutputStream output;
    
    /** 
     * Creates an instance of ClientConnectionImpl.
     */
    public ClientConnectionImpl(Object obj) {
        conn = (HttpConnection)obj;
        input = null;
        output = null;
    }
        
    /**
     * Opens the input/output stream.  <code>setRequestMethod</code> must be 
     * done before calling this method.
     */
    public void prepareConnection() throws IOException {                
            // maybe I should implement all SyncClientAgent protocol specific.
            conn.setRequestMethod("POST");
            conn.setRequestProperty("Cache-Control", "no-store");
            conn.setRequestProperty("User-Agent", this.getClass().toString());
            conn.setRequestProperty("Content-Type", "application/vnd.syncml+xml");            
            
            // The call to setRequestProperty() in connect() turns off 
            // HTTP persistent connections. By default, the client may 
            // reuse outgoing connections to the same server. This does not
            // work correctly with the J2MEWTK emulator when connections
            // are interrupted. Passing a Connection: close header in the 
            // HTTP request turns off this feature.                 
            conn.setRequestProperty("Connection", "close");        
    }
    
    public void send(final Message msg) throws IOException {
        conn.setRequestProperty("Content-Length", "" + msg.toString().length());
        output = conn.openOutputStream();
        output.write(msg.toString().getBytes());
        //output.flush();  
    }
    
    /**
     * Receives from a HttpConnection.  Courtesy of 
     * <a href="http://wireless.java.sun.com/midp/questions/calcbyte/">Ed Ort.</a>
     *
     * @return the bytes from a http server.
     */
    public Message receive() throws IOException {
        input = conn.openInputStream();
        int length = (int)conn.getLength();
        byte[] b = null;
        if (length == -1L) {
            // Reading from an HTTP 1.0 server or a chunked HTTP 1.1
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            int c ;
            while (true) {
                c = input.read();
                if (c == -1)
                    break;
                baos.write(c);
            }
            b = baos.toByteArray();
        } else {
            // Reading a Content-Length labeled payload
             b = new byte[length];
             input.read(b);
        }
        return new Message(new String(b));
    }
    
    public void closeConnection() throws IOException {
        if (input != null) {
            input.close();
        }
        if (output != null) {
            output.close();
        }
        if (conn != null) {
            conn.close();
        }
    }
}
